%%%----------------------------------------------------------------------
%%% File    : hipe_coalesce.erl
%%% Author  : Thorild Seln <d95ths@zeppo.it.uu.se>
%%% Purpose : Combine and join temporaries with each other where
%%%           possible.  We can do this when we find a copy
%%%           instruction.
%%% Created :  18 Mar 2000 by Thorild Seln <d95ths@zeppo.it.uu.se>
%%%----------------------------------------------------------------------

-module(hipe_coalesce).
-export([coalesce/8]).
-author('Thorild Seln').


%%%----------------------------------------------------------------------
% Function:    coalesce
%
% Description: Coalesces nodes in worklist
% Parameters:
%   Moves         -- Current move information
%   IG            -- Interference graph
%   Worklists     -- Current worklists
%   Node_sets     -- Current node information
%   Alias         -- Current aliases for temporaries
%   Select_stack  -- Nodes selected for coloring
%   K             -- Number of registers
%   
% Returns:
%   {Moves, IG, Worklists, Node_sets, Alias}
%                    (Updated versions of above structures,
%                    after coalescing)
%
%%%----------------------------------------------------------------------


coalesce(Moves, IG, Worklists, Node_sets, Alias, Select_stack, K, Target) ->
    case hipe_moves:worklist(Moves) of
	[] ->
	    hipe_message:error(coalesce, 'No moves in worklist');
	[Move|Rest] ->
	    {move, Dest, Source} = Move,

	    hipe_message:debug(coalesce, "Testing nodes ~p and ~p for coalescing~n",[Dest,Source]),

	    Alias_src = hipe_alias:getAlias(Source, Node_sets, Alias),
	    Alias_dst = hipe_alias:getAlias(Dest, Node_sets, Alias),
	    {U, V} = case Target:is_precolored(Alias_dst) of
			 true -> {Alias_dst, Alias_src};
			 false -> {Alias_src, Alias_dst}
		     end,
	    Moves0 = hipe_moves:remove(worklist, Move, Moves),
	    Degree0 = hipe_ig:degree(IG),
	    if U == V ->
		    Moves1 = hipe_moves:add(coalesced, Move, Moves0),
		    Worklists0 = 
			add_worklist(Worklists, U, K, Moves1, Degree0, Target),
		    {Moves1,
		     IG,
		     Worklists0,
		     Node_sets,
		     Alias};
	       true ->
		    case Target:is_precolored(V) or 
			hipe_adj_set:adjacent(U, V, hipe_ig:adj_set(IG)) of 
			true ->
			    Moves1 = 
				hipe_moves:add(constrained, Move, Moves0),
			    Worklists1 = 
				add_worklist(add_worklist(Worklists,
							  U,
							  K,
							  Moves1,
							  Degree0,
							  Target),
					     V,
					     K,
					     Moves1,
					     Degree0,
					     Target),
			    {Moves1,
			     IG,
			     Worklists1,
			     Node_sets,
			     Alias};
			false ->
			    Adj_list = hipe_ig:adj_list(IG),
			    Coalesced_nodes = hipe_node_sets:coalesced(Node_sets),
			    AdjU = hipe_coalescing_regalloc:adjacent(U,
						     Adj_list,
						     Coalesced_nodes,
						     Select_stack),
			    AdjV = hipe_coalescing_regalloc:adjacent(V,
						     Adj_list,
						     Coalesced_nodes,
						     Select_stack),

			    case (Target:is_precolored(U)
				  and all_adjacent_ok(AdjV, U, IG, K, Target))
				or (not(Target:is_precolored(U))
				    and (conservative(ordsets:union(AdjU, AdjV),
						      IG,
						      K))) of
			       true ->
				    Moves1 = hipe_moves:add(coalesced, Move, 
							  Moves0),
				    {IG0,
				     Node_sets0,
				     Worklists1,
				     Moves2,
				     Alias0} = combine(U,
						       V,
						       IG,
						       Node_sets,
						       Worklists,
						       Moves1,
						       Alias,
						       Select_stack,
						       K,
						       Target),
				    

				    Degree1 = hipe_ig:degree(IG0),
				    Worklists2 = add_worklist(Worklists1,
							      U,
							      K,
							      Moves2,
							      Degree1,
							      Target),
				    {Moves2,
				     IG0,
				     Worklists2,
				     Node_sets0,
				     Alias0};
			       false ->
				    {hipe_moves:add(active, Move, Moves0),
				     IG,
				     Worklists,
				     Node_sets,
				     Alias}
			    end
		    end
	    end
    end.



%%%----------------------------------------------------------------------
% Function:    add_worklist
%
% Description: Builds new worklists where U is transferred from freeze
%              to simplify, if possible
%
% Parameters:
%   Worklists     -- Current worklists
%   U             -- Node to operate on
%   K             -- Number of registers
%   Moves         -- Current move information
%   Degree        -- Degree information from interference graph
%   Target        -- The containing the target-specific functions
%   
% Returns:
%   Worklists (updated)
%
%%%----------------------------------------------------------------------

add_worklist(Worklists, U, K, Moves, Degree, Target) ->
    case (not(Target:is_precolored(U))
	  and not(hipe_moves:move_related(U, Moves))
	  and (hipe_degree:is_simple(U, K, Degree))) of
	true ->
	    hipe_reg_worklists:transfer(freeze, simplify, U, Worklists);
	false ->
	    Worklists
    end.

			    			   
	
%%%----------------------------------------------------------------------
% Function:    combine
%
% Description: Combines two nodes into one (used when coalescing)
%
% Parameters:
%   U             -- First node to operate on
%   V             -- Second node to operate on
%   IG            -- Interference graph
%   Node_sets     -- Current node information
%   Worklists     -- Current worklists
%   Moves         -- Current move information
%   Alias         -- Current aliases for temporaries
%   Select_stack  -- Nodes selected for coloring
%   K             -- Number of registers
%
% Returns:
%   {IG, Node_sets, Worklists, Moves, Alias} (updated)
%
%%%----------------------------------------------------------------------

	       
combine(U, V, IG, Node_sets, Worklists, Moves, Alias, Select_stack, K,
	Target) ->
    Worklists1 = case hipe_reg_worklists:member(freeze, V, Worklists) of
		     true -> hipe_reg_worklists:remove(freeze, V, Worklists);
		     false -> hipe_reg_worklists:remove(spill, V, Worklists)
		 end,
    Node_sets1 = hipe_node_sets:add(coalesced, V, Node_sets),

    hipe_message:debug(coalesce, "Coalescing ~p and ~p to ~p~n",[V,U,U]),

    Alias1 = hipe_alias:setAlias(V, U, Alias),

    % NOTE: Here there is an error in the pseudocode for the algorithm!
    Moves1 = hipe_moves:set_movelist(hipe_vectors:set(hipe_moves:movelist(Moves),
					   U,
					   ordsets:union(hipe_moves:node_moves(U, Moves),
						      hipe_moves:node_moves(V, Moves))),
				Moves),

    Adj_list = hipe_ig:adj_list(IG),
    Adjacent = hipe_coalescing_regalloc:adjacent(V,
				 Adj_list,
				 hipe_node_sets:coalesced(Node_sets1),
				 Select_stack),

    {IG1, Worklists2, Moves2} =  combine_edges(Adjacent,
					       U,
					       IG,
					       Node_sets1,
					       Worklists1,
					       Moves1,
					       Select_stack,
					       K,
					       Target),

    New_worklists = case (not(hipe_degree:is_simple(U, K, hipe_ig:degree(IG1)))
			  and hipe_reg_worklists:member(freeze, U, Worklists2)) of
			true -> hipe_reg_worklists:transfer(freeze, spill, U, 
						   Worklists2);
			false -> Worklists2
		    end,
    {IG1, Node_sets1, New_worklists, Moves2, Alias1}.



%%%----------------------------------------------------------------------
% Function:    combine_edges
%
% Description: For each node in a list, make an edge between that node
%              and node U, and decrement its degree by 1
%              (Used when two nodes are coalesced, to connect all nodes
%              adjacent to one node to the other node)
%
% Parameters:
%   [T|Ts]        -- List of nodes to make edges to
%   U             -- Node to make edges from
%   IG            -- Interference graph
%   Node_sets     -- Current node information
%   Worklists     -- Current worklists
%   Moves         -- Current move information
%   Select_stack  -- Nodes selected for coloring
%   K             -- Number of registers
%   
% Returns:
%   {IG, Worklists, Moves} (updated)
%
%%%----------------------------------------------------------------------

combine_edges([], U, IG, Node_sets, Worklists, Moves, Select_stack, K, 
	      Target) -> 
    {IG, Worklists, Moves};
combine_edges([T|Ts], U, IG, Node_sets, Worklists, Moves, Select_stack, K,
	     Target) ->
    IG1 = hipe_ig:add_edge(T, U, IG, Target),
    {IG2, Worklists1, Moves1} = hipe_coalescing_regalloc:decrement_degree([T],
							  IG1,
							  Node_sets,
							  Worklists,
							  Moves,
							  Select_stack,
							  K),
    combine_edges(Ts, U, IG2, Node_sets, Worklists1, Moves1, Select_stack, K,
		  Target).


%%%----------------------------------------------------------------------
% Function:    ok
%
% Description: Checks if a node T is suitable to coalesce with R
%
% Parameters:
%   T             -- Node to test
%   R             -- Other node to test
%   IG            -- Interference graph
%   K             -- Number of registers
%   Target        -- The module containing the target-specific functions
%   
% Returns:
%   true iff coalescing is OK
%
%%%----------------------------------------------------------------------

ok(T, R, IG, K, Target) ->
    ((hipe_degree:is_simple(T, K, hipe_ig:degree(IG)))
     or Target:is_precolored(T)
     or hipe_adj_set:adjacent(T, R, hipe_ig:adj_set(IG))).
     

%%%----------------------------------------------------------------------
% Function:    all_ok
%
% Description: True iff, for every T in the list, OK(T,U)
%
% Parameters:
%   [T|Ts]        -- Nodes to test
%   U             -- Node to test for coalescing
%   IG            -- Interference graph
%   K             -- Number of registers
%   Target        -- The module containing the target-specific functions
%   
% Returns:
%   true iff coalescing is OK for all nodes in the list
%
%%%----------------------------------------------------------------------

all_adjacent_ok([], U, IG, K, Target) -> false;
all_adjacent_ok([T|Ts], U, IG, K, Target) ->
    ok(T, U, IG, K, Target) and all_adjacent_ok(Ts, U, IG, K, Target).


%%%----------------------------------------------------------------------
% Function:    conservative
%
% Description: Checks if nodes can be safely coalesced according to
%              the Briggs' conservative coalescing heuristic
%
% Parameters:
%   Nodes         -- Adjacent nodes
%   IG            -- Interference graph
%   K             -- Number of registers
%   
% Returns:
%   true iff coalescing is safe
%
%%%----------------------------------------------------------------------

conservative(Nodes, IG, K) ->
    conservative_count(Nodes, hipe_ig:degree(IG), K, 0) < K.


%%%----------------------------------------------------------------------
% Function:    conservative_count
%
% Description: Counts degrees for conservative (Briggs' heuristic)
%
% Parameters:
%   Nodes         -- (Remaining) adjacent nodes
%   Degree        -- Degree information from interference graph
%   K             -- Number of registers
%   Cnt           -- Accumulator for counting
%   
% Returns:
%   Final value of accumulator
%
%%%----------------------------------------------------------------------

conservative_count([], Degree, K, Cnt) -> Cnt;
conservative_count([Node|Nodes], Degree, K, Cnt) ->
    case hipe_degree:is_simple(Node, K, Degree) of
	true -> conservative_count(Nodes, Degree, K, Cnt);
	false -> conservative_count(Nodes, Degree, K, Cnt+1)
    end.
