%%%----------------------------------------------------------------------
%%% File    : hipe_coallescing_regalloc.erl
%%% Author  : Andreas Wallin <d96awa@zeppo.it.uu.se>
%%% Purpose : Play paintball with registers on target machine. We win
%%%           if they are all colored. This is an optimistic iterated
%%%           register allocator.
%%% Created :  4 Mar 2000 by Andreas Wallin <d96awa@zeppo.it.uu.se>
%%%----------------------------------------------------------------------

-module(hipe_coalescing_regalloc).
-author(['Andreas Wallin', 'Ingemar berg']).
-export([regalloc/3, 
	 regalloc/4, 
	 adjacent/4, 
	 decrement_degree/7]).

%%%----------------------------------------------------------------------
% Function:    regalloc
%
% Description: Creates a K coloring for a function.
% Parameters:
%   CFG           --  A control flow graph
%   SpillIndex    --  Last index of spill variable
%   SpillLimit    --  Just for compability with old register allocator.
%   Target        --  The module containing the target-specific functions.
%
% Returns:
%   Coloring      --  A coloring for specified CFG
%   SpillIndex0   --  A new spill index
%%%----------------------------------------------------------------------

% The old register allocator doesn't know about other targets than sparc
% so we are keeping this 'regalloc/3' function a while 

regalloc(CFG, SpillIndex, SpillLimit) ->
    regalloc(CFG, SpillIndex, SpillLimit, hipe_sparc_specific).


%%%----------------------------------------------------------------------
% Function:    regalloc
%
% Description: Creates a K coloring for a function.
% Parameters:
%   CFG           --  A control flow graph
%   SpillIndex    --  Last index of spill variable
%   SpillLimit    --  Just for compability with old register allocator.
%   Target        --  Name of target specific module. This module 
%                      contains all target specific functions we need.
%
% Returns:
%   Coloring      --  A coloring for specified CFG
%   Spill_index0  --  A new spill index
%%%----------------------------------------------------------------------

regalloc(CFG, SpillIndex, SpillLimit, Target) ->

    % Build interference graph
    hipe_message:debug(regalloc, "Team Orvars register allocator was called~n", []),
    IG = hipe_ig:build(CFG, Target),
    {Min_temporary, Max_temporary} = Target:var_range(CFG),
    All_colors = Target:allocatable(),
    K = length(All_colors),
    Node_sets = hipe_node_sets:new(Target, Min_temporary, Max_temporary),

    %% Add registers with their own coloring
    Move_sets = hipe_moves:new(IG),
    Worklists = hipe_reg_worklists:new(IG, Node_sets, Move_sets, K),
    Select_stack = [],
    Alias = hipe_alias:initAlias(Max_temporary),
    {IG0, Worklists0, Moves0, Alias0, Node_sets0, Select_stack0} = 
	do_coloring(IG, 
		    Worklists, 
		    Node_sets,
		    Move_sets, 
		    Alias,
		    Select_stack, 
		    K,
		    SpillLimit,
		    Target),
    {Color0,Node_sets1} = 
	hipe_assign:defaultColoring(hipe_node_sets:precolored(Node_sets0),
			       hipe_color:init(Max_temporary),
			       Node_sets0, Target),
    {Color1,Node_sets2} = hipe_assign:assignColors(Select_stack0,Node_sets1,
					      Color0, Alias0,All_colors,
					      Target),
    build_namelist(Node_sets2,SpillIndex,Alias0,Color1).



%%----------------------------------------------------------------------
% Function:    do_coloring
%
% Description: Create a coloring. That is, play paintball.
% Parameters:
%   IG             --  An interference graph
%   Worklists      --  Worklists, that is simplify, spill and freeze
%   Node_sets      --  Node sets, that is spilled, coalesced and so on.
%   Moves          --  Moves sets, that is coalesced, constrained 
%                       and so on.
%   Alias          --  Tells if two temporaries can have their value
%                       in the same register.
%   Select_stack   --  Stack where simplifyed nodes are pushed.
%   K              --  Want to create a K coloring.
%   SpillLimit     --  Try not to spill nodes that are above the spill limit.
%
% Returns:
%   IG             --  Updated interference graph
%   Worklists      --  Updated Worklists structure
%   Moves          --  Updated Moves structure 
%   Alias          --  Updates Alias structure
%   Node_sets      --  Updated Node_sets structure
%   Select_stack   --  Updated Select_stack.
%   
%%----------------------------------------------------------------------
do_coloring(IG, Worklists, Node_sets, Moves, Alias, Select_stack, K, 
	    SpillLimit, Target) ->
    Simplify = not(hipe_reg_worklists:is_empty(simplify, Worklists)),
    Coalesce = not(hipe_moves:is_empty(worklist, Moves)),
    Freeze   = not(hipe_reg_worklists:is_empty(freeze, Worklists)),
    Spill    = not(hipe_reg_worklists:is_empty(spill, Worklists)),
    if Simplify == true ->
	    {IG0, Worklists0, Moves0, Select_stack0} = 
		simplify(hipe_reg_worklists:simplify(Worklists),
			 IG, 
			 Node_sets,
			 Worklists, 
			 Moves, 
			 Select_stack, 
			 K),
	    do_coloring(IG0, Worklists0, Node_sets, Moves0, Alias,
			Select_stack0, K, SpillLimit,Target);
       Coalesce == true ->
	    {Moves0, IG0, Worklists0, Node_sets0, Alias0} = 
		hipe_coalesce:coalesce(Moves, IG, Worklists, Node_sets, Alias, 
				  Select_stack, K, Target),
	    do_coloring(IG0, Worklists0, Node_sets0, Moves0, Alias0, 
			Select_stack, K, SpillLimit,Target);
       Freeze == true ->
	    {Worklists0,Moves0} = 
		hipe_freeze:freeze(K, Worklists, Moves, hipe_ig:degree(IG), 
			      Node_sets, Alias),
	    do_coloring(IG, Worklists0, Node_sets, Moves0, Alias, 
			Select_stack, K, SpillLimit, Target);
       Spill == true ->
	    {Worklists0, Moves0} = 
		hipe_reg_sel:selectSpill(Worklists, Moves, IG,
					 K, Node_sets, Alias, SpillLimit),
	    do_coloring(IG, Worklists0, Node_sets, Moves0, Alias, 
			Select_stack, K, SpillLimit,Target);
       true -> % Catchall case
	    {IG, Worklists, Moves, Alias, Node_sets, Select_stack}
    end.

%%%----------------------------------------------------------------------
% Function:    adjacent
%
% Description: Adjacent nodes that's not coalesced, on the stack or
%               precolored.
% Parameters:
%   Node           --  Node that you want to adjacents of
%   Adj_list       --  An adjacent nodes list (created in IG)
%   Coalesced      --  Nodes that are ready for coalesced
%   Select_stack   --  Nodes that we think will get a coloring
%
%   Returns: 
%     A set with nodes/temporaries that are not coalesced, on the 
%      stack or precolored.
%%%----------------------------------------------------------------------
adjacent(Node, Adj_list, Coalesced, Select_stack) ->
    Adjacent_edges = hipe_adj_list:edges(Node, Adj_list),
    Removed_coalesced = ordsets:subtract(Adjacent_edges, Coalesced),
    Stack_nodes = stack_nodes(Select_stack, []),
    ordsets:subtract(Removed_coalesced, ordsets:from_list(Stack_nodes)).


%%----------------------------------------------------------------------
% Function:    stack_nodes
%
% Description: Extract the nodes only from Stack_elements.
% Parameters:
%   Stack_elements --  A list of stack elements
%   Node_list      --  An empty list
%
%   Returns: 
%     A list of nodes found on stack.
%%----------------------------------------------------------------------
stack_nodes([], Node_list) -> Node_list;
stack_nodes([{Node, _}|Stack_elements], Node_list) -> 
    stack_nodes(Stack_elements, [Node|Node_list]).



%%----------------------------------------------------------------------
% Function:    simplify
%
% Description: Simplify graph by removing nodes of low degree. This
%               function simplify all nodes it can at once.
% Parameters:
%   [Node|Nodes]   --  The simplify worklist
%   IG             --  The interference graph
%   Node_sets      --  The node_sets data-structure
%   Worklists      --  The worklists data-structure
%   Moves          --  The moves data-structure
%   Select_stack   --  The stack data-structure
%   K              --  Produce a K coloring
%
%   Returns: 
%     IG            --  An updated interference graph
%     Worklists     --  An updated worklists data-structure
%     Moves         --  An updated moves data-structure
%     Select_stack  --  An updated stack data-structure
%%----------------------------------------------------------------------
simplify([], IG, Node_sets, Worklists, Moves, Select_stack, K) -> 
    {IG, Worklists, Moves, Select_stack};
simplify([Node|Nodes], IG, Node_sets, Worklists, Moves, Select_stack, K) ->
    Worklists0 = hipe_reg_worklists:remove(simplify, Node, Worklists),
    Adj_nodes = hipe_adj_list:edges(Node, hipe_ig:adj_list(IG)),
    Select_stack0 = [{Node, Adj_nodes} | Select_stack],
    Adjacent = adjacent(Node, 
			hipe_ig:adj_list(IG), 
			hipe_node_sets:coalesced(Node_sets), 
			Select_stack0),
    {New_ig, Worklists1, New_moves} = decrement_degree(Adjacent, 
						       IG, 
						       Node_sets,
						       Worklists0, 
						       Moves, 
						       Select_stack0,
						       K),
    simplify(Nodes, New_ig, Node_sets, Worklists1, New_moves, Select_stack0, 
	     K).




%%%----------------------------------------------------------------------
% Function:    decrement_degree
%
% Description: Decrement the degree on a number of nodes/temporaries.
% Parameters:
%   [Node|Nodes]   --  Decrement degree on these nodes
%   IG             --  The interference graph
%   Node_sets      --  The Node_sets data structure
%   Worklists      --  The Worklists data structure
%   Moves          --  The Moves data structure.
%   Select_stack   --  Nodes that we think will get a coloring
%   K              --  We want to create a coloring with K colors
%
%   Returns: 
%     IG         --  An updated interference graph (the degrees)
%     Worklists  --  Updated Worklists. Changed if one degree goes
%                     down to K.
%     Moves      --  Updated Moves. Changed if a move related temporary
%                     gets degree K.
%%%----------------------------------------------------------------------
decrement_degree([], IG, Node_sets, Worklists, Moves, Select_stack, K) -> 
    {IG, Worklists, Moves};
decrement_degree([Node|Nodes], IG, Node_sets, Worklists, Moves, Select_stack, 
		 K) ->
    Degree0 = hipe_ig:degree(IG),
    Degree1 = hipe_degree:dec(Node, Degree0),
    IG0 = hipe_ig:set_degree(Degree1, IG),
    case (hipe_degree:degree(Node, Degree0) == K) of
	true ->
	    Adjacent = adjacent(Node, 
				hipe_ig:adj_list(IG0), 
				hipe_node_sets:coalesced(Node_sets),
				Select_stack),
	    Moves0 = enable_moves(ordsets:add_element(Node, Adjacent), Moves),
	    Worklists0 = hipe_reg_worklists:remove(spill, Node, Worklists),
	    case hipe_moves:move_related(Node, Moves0) of
		true ->
		    Worklists1 = hipe_reg_worklists:add(freeze, Node, Worklists0),
		    		    decrement_degree(Nodes, 
						     IG0, 
						     Node_sets, 
						     Worklists1, 
						     Moves0, 
						     Select_stack, 
						     K);
		_ ->
		    Worklists1 = hipe_reg_worklists:add(simplify, Node, Worklists0),
		    decrement_degree(Nodes, 
				     IG0, 
				     Node_sets, 
				     Worklists1, 
				     Moves0, 
				     Select_stack, 
				     K)
	    end;
	_ ->
	    decrement_degree(Nodes, 
			     IG0, 
			     Node_sets, 
			     Worklists, 
			     Moves, 
			     Select_stack, 
			     K)

    end.
		    
%%----------------------------------------------------------------------
% Function:    enable_moves
%
% Description: Make (move) nodes thats not yeat considered for
%               coalescing to ready for possible coalescing.
%               
% Parameters:
%   [Node|Nodes]   --  A list of move nodes
%   Moves          --  The moves data-structure
%
%   Returns: 
%     An updated moves data-structure
%%----------------------------------------------------------------------
enable_moves([], Moves) -> Moves;
enable_moves([Node|Nodes], Moves) ->
    Node_moves = hipe_moves:node_moves(Node, Moves),
    New_moves = enable_moves_active_to_worklist(Node_moves, Moves),
    enable_moves(Nodes, New_moves).

%%----------------------------------------------------------------------
% Function:    enable_moves_active_to_worklist
%
% Description: Make (move) nodes thats not yeat considered for
%               coalescing to ready for possible coalescing.
%               
% Parameters:
%   [Node|Nodes]   --  A list of move nodes
%   Moves          --  The moves data-structure
%
%   Returns: 
%     An updated moves data-structure
%%----------------------------------------------------------------------
enable_moves_active_to_worklist([], Moves) -> Moves;
enable_moves_active_to_worklist([Node|Nodes], Moves) ->
    case ordsets:is_element(Node, hipe_moves:active(Moves)) of
	true ->
	    New_moves = hipe_moves:add(worklist,
				     Node,
				     hipe_moves:remove(active, Node, Moves)),
	    enable_moves_active_to_worklist(Nodes, New_moves);
	_ ->
	    enable_moves_active_to_worklist(Nodes, Moves)
    end.


%%% Build the namelists, these functions are fast hacks, they use knowledge 
%%% about data representation that they shouldn't know, bad abstraction.

build_namelist(NodeSets,Index,Alias,Color) ->
    {alias,AliasVector} = Alias,
    AliasList = build_alias_list(hipe_vectors:vector_to_list(AliasVector),1,[]),
    NL1 = build_coalescedlist(AliasList,NodeSets,Color,Alias,[]),
    NL2 = build_reglist(hipe_node_sets:colored(NodeSets),Color,NL1),
    build_spillist(hipe_node_sets:spilled(NodeSets),Index,NL2).

build_spillist([],Index,List) ->
    {List,Index};
build_spillist([Node|Nodes],Index,List) ->
    hipe_message:debug(regalloc, "Spill, (~p to ~p)~n",[Node,Index]),
    build_spillist(Nodes,Index+1,[{Node,{spill,Index}}|List]).

build_coalescedlist([],NodeSets,Color,Alias,List) ->
    List;
build_coalescedlist([Node|Nodes],NodeSets,Color,Alias,List) 
  when integer(Node) ->
    AC = hipe_color:getColor(hipe_alias:getAlias(Node,NodeSets,Alias),Color),
    build_coalescedlist(Nodes,NodeSets,Color,Alias,[{Node,{reg,AC}}|List]);
build_coalescedlist([Node|Nodes],NodeSets,Color,Alias,List) ->
    build_coalescedlist(Nodes,NodeSets,Color,Alias,List).

build_reglist([],Color,List) -> 
    List;
build_reglist([Node|Nodes],Color,List) ->
    build_reglist(Nodes,Color,[{Node,{reg,hipe_color:getColor(Node,Color)}}|List]).

build_alias_list([],I,List) ->
    List;
build_alias_list([Alias|Aliases],I,List) when integer(Alias) ->
    build_alias_list(Aliases,I+1,[I|List]);
build_alias_list([Alias|Aliases],I,List) ->
    build_alias_list(Aliases,I+1,List).

