%%%----------------------------------------------------------------------
%%% File    : hipe_freeze.erl
%%% Author  : Ingemar berg <d95ina@zeppo.it.uu.se>
%%% Purpose : Freeze moves that can't be coalesced.
%%% Created : 20 Mar 2000 by Ingemar berg <d95ina@zeppo.it.uu.se>
%%%----------------------------------------------------------------------

-module(hipe_freeze).
-author('d95ina@zeppo.it.uu.se').
-export([freeze/6, 
	 freezeMoves/7]).

%%%----------------------------------------------------------------------
% Function:    freeze
%
% Description: When both simplifying and coalescing is impossible we 
%              rather freezes a node in stead of spilling, this function
%              selects a node for freezing (it just picks the first one in
%              the list)
%
% Parameters:
%   K              -- The number of available registers
%   WorkLists      -- A datatype containing the different worklists
%   Moves          -- A datatype containing the different movelists
%   Degrees        -- A Datatype containing the degrees of the nodes
%   NodeSets       -- A Datatype containing the different NodeSets
%   Alias          -- An alias mapping, shows the alias of all coalesced 
%                      nodes  
%
% Returns:
%   WorkLists      -- The updated worklists
%   Moves          -- The updated movelists
%
%%%----------------------------------------------------------------------

freeze(K,WorkLists,Moves,Degrees,NodeSets,Alias) ->
    [U|_] = hipe_reg_worklists:freeze(WorkLists),         % Smarter routine?
    hipe_message:debug(freeze, "freezing node ~p~n",[U]),
    WorkLists0 = hipe_reg_worklists:transfer(freeze, simplify, U, WorkLists),
    freezeMoves(U,K,WorkLists0,Moves,Degrees,NodeSets,Alias).

%%%----------------------------------------------------------------------
% Function:    freezeMoves
%
% Description: Make all move related interferences for a certain node 
%              into ordinary interference arcs.
%              
% Parameters:
%   U              -- The node we want to freeze
%   K              -- The number of available registers
%   WorkLists      -- A datatype containing the different worklists
%   Moves          -- A datatype containing the different movelists
%   Degrees        -- A Datatype containing the degrees of the nodes
%   NodeSets       -- A Datatype containing the different NodeSets
%   Alias          -- An alias mapping, shows the alias of all coalesced 
%                     nodes  
%
% Returns:
%   WorkLists      -- The updated worklists
%   Moves          -- The updated movelists
%
%%%----------------------------------------------------------------------

freezeMoves(U,K,WorkLists,Moves,Degrees,NodeSets,Alias) ->
    Nodes = hipe_moves:node_moves(U, Moves),
    freezeEm(U,Nodes,K,WorkLists,Moves,Degrees,NodeSets,Alias).

%%% Find what the other value in a copy instruction is, return false if 
%%% the instruction isn't a move with the first argument in it.

moves (U,{move,U,V}) ->
    V;
moves(U,{move,V,U}) ->
    V;
moves(U,Node) ->
    false.

freezeEm(U,[],K,WorkLists,Moves,Degrees,NodeSets,Alias) -> 
    {WorkLists,Moves};
freezeEm(U,[M|Ms],K,WorkLists,Moves,Degrees,NodeSets,Alias) ->
    case moves(U,M) of
	false ->
	    freezeEm(U,Ms,K,WorkLists,Moves,Degrees,NodeSets,Alias);
	V ->
	    {WorkLists2,Moves2} = freezeEm2(U,V,M,K,WorkLists,
				 Moves,Degrees,NodeSets,Alias),
	    freezeEm(U,Ms,K,WorkLists2,Moves2,Degrees,NodeSets,Alias)
    end.

freezeEm2(U,V,M,K,WorkLists,Moves,Degrees,NodeSets,Alias) ->
    case hipe_moves:member(active, M, Moves) of
	true ->
	    Moves1 = hipe_moves:remove(active, M, Moves),
	    freezeEm3(U,V,M,K,WorkLists,Moves1,Degrees,NodeSets,Alias);	
	false ->
	    Moves1 = hipe_moves:remove(worklist, M, Moves),
	    freezeEm3(U,V,M,K,WorkLists,Moves1,Degrees,NodeSets,Alias)
    end.

freezeEm3(U,V,M,K,WorkLists,Moves,Degrees,NodeSets,Alias) ->
    Moves1 = hipe_moves:add(frozen,M,Moves),
    V1 = hipe_alias:getAlias(V,NodeSets,Alias),
    %% We know that hipe_moves:node_moves/2 returns an ordset (a list).
    case (hipe_moves:node_moves(V1,Moves1) == []) and 
	hipe_degree:is_simple(V1,K,Degrees) of
	true ->
	    hipe_message:debug(freeze, "freezing move between ~p and ~p(~p)~n",
			  [U,V,V]),
	    Worklists1 = hipe_reg_worklists:transfer(freeze, simplify, V1, WorkLists),
	    {Worklists1, Moves1};
	false ->
	    {WorkLists,Moves1}
    end.

%%% END help functions.
