%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Copyright (c) 2001 by Erik Johansson.  All Rights Reserved 
%% Time-stamp: <01/04/10 15:04:02 happi>
%% ====================================================================
%%  Filename : 	hipe_mk_switch.erl
%%  Module   :	hipe_mk_switch
%%  Purpose  :  To implement switching.
%%  Notes    : 
%%  History  :	* 2001-02-28 Erik Johansson (happi@csd.uu.se): 
%%               Created.
%%  CVS      :
%%              $Author: happi $
%%              $Date: 2001/04/10 22:42:26 $
%%              $Revision: 1.3 $
%% ====================================================================
%%  Exports  :
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_rtl_mk_switch).
-export([gen_switch_val/5,gen_switch_tuple/5]).
%-------------------------------------------------------------------------
-include("../main/hipe.hrl").
%-------------------------------------------------------------------------
-define(MINFORJUMPTABLE,4).
-define(WORDSIZE, 4).

gen_switch_val(I, VarMap, ConstTab, Options, ExitInfo) ->
  case property_lists:get_bool(use_jumptable, Options) of
    false -> gen_slow_switch_val(I, VarMap, ConstTab, Options, ExitInfo);
    true -> gen_fast_switch_val(I, VarMap, ConstTab, Options, ExitInfo)
  end.

%% (By Happi)
%%
%% NOTE: 
%%  This is mostly experimental code to test the handling of jumptables
%%  in the const table and in the rest of the code generation...
%%
%% TODO: 
%%  Make sure it can handle all types of values thrown at it.
%%  There should be some better heuristics determining when to use
%%  a jump-table.
%%  Use the tagscheme module to do untagging.
%%  (XXX: OBS: This is hardcoded in this testversion... !!!)
%%
gen_fast_switch_val(I, VarMap, ConstTab, Options, ExitInfo) ->
  {Arg, VarMap0} = 
    hipe_rtl_varmap:icode_var2rtl_var(hipe_icode:switch_val_arg(I), VarMap),
  IcodeFail = hipe_icode:switch_val_fail_label(I),
  {Fail, VarMap1} = 
    hipe_rtl_varmap:icode_label2rtl_label(IcodeFail, VarMap0),

  %% Important that the list of cases is sorted...
  Cases = lists:sort(hipe_icode:switch_val_cases(I)),

  case length(Cases) < ?MINFORJUMPTABLE of
    true -> 
      %% No need to make a jumptable for only MINFORJUMPTABLE args
      %% TODO: Find out the optimal MINFORJUMPTABLE.
      gen_slow_switch_val(I, VarMap, ConstTab, Options, ExitInfo);
    _ -> %% More than 4 cases
      %% ... to get the smallest (integer) in the list ...
      {First,_} = hd(Cases),
      Min = hipe_icode:const_value(First),
      %% ... and to make dense intervals work.

      %% The function dense_interval throws dense_interval if not all 
      %% constants in the jump-table are integers.
      %% Otherwise it makes a sparse set of integers dense.
      case catch dense_interval(Cases, Min, IcodeFail) of
	not_an_integer -> %% Can only handle ints so far...
	  gen_search_switch_val(I, Arg, Cases, 
				Fail, VarMap, ConstTab, Options);
	way_to_sparse -> %% There was a very big gap in the table
	  gen_slow_switch_val(I, VarMap, ConstTab, Options, ExitInfo);
	{_, _, false} ->  %% The table is not dense enough.
	  gen_slow_switch_val(I, VarMap, ConstTab, Options, ExitInfo);

	{Max, DenseTbl, true} -> 
	  %% The switch only contains integers... and it is dense

	  %% DenseTbl now contains a dense mapping from all integers 
	  %% between Min to Max to a label.

	  %% Map is a rtl mapping of Dense
	  {Map,VarMap2} = lbls_from_cases(DenseTbl,VarMap1),

	  %% Make some labels and registers that we need.
	  BelowLab = hipe_rtl:mk_new_label(),
	  UntaggedR = hipe_rtl:mk_new_reg(),
	  StartR = hipe_rtl:mk_new_reg(),

	  %% Generate the code to do the switch...
	  {[
	    %% Untag the index.
	    %% 
	    %% OBS:
	    %% TODO: Use the tagscheme module for this!
	    hipe_rtl:mk_alu(UntaggedR, Arg, sra, hipe_rtl:mk_imm(4))|
	    %% Check that the index is within Min and Max.
	    case Min of
	      0 -> %% First element is 0 this is simple.
		[hipe_rtl:mk_branch(UntaggedR, gtu, hipe_rtl:mk_imm(Max),
				     hipe_rtl:label_name(Fail), 
				     hipe_rtl:label_name(BelowLab) , 0.01),
		 BelowLab,
		 %% StartR contains the index into the jumptable
		 hipe_rtl:mk_switch(UntaggedR, Map)];
	      _ -> %% First element is not 0 
		[hipe_rtl:mk_alu(StartR, UntaggedR, sub, 
				 hipe_rtl:mk_imm(Min)), 
		 hipe_rtl:mk_branch(StartR, gtu, hipe_rtl:mk_imm(Max-Min),
				    hipe_rtl:label_name(Fail), 
				    hipe_rtl:label_name(BelowLab) , 0.01),
		 BelowLab,
		 %% StartR contains the index into the jumptable
		 hipe_rtl:mk_switch(StartR, Map)]
	    end],
	    VarMap2,
	    ConstTab}

      end
  
  end.

%-------------------------------------------------------------------------
%% switch_val without jumtable
%%

gen_slow_switch_val(I, VarMap, ConstTab, Options, ExitInfo) ->
  Is = rewrite_switch_val(I),
  ?IF_DEBUG_LEVEL(3,?msg("Switch: ~w\n",[Is]),no_debug),
  hipe_icode2rtl:translate_instrs(Is, VarMap, ConstTab, Options, ExitInfo).

rewrite_switch_val(I) ->
    Arg = hipe_icode:switch_val_arg(I),
    Fail = hipe_icode:switch_val_fail_label(I),
    Cases = hipe_icode:switch_val_cases(I),
    rewrite_switch_val_cases(Cases, Fail, Arg).

rewrite_switch_val_cases([{C,L}|Cases], Fail, Arg) ->
    Tmp = hipe_icode:mk_new_var(),
    NextLab = hipe_icode:mk_new_label(),
    [hipe_icode:mk_mov(Tmp, C),
     hipe_icode:mk_if(op_exact_eqeq_2, [Arg, Tmp], L,
		      hipe_icode:label_name(NextLab)),
     NextLab |
     rewrite_switch_val_cases(Cases, Fail, Arg)];
rewrite_switch_val_cases([], Fail, Arg) ->
    [hipe_icode:mk_goto(Fail)].



%-------------------------------------------------------------------------
%% switch_val with binarysearch jumtable
%%

gen_search_switch_val(I, Arg, Cases, Default, VarMap, ConstTab, Options) ->
  ValTableR = hipe_rtl:mk_new_reg(),

  {Values,Labels} = split_cases(Cases),
  {NewConstTab,Id} =
       hipe_consttab:insert_sorted_block(ConstTab, 
			      Values),
  {LabMap,VarMap1} = lbls_from_cases(Cases,VarMap),
  
  Code = 
    [hipe_rtl:mk_load_address(ValTableR, Id, constant)|
     tab(Values,LabMap,Arg,ValTableR,Default)],
  {Code, VarMap1, NewConstTab}.




%-------------------------------------------------------------------------
%%
%% tab/6 creates rtl code for a binary search.
%% It requieres two sorted tables one with the keys to search
%% through and one with the corresponding labels to jump to.
%%
%% The implementation is derived from John Bentlys
%% Programming Pearls.
%%
%% Input: 
%%  KeyList - A list of keys to search through. 
%%           (Just used to calculate the number of elements.)
%%  LableList - A list of labels to jump to.
%%  KeyReg - A register containing the key to search for.
%%  TablePntrReg - A register containing a pointer to the
%%                tables with keys
%%  Default - A lable to jump to if the key is not found.
%%
%% Example:
%%  KeyTbl: < a, b, d, f, h, i, z >
%%  Lbls:   < 5, 3, 2, 4, 1, 7, 6 >
%%  Default: 8
%%  KeyReg: v37
%%  TablePntrReg: r41 
%%
%%  should give code like:
%%    r41 <- KeyTbl
%%    r42 <- 0
%%    r43 <- [r41+16]
%%    if (r43 gt v37) then L17 (0.50) else L16
%% L16:
%%    r42 <- 16
%%    goto L17
%% L17: 
%%    r46 <- r42 add 16
%%    r45 <- [r41+r46]
%%    if (r45 gt v37) then L21 (0.50) else L20
%% L20: 
%%    r42 <- r46
%%    goto L21
%% L21: 
%%    r48 <- r42 add 8
%%    r47 <- [r41+r48]
%%    if (r47 gt v37) then L23 (0.50) else L22
%% L22: 
%%    r42 <- r48
%%    goto L23
%% L23: 
%%    r50 <- r42 add 4
%%    r49 <- [r41+r50]
%%    if (r49 gt v37) then L25 (0.50) else L24
%% L24: 
%%    r42 <- r42 add 4
%%    goto L25
%% L25: 
%%    if (r42 gt 28) then L6 (0.50) else L18
%% L18: 
%%    r44 <- [r41+r42]
%%    if (r44 eq v37) then L19 (0.90) else L8
%% L19: 
%%    r42 <- r42 sra 2
%%    switch (r42) <L5, L3, L2, L4, L1, 
%%              L7, L6>

%%   
%% The search is done like a rolled out binary search,
%% but instead of starting in the middle we start at 
%% the power of two closest above the middle.
%%
%% We let IndexReg point to the lower bound of our
%% search, and then we speculatively look at a 
%% position at IndexReg + I where I is a power of 2.
%%
%% Example: Looking for 'h' in 
%%  KeyTbl: < a, b, d, f, h, i, z >
%%
%%  We start with IndexReg=0 and I=4
%%          < a, b, d, f, h, i, z >
%%            ^        ^
%%     IndexReg    +   I
%%
%%  'f' < 'h' so we add I to IndexReg and divide I with 2
%%  IndexReg=4 and I=2
%%          < a, b, d, f, h, i, z >
%%                     ^     ^
%%              IndexReg +  I
%%
%%  'i' > 'h' so we keep IndexReg and divide I with 2
%%  IndexReg=4 and I=1
%%          < a, b, d, f, h, i, z >
%%                     ^  ^
%%              IndexReg+ I
%%  Now we have found 'h' so we add I to IndexReg -> 5
%%  And we can load switch to the label at position 5 in
%%  the label table.
%%
%%  Now since the wordsize is 4 all numbers above are 
%%  Multiples of 4.




tab(KeyList, LabelList, KeyReg, TablePntrReg, Default) ->
  %% io:format("Making switch\n",[]),

  %% Calculate the size of the table: 
  %%  the number of keys * wordsize.
  LastOffset = (length(KeyList)-1)*?WORDSIZE, 

  %% Calculate the power of two closest to the size of the table.
  Pow2 = 
    trunc(math:pow(2,(trunc(math:log(LastOffset) / math:log(2))))),

  %% Create some registers an lables that we need.
  IndexReg = hipe_rtl:mk_new_reg(),
  Temp = hipe_rtl:mk_new_reg(),
  Temp2 = hipe_rtl:mk_new_reg(),
  Lab1 = hipe_rtl:mk_new_label(),
  Lab2 = hipe_rtl:mk_new_label(),
  Lab3 = hipe_rtl:mk_new_label(),
  Lab4 = hipe_rtl:mk_new_label(),
  
  %% Calculate the position to start looking at.
  Init = (LastOffset)-Pow2,

  %% Create the code.
  [hipe_rtl:mk_move(IndexReg,hipe_rtl:mk_imm(0)),
   hipe_rtl:mk_load(Temp,TablePntrReg,hipe_rtl:mk_imm(Init)),
   hipe_rtl:mk_branch(Temp, ge, KeyReg,
		      hipe_rtl:label_name(Lab2), 
		      hipe_rtl:label_name(Lab1) , 0.5),
   Lab1,
   hipe_rtl:mk_alu(IndexReg, IndexReg, add, 
		   hipe_rtl:mk_imm(Init+?WORDSIZE)),
   hipe_rtl:mk_goto(hipe_rtl:label_name(Lab2)),
   Lab2] ++

    step(Pow2 div 2,TablePntrReg,IndexReg,KeyReg) ++

    [hipe_rtl:mk_branch(IndexReg, gt, hipe_rtl:mk_imm(LastOffset),
		       hipe_rtl:label_name(Default), 
		       hipe_rtl:label_name(Lab3) , 0.5),
     Lab3,
     hipe_rtl:mk_load(Temp2,TablePntrReg,IndexReg),
     hipe_rtl:mk_branch(Temp2, eq, KeyReg,
			hipe_rtl:label_name(Lab4), 
			hipe_rtl:label_name(Default) , 0.9),
     Lab4,
     hipe_rtl:mk_alu(IndexReg, IndexReg, sra, 
		   hipe_rtl:mk_imm(2)),
     hipe_rtl:mk_sorted_switch(IndexReg, LabelList, KeyList)]
    .

step(?WORDSIZE,TablePntrReg,IndexReg,KeyReg) ->
  Temp = hipe_rtl:mk_new_reg(),
  TempIndex = hipe_rtl:mk_new_reg(),
  Lab1 = hipe_rtl:mk_new_label(),
  Lab2 = hipe_rtl:mk_new_label(),

  [hipe_rtl:mk_alu(TempIndex, IndexReg, add, 
		   hipe_rtl:mk_imm(?WORDSIZE)),

   hipe_rtl:mk_load(Temp,TablePntrReg,TempIndex),
   hipe_rtl:mk_branch(Temp, gt, KeyReg,
		      hipe_rtl:label_name(Lab2), 
		      hipe_rtl:label_name(Lab1) , 0.5),
   Lab1,
   hipe_rtl:mk_alu(IndexReg, IndexReg, add, 
		   hipe_rtl:mk_imm(?WORDSIZE)),
   hipe_rtl:mk_goto(hipe_rtl:label_name(Lab2)),
   Lab2
  ];
step(I,TablePntrReg,IndexReg,KeyReg) ->
  Temp = hipe_rtl:mk_new_reg(),
  TempIndex = hipe_rtl:mk_new_reg(),
  Lab1 = hipe_rtl:mk_new_label(),
  Lab2 = hipe_rtl:mk_new_label(),

  [hipe_rtl:mk_alu(TempIndex, IndexReg, add, hipe_rtl:mk_imm(I)),
   hipe_rtl:mk_load(Temp,TablePntrReg,TempIndex),
   hipe_rtl:mk_branch(Temp, gt, KeyReg,
		      hipe_rtl:label_name(Lab2), 
		      hipe_rtl:label_name(Lab1) , 0.5),
   Lab1,
   hipe_rtl:mk_move(IndexReg, TempIndex),
   hipe_rtl:mk_goto(hipe_rtl:label_name(Lab2)),
   Lab2 |
   step(I div 2,TablePntrReg,IndexReg,KeyReg)].





%% -----------------------------------------------------------------------
%%
%% dense_interval(Cases, Min, IcodeFail) -> {Max, Tbl, IsDense}
%%
%% Make sure that all integers in the interval [Min,Max]
%% has a label in the jumptable (inserting the fail-label
%% at positions not given in the list of pairs).
%%
%% As a side effect this function ensures that all 
%% constants in the table are integers, and throws the
%% exception not_an_integer othewise.
%%
%% The value of IsDense is true or false depending on 
%% whether the mapping is dense (Currently if at least 
%% half the entries in the range are present.)
%%
%% If the source would look like e.g.:
%%  case A of
%%    1 -> a;
%%    1000000000 -> b
%%  end,
%% then we don't want to create the table, 
%% it will be to sparce anyway, and the list would use
%% a lot of memory at compiletime, so we abort early.
%% This is don by ensuring that two values in the switch
%% are not further apart than MAXIMALGAP.
-define(MAXIMALGAP, 1000).

%% TODO: Extend this to handle clusters of dense intervals.
%% See:
%%  Producing Good Code for the Case Statement, Robert L. Bernstein
%%  Software - Practice and Experience vol 15, 1985, no 10, pp 1021--1024
%% And
%%  Correction to "Producing Good Code for the Case Statement"
%%  Sampath Kannan and Todd A. Proebsting,
%%  Software - Practice and Experience vol 24, 1994, no 2, p 233

dense_interval(Cases, Min, IcodeFail) ->
  dense_interval(Cases, Min, IcodeFail, 0, 0).
dense_interval([Pair = {Const,L}|Rest], Pos, Fail, Range, NoEntries) ->
  case hipe_icode:const_value(Const) of
    Val when integer(Val) ->
      if 
	Pos > Val -> 
	  %% Oooops Val seems to occure twice
	  %%  (Or we forgot to sort Cases...)
	  ?EXIT({multiple_occurences_of_val_in_switchval,Val});
	Pos < (Val - ?MAXIMALGAP) ->
	  %% We dont want to even try to create a table with 
	  %% gaps of size MAXIMALGAP.
	  throw(way_to_sparse);	  
	 Pos < Val ->
	  {Max, Res, IsDense} = 
	    dense_interval([Pair|Rest], Pos+1, Fail, Range+1, NoEntries),
	  {Max,[{hipe_icode:mk_const(Pos), Fail}|Res], IsDense};
	 true ->
	  {Max, Res, IsDense} = 
	    dense_interval(Rest, Pos+1, Fail,
			  Range+1, NoEntries+1),
	  {Max, [Pair | Res], IsDense}
      end;
    Other ->
      throw(not_an_integer)
  end;
dense_interval([], Max, _, Range, NoEntries) -> 
  {Max-1, [], NoEntries > (Range div 2)}.


%-------------------------------------------------------------------------
lbls_from_cases([{_,L}|Rest], VarMap) ->
  {Map,VarMap1} = lbls_from_cases(Rest, VarMap),
  {RtlL, VarMap2} = hipe_rtl_varmap:icode_label2rtl_label(L,VarMap1),
%  {[{label,hipe_rtl:label_name(RtlL)}|Map],VarMap2};
  {[hipe_rtl:label_name(RtlL)|Map],VarMap2};
lbls_from_cases([],VarMap) -> {[], VarMap}.

%-------------------------------------------------------------------------
split_cases(L) -> 
  split_cases(L,[],[]).
split_cases([],Vs,Ls) -> {lists:reverse(Vs),lists:reverse(Ls)};
split_cases([{V,L}|Rest], Vs, Ls) ->
  split_cases(Rest, [hipe_icode:const_value(V)|Vs], [L|Ls]).


%-------------------------------------------------------------------------
%
%
%-------------------------------------------------------------------------

%
% {switch_tuple_arity,X,Fail,N,[{A1,L1},...,{AN,LN}]}
%
% if not boxed(X) goto Fail
% Hdr := *boxed_val(X)
% switch_int(Hdr,Fail,[{H(A1),L1},...,{H(AN),LN}])
% where H(Ai) = make_arityval(Ai)
% 

gen_switch_tuple(I, Map, ConstTab, Options, ExitInfo) ->
    {X, Map1} = 
    hipe_rtl_varmap:icode_var2rtl_var(hipe_icode:switch_tuple_arity_arg(I), Map),
    Fail0 = hipe_icode:switch_tuple_arity_fail_label(I),
    {Fail1, Map2} = 
    hipe_rtl_varmap:icode_label2rtl_label(Fail0, Map1),
    FailLab = hipe_rtl:label_name(Fail1),
    {Cases, Map3} =
	lists:foldr(fun({A,L}, {Rest,M}) ->
			    {L1,M1} = hipe_rtl_varmap:icode_label2rtl_label(L, M),
			    L2 = hipe_rtl:label_name(L1),
			    A1 = hipe_icode:const_value(A),
			    H1 = hipe_tagscheme:mk_arityval(A1),
			    {[{H1,L2}|Rest], M1} end,
		    {[], Map2},
		    hipe_icode:switch_tuple_arity_cases(I)),
    Hdr = hipe_rtl:mk_new_reg(),
    IsBoxedLab = hipe_rtl:mk_new_label(),
    {[hipe_tagscheme:test_is_boxed(X, hipe_rtl:label_name(IsBoxedLab),
				   FailLab, 0.9),
      IsBoxedLab,
      hipe_tagscheme:get_header(Hdr, X) |
      gen_switch_int(Hdr, FailLab, Cases)],
     Map3, ConstTab}.

%
% rtl-level switch-on-int
% XXX: move to a module of its own and extend massively
%

gen_switch_int(X, FailLab, [{C,L}|Rest]) ->
    NextLab = hipe_rtl:mk_new_label(),
    [hipe_rtl:mk_branch(X, eq, hipe_rtl:mk_imm(C), L,
			hipe_rtl:label_name(NextLab), 0.5),
     NextLab |
     gen_switch_int(X, FailLab, Rest)];
gen_switch_int(_, FailLab, []) ->
    [hipe_rtl:mk_goto(FailLab)].
