%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%		  GRAPH COLORING REGISTER ALLOCATOR
%
% A simple graph coloring register allocator:
%
% - build interference graph + estimate spill costs
% - simplify graph (push on stack + spill)
% - select colors
%
% Emits a coloring: a list of {TempName,Location}
%  where Location is {reg,N} or {spill,M}.
%   and {reg,N} denotes some register N
%   and {spill,M} denotes the Mth spilled node
% You have to figure out how to rewrite the code yourself.
%
% This version uses vectors rather than hash tables, and uses
% faster algorithms since all vars are known at the start.
%   The result should be considerably quicker than earlier versions.
%
% be considerably faster as a result.
% Deficiencies:
% - no iterated graph coloring    (means more registers must be reserved)
%   * requires code rewriting to work
% - no rewriting of code          (spills, etc.)
% - no renaming                   (reduce unnecessary register pressure)
% - spill costs are naive         (should use better exec.estimates)
% - no biased coloring            (coalesce moves)
% - no live range splitting       (possibly not critical)
%
% *** NOTE ***
% You will have to rewrite some interface calls at the end of the file!

-module('hipe_optimistic_regalloc').
-export([regalloc/2]).

% Define these as 'ok' or 'report(X,Y)' depending on how much output you want.
%-define(report(X,Y),report(X,Y)).
-define(report(X,Y),ok).
-define(report2(X,Y),report(X,Y)).
-define(report3(X,Y),report(X,Y)).

% Given CFG and number of colors K, produce a coloring list
% of items {reg,N} (1 =< N =< K) and {spill,M}, where M is
% an index denoting 'a location'.
% (You might use it as a stack index, perhaps.)
%
% You can in principle delete check_coloring/2; it merely checks
% that the coloring agrees with the interference graph (that is, that
% no neighbors have the same register or spill location).

regalloc(CFG,K) ->
    ?report('building IG~n',[]),
    {IG,Spill,Phys} = build_ig(CFG),
    check_ig(IG),
    ?report('graph: ~p~nphysical regs: ~p~n',[list_ig(IG),Phys]),
    ?report('coloring IG with ~p colors~n',[K]),
    Cols = color(IG,Spill,Phys,K),
    check_coloring(Cols,IG),
    Cols.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** BUILDING THE INTERFERENCE GRAPH ***

% Returns {Interference_graph, Spill_cost_dictionary, Bag_of_physical_regs}

build_ig(CFG) ->
    case catch build_ig_0(CFG) of
	{'EXIT',Rsn} ->
	    error(build_ig,'failed with reason ~p~n',[Rsn]),
	    exit(Rsn);
	Else ->
	    Else
    end.

build_ig_0(CFG) ->
    Live = liveness(CFG),
    {_,NumN} = cfg:var_range(CFG),  % poss. N-1?
    {IG,Spill,Phys} = build_ig_bbs(labels(CFG),CFG,Live,
				   empty_ig(NumN),empty_spill(NumN),[]),
    {IG,Spill,ordsets:from_list(Phys)}.

build_ig_bbs([],CFG,Live,IG,Spill,Phys) -> {IG,Spill,Phys};
build_ig_bbs([L|Ls],CFG,Live,IG,Spill,Phys0) ->
    Xs = bb(CFG,L),
    {_,NewIG,NewSpill,Phys1} = build_ig_bb(Xs,liveout(Live,L),IG,Spill,Phys0),
    build_ig_bbs(Ls,CFG,Live,NewIG,NewSpill,Phys1).

build_ig_bb([],LiveOut,IG,Spill,Phys) -> {LiveOut,IG,Spill,Phys};
build_ig_bb([X|Xs],LiveOut,IG,Spill,Phys0) ->
    {Live,NewIG,NewSpill,Phys1} = build_ig_bb(Xs,LiveOut,IG,Spill,Phys0),
    build_ig_instr(X,Live,NewIG,NewSpill,Phys1).

% Note: We could add move-related arcs here as well.
%
% Note: Ideally, we would like to add all registers to the IG
%  at once rather than doing 'add_nodes' for each instruction.
% (This is costly, since nodes that already are present are checked!)

build_ig_instr(X,Live,IG,Spill,Phys) ->
    {D,U} = def_use(X),
    NewSpill = inc_spill_costs(D,inc_spill_costs(U,Spill)),
    NewIG = interference_arcs(D,Live,IG),
    NewLive = union(U,diff(Live,D)),
    NewPhys = collect_physical_regs(D,collect_physical_regs(U,Phys)),
    {NewLive,NewIG,NewSpill,NewPhys}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

collect_physical_regs([],Phys) -> Phys;
collect_physical_regs([X|Xs],Phys) ->
   case physical_reg(X) of
      true ->
	 collect_physical_regs(Xs,[X|Phys]);
      true ->
	 collect_physical_regs(Xs,Phys)
   end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

interference_arcs([],Live,IG) -> IG;
interference_arcs([X|Xs],Live,IG) ->
   interference_arcs(Xs,Live,i_arcs(X,Live,IG)).

i_arcs(X,[],IG) -> IG;
i_arcs(X,[Y|Ys],IG) ->
   i_arcs(X,Ys,add_edge(X,Y,IG)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

inc_spill_costs([],Spill) -> Spill;
inc_spill_costs([X|Xs],Spill) ->
   inc_spill_costs(Xs,inc_spill_cost(X,Spill)).

inc_spill_cost(X,Spill) ->
   set_spill_cost(X,get_spill_cost(X,Spill)+1,Spill).

get_spill_cost(X,Spill) ->
   spill_cost_lookup(X,Spill).

set_spill_cost(X,N,Spill) ->
   spill_cost_update(X,N,Spill).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** COLORING ***
%
% Coloring is done straightforwardly:
% - find the low-degree nodes, put them in low
% - while low non-empty:
%   * remove x from low
%   * push x on stack
%   * decrement degree of neighbors of x
%   * for each neighbor y of low degree, put y on low
% - when low empty:
%   - if graph empty, return stack
%   - otherwise
%     * select a node z to spill
%     * push z on stack
%     * decrement degree of neighbors of z
%     * add low-degree neighbors of z to low
%     * restart the while-loop above

color(IG,Spill,Phys,K) ->
    case catch color_0(IG,Spill,Phys,K) of
	{'EXIT',Rsn} ->
	    error(color,'failed with ~p~n',[Rsn]),
	    exit(Rsn);
	Else ->
	    Else
    end.

color_0(IG,Spill,Phys,K) ->
    ?report('simplification of IG~n',[]),
    Nodes = list_ig(IG),
    NumNodes = length(Nodes),
    Low = low_degree_nodes(Nodes,K),
    ?report(' starting with low degree nodes ~p~n',[Low]),
    NxtSpill = 0,
    EmptyStk = [],
    Stk = simplify(Low,NumNodes,Phys,IG,Spill,K,NxtSpill,EmptyStk),
    ?report('selecting colors~n',[]),
    select(Stk,Phys,IG,K).

%%%%%%%%%%%%%%%%%%%%
%
% Simplification: push all easily colored nodes on a stack;
%  when the list of easy nodes becomes empty, see if graph is
%  empty as well. If it is not, spill a node and continue.
%  If it is empty, return the stack.
%
% Notes:
% - We keep the set of visited nodes around for spill purposes
%   (visited nodes are not considered for spilling)
%
% - At present, nodes can be pushed onto the stack even if they
%   already are on the stack. This can be fixed by another 'Vis'
%   dictionary that keeps track of what is on the stack.
%   Currently, we just skip already colored nodes.
%
% - Arguments:
%   Low: low-degree nodes (ready to color)
%   NumNodes: number of remaining nodes in graph
%   IG: interference graph
%   Spill: spill costs of nodes
%   K: number of colors
%   Ix: next spill index
%   Stk: stack of already simplified nodes
%
% Physical registers are marked as 'visited' prior to simplify.
% This has the following effect:
% - they are not considered for spilling
% - they are not pushed on the stack
% - since we do NOT decrement degrees of surrounding vars, the
%   non-physreg variables must still take them into account.

simplify(Low,NumNodes,Phys,IG,Spill,K,Ix,Stk) ->
    ?report('physical regs: ~p~n',[Phys]),
    Vis = visit_all(Phys,none_visited(NumNodes)),
    ActualNumNodes = NumNodes-length(Phys),
    ?report('actual number of nodes: ~p~n',[ActualNumNodes]),
    simplify_ig(Low,ActualNumNodes,IG,Spill,K,Ix,Stk,Vis).

simplify_ig([],0,IG,Spill,K,Ix,Stk,Vis) -> 
    ?report('done~n',[]),
    Stk;
simplify_ig([],N,IG,Spill,K,Ix,Stk,Vis) when N > 0 ->
    ?report('  *** spill required ***~n',[]),
    {X,Low,NewIG} = spill(IG,Vis,Spill,K),
    NewVis = visit(X,Vis),
    {NewStk,NewIx} = push_spill_node(X,Ix,Stk),
    ?report('  node ~p spilled (~p now ready)~n',[X,Low]),
    simplify_ig(Low,N-1,NewIG,Spill,K,NewIx,NewStk,NewVis);
simplify_ig([X|Xs],N,IG,Spill,K,Ix,Stk,Vis) ->
    case visited(X,Vis) of
       true ->
	  ?report('  node ~p already visited~n',[X]),
	  simplify_ig(Xs,N,IG,Spill,K,Ix,Stk,Vis);
       true ->
	  {NewLow,NewIG} = decrement_neighbors(X,Xs,IG,Vis,K),
	  ?report('  node ~p pushed (~p now ready)~n',[X,NewLow]),
	  NewStk = push_colored(X,Stk),
	  simplify_ig( NewLow, N-1, NewIG, Spill, 
		      K, Ix, NewStk, visit(X,Vis) )
    end.

% Returns { NowLowDegreeNeighbors, NewIG }

decrement_neighbors(X,Xs,IG,Vis,K) ->
    Ns = unvisited_neighbors(X,Vis,IG),
    ?report('  node ~p has neighbors ~p (unvisited ~p)~n',
	   [X,neighbors(X,IG),Ns]),
    decrement_each(Ns,Xs,IG,Vis,K).

% For each node, decrement its degree and check if it is now
% a low-degree node. In that case, add it to the 'low list'.

decrement_each([],Low,IG,Vis,K) -> { Low, IG };
decrement_each([N|Ns],OldLow,IG,Vis,K) ->
    { Low, CurrIG } = decrement_each(Ns,OldLow,IG,Vis,K),
    case visited(N,Vis) of
	true ->
	    { Low, CurrIG };
	false ->
	    {D,NewIG} = decrement_degree(N,CurrIG),
	    if
		D < K ->
		    { [N|Low], NewIG };
		true ->
		    { Low, NewIG }
	    end
    end.

%%%%%%%%%%%%%%%%%%%%
%
% The spill cost of a node is:
%    est_spill_cost / current_degree
%
% For all unvisited nodes, compute spill cost and select the minimum.
% This node is chosen to be spilled. Then decrement the degree of its
% neighbors, and return those of low degree.
%
% Notes:
% - A better method for computing spill costs is to just keep the
%   minimum cost node. But for debugging purposes, we compute a list
%   of {node,spillcost} pairs and select the minimum.
%
% Returns:
%  { Spilled_node, Low_degree_neighbors, New_interference_graph }

spill(IG,Vis,Spill,K) ->
    Ns = list_ig(IG),
    Costs = spill_costs(Ns,IG,Vis,Spill),
    ?report('spill costs are ~p~n',[Costs]),
    ActualCosts = lists:sort(select_actual_costs(Costs)),
    ?report('actual costs are ~p~n',[ActualCosts]),
    case ActualCosts of
	[] ->
	    error(spill,'there is no node to spill');
	[{Cost,N}|Xs] ->
	    {Low,NewIG} = decrement_neighbors(N,[],IG,Vis,K),
	    ?report('spilled node ~p at cost ~p (~p now ready)~n',[N,Cost,Low]),
	    {N,Low,NewIG}
    end.

spill_costs(Ns,IG,Vis,Spill) -> 
    [ { spill_cost(N,Info,Vis,Spill), N } || {N,Info} <- Ns ].

spill_cost(N,Info,Vis,Spill) ->
    case visited(N,Vis) of 
       true ->
	  visited;
       true ->
	  spill_cost_of(N,Spill) / degree(Info)
    end.

select_actual_costs([]) -> [];
select_actual_costs([{visited,N}|Xs]) ->
    select_actual_costs(Xs);
select_actual_costs([{C,N}|Xs]) when number(C) ->
    [{C,N}|select_actual_costs(Xs)];
select_actual_costs([X|Xs]) ->
    warning(select_actual_costs,'undefined spill cost type ~p',[X]),
    select_actual_costs(Xs).

%%%%%%%%%%%%%%%%%%%%
%
% Note: we use pessimistic coloring here.
% - we could use optimistic coloring: for spilled node, check if there is
%   an unused color among the neighbors and choose that.

select(Stk,Phys,IG,K) ->
    NumNodes = length(Stk)+length(Phys),
    {PhysColors,Cols} = precolor(Phys,none_colored(NumNodes)),
    ?report('precoloring has yielded ~p~n',[list_coloring(Cols)]),
    PhysColors ++ select_colors(Stk,IG,Cols,K).

% We use optimistic coloring in this version:
% - When a nominally spilled variable is found, collect the colors
%   of neighbors. 
% - If there is a color not used by neighbors, choose that.
% - Otherwise spill node.

select_colors([],IG,Cols,K) -> 
    ?report('all nodes colored~n',[]),
    [];
select_colors([{X,colorable}|Xs],IG,Cols,K) ->
    ?report('color of ~p: ',[X]),
    {Reg,NewCols} = select_color(X,IG,Cols,K),
    ?report('~p~n',[Reg]),
    [{X,{reg,Reg}}|select_colors(Xs,IG,NewCols,K)];
select_colors([{X,{spill,M}}|Xs],IG,Cols,K) ->
    ?report('spilled: ~p~n',[X]),
    % Check if optimistic coloring could have found a color
    case catch select_color(X,IG,Cols,K) of
	{'EXIT',_} ->   % no color possible
	    ?report('(no optimistic color)~n',[]),
	    [{X,{spill,M}}|select_colors(Xs,IG,Cols,K)];
        {Reg,NewCols} ->
	    ?report('(optimistic color: ~p)~n',[Reg]),
	    [{X,{reg,Reg}}|select_colors(Xs,IG,NewCols,K)]
    end.

select_color(X,IG,Cols,K) ->
    Ns = neighbors(X,IG),
    ?report('neighbors of ~p are ~p~n',[X,Ns]),
    UsedColors = get_colors(neighbors(X,IG),Cols),
    Reg = select_unused_color(UsedColors,K),
    { Reg, set_color(X,Reg,Cols) }.

%%%%%%%%%%%%%%%%%%%%

get_colors([],Cols) -> [];
get_colors([X|Xs],Cols) ->
    case color_of(X,Cols) of
	uncolored ->
	    get_colors(Xs,Cols);
	{color,R} ->
	    [R|get_colors(Xs,Cols)]
    end.

select_unused_color(UsedColors,K) ->
    Summary = ordsets:from_list(UsedColors),
    ?report('neighbor colors: ~p~nused colors: ~p~n',[UsedColors,Summary]),
    select_unused(first_reg(),Summary,K).

% Assume first register number is "1".

first_reg() -> 1.

select_unused(R,[],K) when R =< K -> R;
select_unused(R,[X|Xs],K) when R < X -> R;
select_unused(R,[R|Xs],K) ->
    select_unused(R+1,Xs,K).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

push_spill_node(X,M,Stk) ->
    {  [{X,{spill,M}}|Stk], M+1 }.

push_colored(X,Stk) ->
    [ {X, colorable} | Stk ].

%%%%%%%%%%%%%%%%%%%%

low_degree_nodes([],K) -> [];
low_degree_nodes([{N,Info}|Xs],K) ->
    ?report('node ~p has degree ~p~n',[N,degree(Info)]),
    Deg = degree(Info),
    if
	Deg < K ->
	    [N|low_degree_nodes(Xs,K)];
	true ->
	    low_degree_nodes(Xs,K)
    end.

%%%%%%%%%%%%%%%%%%%%

unvisited_neighbors(X,Vis,IG) ->
    ordsets:from_list(unvisited(neighbors(X,IG),Vis)).

unvisited([],Vis) -> [];
unvisited([X|Xs],Vis) ->
    case visited(X,Vis) of
       true ->
	  unvisited(Xs,Vis);
       false ->
	  [X|unvisited(Xs,Vis)]
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** ABSTRACT DATATYPES ***

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Set operations used for liveness.

union([],Ys) -> Ys;
union([X|Xs],Ys) ->
    case member(X,Ys) of
       true ->
	  union(Xs,Ys);
       false ->
	  [X|union(Xs,Ys)]
    end.

diff([],Ys) -> [];
diff([X|Xs],Ys) ->
    case member(X,Ys) of
       true ->
	  diff(Xs,Ys);
       false ->
	  [X|diff(Xs,Ys)]
    end.

member(X,[]) -> false;
member(X,[X|_]) -> true;
member(X,[_|Ys]) -> member(X,Ys).
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The ig datatype:
%
% Note: if we know the number of temps used, we can use a VECTOR
% instead, which will speed up things.
%
% Note: later on, we may wish to add 'move-related' support.

-record(ig_info,{neighbors=[],degree=0}).

empty_ig(NumNodes) ->
    hipe_vectors:init(NumNodes,#ig_info{neighbors=[],degree=0}).

degree(Info) ->
    Info#ig_info.degree.

neighbors(Info) ->
    Info#ig_info.neighbors.

add_edge(X,X,IG) -> IG;
add_edge(X,Y,IG) ->
    add_arc(X,Y,add_arc(Y,X,IG)).

add_arc({var,X},{var,Y},IG) ->
   Info = hipe_vectors:get(IG,X),
   Old = neighbors(Info),
   Deg = degree(Info),
   case member(Y,Old) of
      true ->
	 New = Info;
      false ->
	 New = Info#ig_info{neighbors=[Y|Old],degree=Deg+1}
   end,
   hipe_vectors:set(IG,X,New).

degree(X,IG) when integer(X) ->
    Info = hipe_vectors:get(IG,X),
    Info#ig_info.degree;
degree({var,X},IG) when integer(X) ->
    Info = hipe_vectors:get(IG,X),
    Info#ig_info.degree.

neighbors(X,IG) when integer(X) ->
    Info = hipe_vectors:get(IG,X),
    Info#ig_info.neighbors;
neighbors({var,X},IG) when integer(X) ->
    Info = hipe_vectors:get(IG,X),
    Info#ig_info.neighbors.

% Returns {NewDegree,NewIG}

decrement_degree(X,IG) when integer(X) ->
    Info = hipe_vectors:get(IG,X),
    Degree = degree(Info),
    NewDegree = Degree-1,
    NewInfo = Info#ig_info{degree=NewDegree},
    {NewDegree,hipe_vectors:set(IG,X,NewInfo)};
decrement_degree({var,X},IG) when integer(X) ->
    Info = hipe_vectors:get(IG,X),
    Degree = degree(Info),
    NewDegree = Degree-1,
    NewInfo = Info#ig_info{degree=NewDegree},
    {NewDegree,hipe_vectors:set(IG,X,NewInfo)}.

list_ig(IG) ->
    hipe_vectors:list(IG).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The spill cost datatype:
%
% Note: if we know the number of temps used, we can use a VECTOR
% instead, which will speed up things.

empty_spill(NumNodes) ->
    hipe_vectors:init(NumNodes,0).

spill_cost_of(X,Spill) when integer(X) ->
    hipe_vectors:get(Spill,X);
spill_cost_of({var,X},Spill) when integer(X) ->
    hipe_vectors:get(Spill,X).

spill_cost_lookup({var,X},Spill) when integer(X) ->
    spill_cost_of(X,Spill);
spill_cost_lookup(X,Spill) when integer(X) ->
    spill_cost_of(X,Spill).

spill_cost_update(X,N,Spill) when integer(X) ->
    hipe_vectors:set(Spill,X,N);
spill_cost_update({var,X},N,Spill) when integer(X) ->
    hipe_vectors:set(Spill,X,N).

list_spill_costs(Spill) -> 
    hipe_vectors:list(Spill).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The coloring datatype:

none_colored(NumNodes) ->
    hipe_vectors:init(NumNodes,uncolored).

color_of(X,Cols) when integer(X) ->
    hipe_vectors:get(Cols,X);
color_of({var,X},Cols) when integer(X) ->
    hipe_vectors:get(Cols,X).

set_color(X,R,Cols) when integer(X) ->
    hipe_vectors:set(Cols,X,{color,R});
set_color({var,X},R,Cols) when integer(X) ->
    hipe_vectors:set(Cols,X,{color,R}).

list_coloring(Cols) ->
    hipe_vectors:list(Cols).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Note: there might be a slight gain in separate the two versions
% of visit/2 and visited/2. (So that {var,X} selects X and calls
% the integer version.

none_visited(NumNodes) ->
    hipe_vectors:init(NumNodes).

visit(X,Vis) when integer(X) ->
    hipe_vectors:set(Vis,X,visited);
visit({var,X},Vis) when integer(X) ->
    hipe_vectors:set(Vis,X,visited).

visited(X,Vis) when integer(X) ->
    vis0(X,Vis);
visited({var,X},Vis) when integer(X) ->
    vis0(X,Vis).

vis0(X,Vis) ->
    case hipe_vectors:get(Vis,X) of
	visited ->
	    true;
	_ ->
	    false
    end.

visit_all([],Vis) -> Vis;
visit_all([X|Xs],Vis) ->
    visit_all(Xs,visit(X,Vis)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** TESTING ***

% regalloc a procedure

regalloc_proc(MFA,K) ->
    case catch cfg:init(translate:translate(MFA)) of
	{'EXIT',Rsn} ->
	    warning(regalloc_proc,'translation to cfg failed with ~p~n',[Rsn]),
	    exit(Rsn);
	CFG ->
	    regalloc(CFG,K)
    end.

regalloc_all([],K) -> ?report3('done~n',[]);
regalloc_all([MFA|MFAs],K) ->
    ?report3('regalloc ~p~n',[MFA]),
    case catch regalloc_proc(MFA,K) of
	{'EXIT',Rsn} ->
	    ?report3(' failed with ~p~n',[Rsn]);
	_ ->
	    ?report3(' successful~n',[])
    end,
    regalloc_all(MFAs,K).

%%%%%%%%%%%%%%%%%%%%
% Check that all arcs in IG are bidirectional + degree is correct

check_ig(IG) ->
    check_ig(list_ig(IG),IG).

check_ig([],IG) -> ok;
check_ig([{N,Info}|Xs],IG) ->
    Ns = neighbors(Info),
    NumNs = length(Ns),
    D = degree(Info),
    if
	D == NumNs ->
	    ok;
	true ->
	    warning('node ~p has degree ~p but ~p neighbors~n',[N,D,NumNs])
    end,
    check_neighbors(N,Ns,IG),
    check_ig(Xs,IG).

check_neighbors(N,[],IG) -> ok;
check_neighbors(N,[M|Ms],IG) ->
    Ns = neighbors(M,IG),
    case member(N,Ns) of
       true ->
	  ok;
       false ->
	  warning('node ~p should have ~p as neighbor (has ~p)~n',[M,N,Ns])
    end,
   check_neighbors(N,Ms,IG).

%%%%%%%%%%%%%%%%%%%%
% Check that the coloring is correct (if the IG is correct):
%
% NOTE: does NOTNOTNOT use same representation of coloring as
%  the standard allocator!

check_coloring(Coloring,IG) ->
    ?report('checking coloring ~p~n',[Coloring]),
    check_cols(list_ig(IG),init_coloring(Coloring)).

init_coloring(Xs) -> hash:init(Xs).

check_color_of(X,Cols) ->
%    if
%	physical_reg(X) ->
%	    phys_reg_color(X,Cols);
%	true ->
	    case hash:lookup(X,Cols) of
		not_found ->
		    warning(check_color_of,'node ~p: color not found~n',[X]),
		    uncolored;
		{found,C} ->
		    C
	    end.
%    end.

check_cols([],Cols) ->
    ?report('coloring valid~n',[]),
    ok;
check_cols([{X,Info}|Xs],Cols) ->
    Cs = [ {N,check_color_of(N,Cols)} || N <- neighbors(Info) ],
    C = check_color_of(X,Cols),
    case valid_coloring(X,C,Cs) of
	yes ->
	    check_cols(Xs,Cols);
	{no,Invalids} ->
	    warning(check_cols,'node ~p has same color (~p) as ~p~n',
		    [X,C,Invalids]),
	    check_cols(Xs,Cols)
    end.

valid_coloring(X,C,[]) ->
    yes;
valid_coloring(X,C,[{Y,C}|Ys]) ->
    case valid_coloring(X,C,Ys) of
	yes -> {no,[Y]};
	{no,Zs} -> {no,[Y|Zs]}
    end;
valid_coloring(X,C,[_|Ys]) ->
    valid_coloring(X,C,Ys).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

vcg_ig(IG) ->
    Ns = list_ig(IG),
    io:format('~ngraph: {~n',[]),
    vcg_show_nodes(Ns),
    vcg_show_arcs(Ns),
    io:format('~n}~n',[]).

vcg_show_nodes([]) -> ok;
vcg_show_nodes([{N,Info}|Xs]) ->
    io:format('node: {title:\"~w\" label:\"~w ~w\" shape: ~w}~n',
	      [N,N,degree(Info),vcg_shape_node(N)]),
    vcg_show_nodes(Xs).

vcg_shape_node(N) ->
    case physical_reg(N) of 
       true ->
	  rhomb;
       false ->
	  ellipse
    end.

vcg_show_arcs([]) -> ok;
vcg_show_arcs([{N,Info}|Xs]) ->
    Ns = neighbors(Info),
    vcg_node_arcs(N,Ns),
    vcg_show_arcs(Xs).

vcg_node_arcs(N,[]) -> ok;
vcg_node_arcs(N,[M|Ms]) ->
    io:format('edge: { sourcename: \"~w\" targetname: \"~w\" }~n',[N,M]),
    vcg_node_arcs(N,Ms).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** ERROR HANDLING ***

warning(Fun,Msg) ->
    io:format('WARNING (~p): ~p~n',[Fun,Msg]).

warning(Fun,Msg,Xs) ->
    io:format('WARNING (~p):',[Fun]),
    io:format(Msg,Xs),
    io:format('~n',[]).

error(Fun,Msg) ->
    io:format('ERROR (~p): ~p~n',[Fun,Msg]),
    exit({error,Fun,Msg}).

error(Fun,Msg,Xs) ->
    io:format('ERROR (~p):n',[Fun]),
    io:format(Msg,Xs),
    io:format('~n',[]),
    exit({error,Fun,Msg}).

report(Msg) ->
    report(Msg,[]).

% XXX
%report(Msg,Args) -> ok;
report(Msg,Args) ->
    io:format(Msg,Args).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% *** INTERFACES TO OTHER MODULES ***

liveness(CFG) ->
    liveness:analyze(CFG).

liveout(CFG,L) ->
    liveness:liveout(CFG,L).

labels(CFG) ->
    cfg:labels(CFG).

bb(CFG,L) ->
    {bb,Xs} = cfg:bb(CFG,L),
    Xs.

def_use(X) ->
    {hipe_icode:defines(X),hipe_icode:uses(X)}.

% physical_reg(X):
%  returns true if X is a precolored/physical/actual register
%  and false otherwise.
%
% (The current version is a mock-up that just classifies every
% k'th register name as 'physical')

physical_reg({var,N}) when integer(N) ->
    PhysRegMod = phys_reg_modulus(),
    if
	N rem PhysRegMod == 0 ->
	    true;
	true ->
	    false
    end;
physical_reg(N) when integer(N) ->
    PhysRegMod = phys_reg_modulus(),
    if
	N rem PhysRegMod == 0 ->
	    true;
	true ->
	    false
    end;    
physical_reg(X) -> false.

% Every 4th register name is considered a physical name.
% NOTE: this can lead to uncolorable graphs! (We have no guarantees
% that every 4th reg has a 'well-behaved' IG neighborhood.)

phys_reg_modulus() -> 8.

%%%%%%%%%%%%%%%%%%%%
%
% Precoloring: use this version when a proper implementation of
%  physical_name(X) is available!
%
% precolor(Xs,Cols) -> precolor(Xs,[],Cols).
%
% precolor([],Cs,Cols) -> {Cs,Cols};
%precolor([X|Xs],Cs,Cols) ->
%    {C,NewCols} = precolor_reg(X,Cols1),
%    precolor(Xs,[{X,C}|Cs],Cols).
%
%precolor_reg(X,Cols) ->
%    {{X,physical_name(X)},set_color(X,physical_name(X),Cols)}.

precolor(Xs,Cols) ->
    ?report('precoloring ~p~n',[Xs]),
    {Cs,NewCol} = precolor(Xs,1,Cols),
    ?report('   yielded ~p~n',[Cs]),
    {Cs,NewCol}.

precolor([],N,Cols) -> {[],Cols};
precolor([X|Xs],N,Cols) ->
    R = reg_name(X),
    {Cs,Cols1} = precolor(Xs,N+1,Cols),
    {[{R,{reg,N}}|Cs],set_color(X,N,Cols1)}.

reg_name(X) when integer(X) -> X;
reg_name({var,X}) when integer(X) -> X.

%    precolor(Xs,N+1,set_color(X,N,Cols)).
