%%% Coalescing register allocator for x86
%%%
%%% Rules:
%%% - Before RA, every Temp (precoloured or pseudo) is semantically
%%%   equivalent to Reg. Any operand may be Temp.
%%% - Before RA, only FIXED registers may occur in precoloured Temps.
%%%   The exception is move: src or dst may be %eax.
%%% - After RA, an operand (src or dst) may refer to at most one memory cell.
%%%   Therefore, a pseudo-Temp MAY NOT occur as base or offset in an
%%%   explicit memory operand after RA.
%%% - After RA, a binary operation (alu, cmp, move) may refer to at most
%%%   one memory cell. Therefore, AT MOST ONE of src and dst may be a
%%%   pseudo-Temp after RA.
%%% - After RA, the return values in call and return must be the
%%%   %eax and %edx registers.
%%% - After RA, the index in a jmp_switch must be a register.

-module(hipe_x86_ra_coalescing).
-export([ra/1]).
-include("hipe_x86.hrl").
-include("../main/hipe.hrl").

-define(countspills,true).

-ifdef(countspills).
-define(count_spills(X), X).
-define(update_spillsum,
	case get(totalspill) of 
	  {__Loads,__Stores} ->
	    put(totalspill,{__Loads+get(loads),
			    __Stores+get(stores)}),
	    case get(spilledtemps) of
	      __Temps -> 
		put(spilledtemps, __Temps+get(temps));
	      _ ->
		true
	    end;
	  _ -> true 
	end).
-else.
-define(count_spills(X), true).
-define(update_spillsum,true).
-endif.

ra(X86Defun) ->
    io:format("Starting coalescing regalloc!\n"),
    ?count_spills({put(loads,0),put(stores,0),put(temps,0)}),
    #defun{code=Code0,label_range=Labels0} = X86Defun,

    CFG1 = hipe_x86_cfg:init(X86Defun),
    hipe_gensym:set_label(x86, hipe_gensym:get_label(rtl)),
    {_, SpillLimit} = hipe_x86_cfg:var_range(CFG1),
    NewCFG = alloc(CFG1, unspilled, 0, SpillLimit, 0),
    Code1 = hipe_x86_cfg:linearise( NewCFG ),
    {LabelsLo,_} = Labels0,
    LabelsHi = hipe_gensym:get_label(x86),
    Labels1 = {LabelsLo,LabelsHi},
    io:format("Finished coalescing regalloc\n"),
    X86Defun#defun{code=Code1,label_range=Labels1}.


alloc( X86Cfg, PrevSpillArea, SpillIndex, SpillLimit, Locs) ->
    {Coloring, NewSpillIndex} = 
	hipe_coalescing_regalloc:regalloc(X86Cfg, SpillIndex, SpillLimit, hipe_x86_specific ),

     ?count_spills(put(temps, NewSpillIndex)),
    case spilled(Coloring) of
	false ->
	    ?update_spillsum,
	    ColTuple = cols2tuple(Coloring),
	    Labels = hipe_x86_cfg:labels(X86Cfg),
	    CFG0 = rewrite_bbs(Labels, X86Cfg, ColTuple),
	    NewCFG = hipe_x86_cfg:var_range_update(CFG0, {0, 7}), %FIXME
	    case SpillIndex of
		0 ->
		    NewCFG;
		_ ->
		    %% io:format("Spilled: ~w register(s) @ ~w sites (~w)~n", 
		    %%	 [SpillIndex, Locs, hipe_x86_cfg:function(X86Cfg)]),
		    NewConstTab =
			hipe_consttab:update_block(hipe_x86_cfg:data(NewCFG),
						   PrevSpillArea,
						   4,
						   word,
						   hipe_consttab:repeat(SpillIndex, 0)),
		    hipe_x86_cfg:update_data(NewCFG, NewConstTab)
	    end;
	true ->
	    Labels = hipe_x86_cfg:labels(X86Cfg),
	    {Low, High} = hipe_x86_cfg:var_range(X86Cfg),
	    hipe_gensym:set_var(High),
	    {CFG1, SpillArea} = 
		case PrevSpillArea of
		    unspilled ->
			{NewCTab, SpillA} = 
			    hipe_consttab:insert_block(hipe_x86_cfg:data(X86Cfg),
						       4, word, [0]),
			{hipe_x86_cfg:update_data(X86Cfg, NewCTab), SpillA};
		    SpillA ->
			{X86Cfg, SpillA}
		end,
	    
	    Spills = spill_regs(Coloring),
	    {CFG2, Locs0} = spill_rewrite_bbs(Labels, CFG1, Spills, 
					      SpillArea, Locs),
	    CFG3 = hipe_x86_cfg:var_range_update(CFG2,
						   {Low, hipe_gensym:get_var()}),
	    alloc(CFG3, SpillArea, NewSpillIndex, SpillLimit,Locs0)
    end.

%
% Check if a coloring spilled any values
%

spilled([]) ->
   false;
spilled([{_, {spill, _}} | Cs]) ->
   true;
spilled([_ | Cs]) ->
   spilled(Cs).

%
% Returns a list of {SpilledReg, {Tmp, LoadInstr, StoreInstr}}
%

spill_regs([]) ->
   [];
spill_regs([{RegNr, {reg, _}} | Colors]) ->
   spill_regs(Colors);
spill_regs([{RegNr, {spill, SpillIndex}} | Colors]) ->
   SpillIndexImm = hipe_x86:mk_imm(SpillIndex*4),
   Info = {hipe_x86:mk_reg(RegNr), SpillIndexImm},
   [Info | spill_regs(Colors)].


%
% Rewrite a cfg where spills occured
%

spill_rewrite_bbs([], CFG, Spills, SpillArea, Locs) ->
   {CFG, Locs};
spill_rewrite_bbs([Lbl|Lbls], CFG, Spills, SpillArea, Locs) ->
   BB = hipe_x86_cfg:bb(CFG, Lbl),
   Code = hipe_bb:code(BB),
   {NewCode, Locs0} = spill_rewrite_instrs(Code, Spills, SpillArea),
   NewCFG = hipe_x86_cfg:bb_update(CFG, Lbl, hipe_bb:code_update(BB, NewCode)),
   spill_rewrite_bbs(Lbls, NewCFG, Spills, SpillArea, Locs+Locs0).

%
% Convert a list of [{R1, C1}, {R2, C2}, ...} to a tuple {C17, C23, ...}.
%
% The N's must be unique but do not have to be sorted and they can be sparse.
%

cols2tuple(Map) ->
   cols2tuple(1, lists:keysort(1, Map), []).

cols2tuple(N, [], Vs) ->
   list_to_tuple(lists:reverse(Vs));
cols2tuple(N, [{R, C}|Ms], Vs) when N =:= R ->
   cols2tuple(N+1, Ms, [C|Vs]);
cols2tuple(N, Ms, Vs) ->
   cols2tuple(N+1, Ms, [unknown|Vs]).


%
% Rewrite a cfg to use the allocated registers
%

rewrite_bbs([], CFG, ColTuple) ->
   CFG;
rewrite_bbs([Lbl|Lbls], CFG, ColTuple) ->
   BB = hipe_x86_cfg:bb(CFG, Lbl),
   Code = hipe_bb:code(BB),
   NewCode = rewrite_instrs(Code, ColTuple),
   NewCFG = hipe_x86_cfg:bb_update(CFG, Lbl, hipe_bb:code_update(BB, NewCode)),
   rewrite_bbs(Lbls, NewCFG, ColTuple).

get_spills([], _, _) ->
   [];
get_spills([R|Rs], [], Spills) ->
   get_spills(Rs, Spills, Spills);
get_spills([R|Rs], [{R, Info}|_], Spills) ->
   [{R, Info} | get_spills(Rs, Spills, Spills)];
get_spills(Rs, [_|Ss], Spills) ->
   get_spills(Rs, Ss, Spills).

spill_rewrite_instrs([], Spills, SpillArea) ->
   {[], 0};
spill_rewrite_instrs([I|Is], Spills, SpillArea) ->
   {NewI, Locs0} = spill_rewrite_instr(I, Spills, SpillArea),
   {NewIs, Locs1} = spill_rewrite_instrs(Is, Spills, SpillArea),
   {NewI++NewIs, Locs0+Locs1}.


spill_rewrite_instr(I, Spills, SpillArea) ->
    Defs = hipe_x86_defuse:insn_def(I),
    Uses = hipe_x86_defuse:insn_use(I),
    SpillUses = get_spills(Uses, Spills, Spills),
    SpillDefs = get_spills(Defs, Spills, Spills),
    case SpillUses++SpillDefs of
	[] ->
	    {[I], 0};
	_ ->
	    SpillAreaReg = hipe_x86:mk_new_temp( untagged ),
	    %LA = hipe_x86:load_address_create(SpillAreaReg,
		%			   SpillArea, constant, []),
	    %LA = hipe_x86:mk_lea( mk_mem(
	    {Loads,UseSubst} = 
		case SpillUses of
		    [] ->
			{[],[]};
		    [{R1u, Offset1u}] ->
			Tmp1u = hipe_x86:mk_new_temp( untagged ),
			{%[hipe_x86:load_create(Tmp1u, uw,
				%	    SpillAreaReg, Offset1u, [])],
			 [hipe_x86:mk_move(hipe_x86:mk_mem(SpillAreaReg, Offset1u, untagged), Tmp1u)],
			 [{R1u, Tmp1u}]};
		    [{R1u, Offset1u}, {R2u, Offset2u}] ->
			Tmp1u = hipe_x86:mk_new_temp(untagged),
			Tmp2u = hipe_x86:mk_new_temp(untagged),
			{%[hipe_x86:load_create(Tmp1u, uw,
				%	    SpillAreaReg, Offset1u, []),
			 [hipe_x86:mk_move(hipe_x86:mk_mem(SpillAreaReg, Offset1u, untagged), Tmp1u)],
			 % hipe_x86:load_create(Tmp2u, uw,
				%	    SpillAreaReg, Offset2u, [])],
                         [hipe_x86:mk_move(hipe_x86:mk_mem(SpillAreaReg, Offset2u, untagged), Tmp2u)],
			 [{R1u, Tmp1u}, {R2u, Tmp2u}]}
		end,
	    {Stores,DefSubst} = 
		case SpillDefs of
		    [] ->
			{[],[]};
		    [{R1d, Offset1d}] ->
			Tmp1d = hipe_x86:mk_new_temp(untagged),
			{%[hipe_x86:store_create(SpillAreaReg,
				%	     Offset1d, uw, Tmp1d, [])],
			 [hipe_x86:mk_move(Tmp1d, hipe_x86:mk_mem(SpillAreaReg, Offset1d, untagged))],
			 [{R1d, Tmp1d}]}
		end,
       ?count_spills(
	  {put(stores,get(stores)+length(Stores)),
	   put(loads,get(loads)+length(Loads))}),
	 C1 = hipe_x86:mk_comment('** SPILL START **'),
	 C2 = hipe_x86:mk_comment('** SPILL END **'),
	 NewI = hipe_x86:subst_defines(hipe_x86:subst_uses(I, UseSubst), DefSubst),
	 {[C1] ++ Loads ++ [NewI] ++ Stores ++ [C2], length(UseSubst)+length(DefSubst)} %[C1,LA]
   end.

rewrite_instrs([], ColTuple) ->
   [];
rewrite_instrs([I|Is], ColTuple) ->
   [rewrite_instr(I, ColTuple) | rewrite_instrs(Is, ColTuple)].

% This function may have to check:
% * both src,dst in binary op is not memory
rewrite_instr(Ins, ColTuple) ->
    case Ins of
	#alu{aluop=Aluop,src=Src,dst=Dst} ->
	    NewSrc = color_arg(hipe_x86:alu_src(Ins), ColTuple),
	    NewDst = color_arg(hipe_x86:alu_dst(Ins), ColTuple),
	    Ins#alu{aluop=Aluop,src=NewSrc,dst=NewDst};
	#cmovcc{cond=Cond,src=Src,dst=Dst} ->
	    NewSrc = color_arg(hipe_x86:cmovcc_src(Ins), ColTuple),
	    NewDst = color_arg(hipe_x86:cmovcc_dst(Ins), ColTuple),
	    Ins#cmovcc{cond=Cond,src=NewSrc,dst=NewDst};
	#cmp{src=Src,dst=Dst} ->
	    NewSrc = color_arg(hipe_x86:cmp_src(Ins), ColTuple),
	    NewDst = color_arg(hipe_x86:cmp_dst(Ins), ColTuple),
	    Ins#cmp{src=NewSrc,dst=NewDst};
	#dec{dst=Dst} ->
	    NewDst = color_arg(hipe_x86:dec_dst(Ins), ColTuple),
	    Ins#dec{dst=NewDst};
	#inc{dst=Dst} ->
	    NewDst = color_arg(hipe_x86:inc_dst(Ins), ColTuple),
	    Ins#inc{dst=NewDst};
	#jmp_switch{temp=Temp,jtab=JTab,labels=Labels} ->
	    NewTemp = color_arg(hipe_x86:jmp_switch_temp(Ins), ColTuple),
	    Ins#jmp_switch{temp=Temp,jtab=JTab,labels=Labels};
	#lea{mem=Mem,temp=Temp} ->
	    NewMem  = color_arg(hipe_x86:lea_mem(Ins), ColTuple),
	    NewTemp = color_arg(hipe_x86:lea_temp(Ins), ColTuple),
	    Ins#lea{mem=NewMem,temp=NewTemp};
	#move{src=Src,dst=Dst} ->
	    NewSrc = color_arg(hipe_x86:move_src(Ins), ColTuple),
	    NewDst = color_arg(hipe_x86:move_dst(Ins), ColTuple),
	    Ins#move{src=NewSrc,dst=NewDst};
	#push{src=Src} ->
	    NewSrc = color_arg(hipe_x86:push_src(Ins), ColTuple),
	    Ins#push{src=NewSrc};
	_ -> Ins
   end.


% should check if a register is involved i.e both temp and mem operands FIXME
color_arg(Arg, ColTuple) ->
    io:format("color_arg - Arg:~w ColTuple:~w\n",[Arg,ColTuple]),
    case hipe_x86:is_temp(Arg) of
	true ->
	    case hipe_x86:temp_reg(Arg) < 8 of
		true ->
		    case element(hipe_x86:temp_reg(Arg), ColTuple) of %reg_nr
			{reg, NewRgNr} ->
			    hipe_x86:mk_temp(NewRgNr,untagged);
			{spill, SpillIndex} ->
			    exit({x86, spilled})
		    end;
		false ->
		    Arg
	    end;
	false ->
	    Arg
    end.
