%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%%----------------------------------------------------------------------
%% Purpose: Transform internal form of Megaco/H.248 messages
%%----------------------------------------------------------------------

-module(megaco_transformer).

-include_lib("megaco/include/megaco.hrl").
-include_lib("megaco/include/megaco_message_v1.hrl").

-export([
	 to_state/2,
	 transform_MegacoMessage/2
        ]).

-record(state, {mode             = verify, % verify | encode | decode
                resolver_module  = megaco_binary_name_resolver,
                resolver_options = [8, 8, 8]}).

to_state(Mode, Config) ->
    case Config of
	[] ->
	    #state{mode = Mode};
	undefined ->
	    #state{mode = Mode};
	native ->
	    #state{mode = Mode};
	[{Mod, Opt}] ->
	    #state{mode = Mode, resolver_module = Mod, resolver_options = Opt}
    end.

resolve(Type, Item, State) ->
    case State#state.mode of
        verify ->
            Item;
        encode ->
            Mod = State#state.resolver_module,
            Opt = State#state.resolver_options,
            Mod:encode_name(Opt, Type, Item);
        decode ->
            Mod = State#state.resolver_module,
            Opt = State#state.resolver_options,
            Mod:decode_name(Opt, Type, Item)
    end.

transform_MegacoMessage(#'MegacoMessage'{authHeader = Auth,
                                         mess       = Mess},
                        State) ->
    #'MegacoMessage'{authHeader = transform_opt_AuthenticationHeader(Auth, State),
                     mess       = transform_Message(Mess, State)}.

transform_opt_AuthenticationHeader(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_AuthenticationHeader(#'AuthenticationHeader'{secParmIndex = SPI,
                                                           seqNum       = SN,
                                                           ad           = AuthData},
                                   State) ->
    #'AuthenticationHeader'{secParmIndex = transform_SecurityParmIndex(SPI, State),
                            seqNum       = transform_SequenceNum(SN, State),
                            ad           = transform_AuthData(AuthData, State)}.

transform_SecurityParmIndex(SPI, State) ->
    transform_HEXDIG(SPI, State, 8, 8). % BUGBUG: Mismatch between ASN.1 and ABNF

transform_SequenceNum(SN, State) ->
    transform_HEXDIG(SN, State, 8, 8).  % BUGBUG: Mismatch between ASN.1 and ABNF

transform_AuthData(AuthData, State) ->
    transform_HEXDIG(AuthData, State, 32, 64).  % BUGBUG: Mismatch between ASN.1 and ABNF

transform_Message(#'Message'{version     = Version,
                             mId         = MID,
                             messageBody = Body},
                  State) ->
    #'Message'{version     = transform_version(Version, State),
               mId         = transform_MId(MID, State),
               messageBody = transform_Message_messageBody(Body, State)}.

transform_version(Version, State) ->
    transform_DIGIT(Version, State, 0, 99).

transform_Message_messageBody({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            messageError -> transform_ErrorDescriptor(Val, State);
            transactions when list(Val) -> [transform_Transaction(T, State) || T <- Val]
        end,
    {Tag, Val2}.

transform_MId({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            ip4Address -> transform_IP4Address(Val, State);
            ip6Address -> transform_IP6Address(Val, State);
            domainName -> transform_DomainName(Val, State);
            deviceName -> transform_PathName(Val, State);
            mtpAddress -> transform_mtpAddress(Val, State)
        end,
    {Tag, Val2}.

transform_mtpAddress(MtpAddr, State) ->
    transform_OCTET_STRING(MtpAddr, State, 5, 5). % BUBGBUG: length?

transform_DomainName(#'DomainName'{name       = Name,
                                   portNumber = Port},
                     State) ->
    %% BUGBUG: (ALPHA / DIGIT) *63(ALPHA / DIGIT / "-" / ".")
    Domain = #'DomainName'{name       = transform_STRING(Name, State, 1, 64),
			   portNumber = transform_opt_portNumber(Port, State)},
    {domainName, Domain2} = resolve(mid, {domainName, Domain}, State),
    Domain2.

transform_IP4Address(#'IP4Address'{address    = [A1, A2, A3, A4],
                                   portNumber = Port},
                     State) ->
    #'IP4Address'{address    = [transform_V4hex(A1, State),
                                transform_V4hex(A2, State),
                                transform_V4hex(A3, State),
                                transform_V4hex(A4, State)],
                  portNumber = transform_opt_portNumber(Port, State)}.

transform_V4hex(Val, State) ->
    transform_DIGIT(Val, State, 0, 255).

transform_IP6Address(Val, State) ->
    exit(ipv6_not_supported). %% BUGBUG: nyi

transform_PathName(Path, State) ->
    %% BUGBUG: ["*"] NAME *("/" / "*"/ ALPHA / DIGIT /"_" / "$" ) 
    %% BUGBUG: ["@" pathDomainName ]
    Device = transform_STRING(Path, State, 1, 64),
    {deviceName, Device2} = resolve(mid, {deviceName, Device}, State),
    Device2.

transform_Transaction({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            transactionRequest ->     transform_TransactionRequest(Val, State);
            transactionPending ->     transform_TransactionPending(Val, State);
            transactionReply ->       transform_TransactionReply(Val, State);
            transactionResponseAck -> transform_TransactionResponseAck(Val, State)
        end,
    {Tag, Val2}.

transform_TransactionResponseAck(#'TransactionResponseAck'{firstAck = First,
                                                           lastAck  = Last},
                                 State) ->
    #'TransactionResponseAck'{firstAck = transform_TransactionId(First, State),
                              lastAck  = transform_opt_TransactionId(Last, State)}.

transform_opt_TransactionId(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_TransactionId(Id, State) ->
    transform_TransactionId(Id, State).

transform_TransactionId(Id, State) ->
    transform_UINT32(Id, State).

transform_TransactionRequest(#'TransactionRequest'{transactionId = Id,
                                                   actions       = Actions},
                             State) when list(Actions) ->

    #'TransactionRequest'{transactionId = transform_TransactionId(Id, State),
                          actions       = [transform_ActionRequest(ActReq, State) || ActReq <- Actions]}.

transform_TransactionPending(#'TransactionPending'{transactionId = Id},
                             State) ->
    #'TransactionPending'{transactionId = transform_TransactionId(Id, State)}.

transform_TransactionReply(#'TransactionReply'{transactionId     = Id,
                                               immAckRequired    = ImmAck,
                                               transactionResult = TransRes},
                           State) ->
    #'TransactionReply'{transactionId     = transform_TransactionId(Id, State),
                        immAckRequired    = transform_opt_null(ImmAck, State),
                        transactionResult = transform_TransactionReply_transactionResult(TransRes, State)}.

transform_opt_null(asn1_NOVALUE, State) -> asn1_NOVALUE;
transform_opt_null('NULL', State)       -> 'NULL'.

transform_TransactionReply_transactionResult({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            transactionError ->
                transform_ErrorDescriptor(Val, State);
            actionReplies when list(Val), Val /= [] ->
                [transform_ActionReply(ActRep, State) || ActRep <- Val]
        end,
    {Tag, Val2}.

transform_opt_ErrorDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_ErrorDescriptor(ErrDesc, State) ->
    transform_ErrorDescriptor(ErrDesc, State).

transform_ErrorDescriptor(#'ErrorDescriptor'{errorCode = Code,
                                             errorText = Text},
                          State) ->
    #'ErrorDescriptor'{errorCode = transform_ErrorCode(Code, State),
                       errorText = transform_opt_ErrorText(Text, State)}.

transform_ErrorCode(Code, State) ->
    transform_DIGIT(Code, State, 0, 999).

transform_opt_ErrorText(asn1_NOVALUE, State)  ->
    asn1_NOVALUE;
transform_opt_ErrorText(Text, State)  ->
    transform_QUOTED_STRING(Text, State).

transform_ContextID(CtxId, State) ->
    case CtxId of
        ?megaco_all_context_id    -> ?megaco_all_context_id;
        ?megaco_null_context_id   -> ?megaco_null_context_id;
        ?megaco_choose_context_id -> ?megaco_choose_context_id;
        Int when integer(Int)     -> transform_UINT32(Int, State)
    end.

transform_ActionRequest(#'ActionRequest'{contextId           = CtxId,
                                         contextRequest      = CtxReq,
                                         contextAttrAuditReq = CtxAuditReq,
                                         commandRequests     = CmdReqList},
                        State) ->
    #'ActionRequest'{contextId           = transform_ContextID(CtxId, State),
                     contextRequest      = transform_opt_ContextRequest(CtxReq, State),
                     contextAttrAuditReq = transform_opt_ContextAttrAuditRequest(CtxAuditReq, State),
                     commandRequests     = [transform_CommandRequest(CmdReq, State) || CmdReq <- CmdReqList]}.

transform_ActionReply(#'ActionReply'{contextId       = CtxId,
                                     errorDescriptor = ErrDesc,
                                     contextReply    = CtxRep,
                                     commandReply    = CmdRepList},
                      State) ->
    CmdRepList2 = [transform_CommandReply(CmdRep, State) || CmdRep <- CmdRepList],
    #'ActionReply'{contextId       = transform_ContextID(CtxId, State),
                   errorDescriptor = transform_opt_ErrorDescriptor(ErrDesc, State),
                   contextReply    = transform_opt_ContextRequest(CtxRep, State),
                   commandReply    = CmdRepList2}.

transform_opt_ContextRequest(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_ContextRequest(#'ContextRequest'{priority    = Prio,
                                               emergency   = Em,
                                               topologyReq = TopReqList},
                             State) ->
    Prio2 = 
        case Prio of
            asn1_NOVALUE -> asn1_NOVALUE;
            _            -> transform_integer(Prio, State, 0, 15)
        end,
    Em2 = 
        case Em of
            asn1_NOVALUE -> asn1_NOVALUE;
            false        -> false;
            true         -> true
        end,
    TopReqList2 = 
        case TopReqList of
            asn1_NOVALUE -> asn1_NOVALUE;
            _            -> [transform_TopologyRequest(TopReq, State) ||
                                TopReq <- TopReqList]
        end,
    #'ContextRequest'{priority    = Prio2,
                      emergency   = Em2,
                      topologyReq = TopReqList2}.

transform_opt_ContextAttrAuditRequest(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_ContextAttrAuditRequest(#'ContextAttrAuditRequest'{topology = Top,
                                                                 emergency = Em,
                                                                 priority = Prio},
                                      State) ->
    #'ContextAttrAuditRequest'{topology  = transform_opt_null(Top, State),
                               emergency = transform_opt_null(Em, State),
                               priority  = transform_opt_null(Prio, State)}.

transform_CommandRequest(#'CommandRequest'{command        = Cmd,
                                           optional       = Opt,
                                           wildcardReturn = Wild},
                         State) ->
    #'CommandRequest'{optional       = transform_opt_null(Opt, State),
                      wildcardReturn = transform_opt_null(Wild, State),
                      command        = transform_Command(Cmd, State)}.

transform_Command({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            addReq ->            transform_AmmRequest(Val, State);
            moveReq ->           transform_AmmRequest(Val, State);
            modReq ->            transform_AmmRequest(Val, State);
            subtractReq ->       transform_SubtractRequest(Val, State);
            auditCapRequest ->   transform_AuditRequest(Val, State);
            auditValueRequest -> transform_AuditRequest(Val, State);
            notifyReq ->         transform_NotifyRequest(Val, State);
            serviceChangeReq ->  transform_ServiceChangeRequest(Val, State)
        end,
    {Tag, Val2}.

transform_CommandReply({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            addReply ->           transform_AmmsReply(Val, State);
            moveReply ->          transform_AmmsReply(Val, State);
            modReply ->           transform_AmmsReply(Val, State);
            subtractReply ->      transform_AmmsReply(Val, State);
            auditCapReply ->      transform_AuditReply(Val, State);
            auditValueReply ->    transform_AuditReply(Val, State);
            notifyReply ->        transform_NotifyReply(Val, State);
            serviceChangeReply -> transform_ServiceChangeReply(Val, State)
        end,
    {Tag, Val2}.

transform_TopologyRequest(#'TopologyRequest'{terminationFrom   = From,
                                             terminationTo     = To,
                                             topologyDirection = Dir},
                          State) ->
    Dir2 = 
        case Dir of
            bothway -> bothway;
            isolate -> isolate;
            oneway ->  oneway
        end,
    #'TopologyRequest'{terminationFrom   = transform_TerminationID(From, State),
                       terminationTo     = transform_TerminationID(To, State),
                       topologyDirection = Dir2}.

transform_AmmRequest(#'AmmRequest'{terminationID = IdList,
                                   descriptors   = DescList},
                     State) ->
    #'AmmRequest'{terminationID = [transform_TerminationID(Id, State) ||
                                      Id <- IdList],
                  descriptors   = [transform_ammDescriptor(Desc, State) ||
                                      Desc <- DescList]}.

transform_ammDescriptor({Tag, Desc}, State) ->
    Desc2 = 
        case Tag of
            mediaDescriptor       -> transform_MediaDescriptor(Desc, State);
            modemDescriptor       -> transform_ModemDescriptor(Desc, State);      
            muxDescriptor         -> transform_MuxDescriptor(Desc, State);   
            eventsDescriptor      -> transform_EventsDescriptor(Desc, State);      
            eventBufferDescriptor -> transform_EventBufferDescriptor(Desc, State); 
            signalsDescriptor     -> transform_SignalsDescriptor(Desc, State);    
            digitMapDescriptor    -> transform_DigitMapDescriptor(Desc, State);    
            auditDescriptor       -> transform_AuditDescriptor(Desc, State)
        end,
    {Tag, Desc2}.

transform_AmmsReply(#'AmmsReply'{terminationID    = IdList,
                                 terminationAudit = TermAudit},
                    State) ->
    TermAudit2 =
        case TermAudit of
            asn1_NOVALUE -> asn1_NOVALUE;
            _            -> transform_TerminationAudit(TermAudit, State)
        end,
    #'AmmsReply'{terminationID    = [transform_TerminationID(Id, State) ||
                                        Id <- IdList],
                 terminationAudit = TermAudit2}.

transform_SubtractRequest(#'SubtractRequest'{terminationID   = IdList,
                                             auditDescriptor = Desc},
                          State) ->
    Desc2 =
        case Desc of
            asn1_NOVALUE -> asn1_NOVALUE;
            _            -> transform_AuditDescriptor(Desc, State)
        end,
    #'SubtractRequest'{terminationID   = [transform_TerminationID(Id, State) ||
                                             Id <- IdList],
                       auditDescriptor = Desc2}.

transform_AuditRequest(#'AuditRequest'{terminationID = IdList,
                                       auditDescriptor = Desc},
                       State) ->
    Desc2 =
        case Desc of
            asn1_NOVALUE -> asn1_NOVALUE;
            _            -> transform_AuditDescriptor(Desc, State)
        end,
    #'AuditRequest'{terminationID = [transform_TerminationID(Id, State) ||
                                        Id <- IdList],
                    auditDescriptor = Desc2}.

%% auditReply           = (AuditValueToken / AuditCapToken ) 
%%                        ( contextTerminationAudit  / auditOther)
%% auditOther           = EQUAL TerminationID LBRKT 
%%                        terminationAudit RBRKT
%% terminationAudit     = auditReturnParameter *(COMMA auditReturnParameter) 
%% 
%% contextTerminationAudit = EQUAL CtxToken ( terminationIDList / 
%%                        LBRKT errorDescriptor RBRKT )
transform_AuditReply(#'AuditReply'{terminationID = Id,
                                   auditResult   = AuditRes},
                     State) ->
    #'AuditReply'{terminationID = transform_TerminationID(Id, State),
                  auditResult   = transform_auditResult(AuditRes, State)}.

transform_auditResult({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            contextAuditResult ->
                %% contextTerminationAudit
                %% BUGBUG: Mismatch between ASN.1 and ABNF
                %% BUGBUG: LBRKT errorDescriptor RBRKT
                [transform_TerminationID(Id, State) || Id <- Val];
            terminationAuditResult ->
                %% auditOther
                transform_TerminationAudit(Val, State)
        end,
    {Tag, Val2}.

transform_AuditDescriptor(#'AuditDescriptor'{auditToken = Tokens},
                          State) ->
    case Tokens of
        asn1_NOVALUE -> asn1_NOVALUE;
        _            -> [transform_auditItem(Token, State) || Token <- Tokens]
    end.

transform_auditItem(Token, State) ->
    case Token of
        muxToken               -> muxToken;
        modemToken             -> modemToken;
        mediaToken             -> mediaToken;
        eventsToken            -> eventsToken;
        signalsToken           -> signalsToken;
        digitMapToken          -> digitMapToken;
        statsToken             -> statsToken;
        observedEventsToken    -> observedEventsToken;
        packagesToken          -> packagesToken;
        eventBufferToken       -> eventBufferToken
    end.

transform_TerminationAudit(ParmList, State) when list(ParmList) ->
    [transform_AuditReturnParameter(Parm, State) || Parm <- ParmList].

transform_AuditReturnParameter({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            errorDescriptor ->
                transform_ErrorDescriptor(Val, State);
            mediaDescriptor ->
                transform_MediaDescriptor(Val, State);
            modemDescriptor ->
                transform_ModemDescriptor(Val, State);
            muxDescriptor ->
                transform_MuxDescriptor(Val, State);
            eventsDescriptor ->
                transform_EventsDescriptor(Val, State);
            eventBufferDescriptor ->
                transform_EventBufferDescriptor(Val, State);
            signalsDescriptor ->
                transform_SignalsDescriptor(Val, State);
            digitMapDescriptor ->
                transform_DigitMapDescriptor(Val, State);
            observedEventsDescriptor ->
                transform_ObservedEventsDescriptor(Val, State);
            statisticsDescriptor ->
                transform_StatisticsDescriptor(Val, State);
            packagesDescriptor ->
                transform_PackagesDescriptor(Val, State);
            emptyDescriptors ->
                transform_EmptyDescriptors(Val, State)
        end,
    {Tag, Val2}.

transform_EmptyDescriptors(#'AuditDescriptor'{auditToken = Tokens},
                           State) ->
    case Tokens of
        asn1_NOVALUE -> asn1_NOVALUE;
        _            -> [transform_auditItem(Token, State) || Token <- Tokens]
    end.

transform_NotifyRequest(#'NotifyRequest'{terminationID            = IdList,
                                         observedEventsDescriptor = ObsDesc,
                                         errorDescriptor          = ErrDesc},
                        State) ->
    %% BUGBUG: Mismatch between ASN.1 and ABNF
    %% BUGBUG: The following ought to be a 'choice'
    #'NotifyRequest'{terminationID            = [transform_TerminationID(Id, State) ||
                                                    Id <- IdList],
                     observedEventsDescriptor = transform_ObservedEventsDescriptor(ObsDesc, State),
                     errorDescriptor          = transform_opt_ErrorDescriptor(ErrDesc, State)}.

transform_NotifyReply(#'NotifyReply'{terminationID = IdList,
                                     errorDescriptor = ErrDesc},
                      State) ->
    #'NotifyReply'{terminationID = transform_opt_TerminationIDList(IdList, State),
                   errorDescriptor = transform_opt_ErrorDescriptor(ErrDesc, State)}.

transform_opt_TerminationIDList(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_TerminationIDList(IdList, State) when list(IdList) ->
    [transform_TerminationID(Id, State) || Id <- IdList].

transform_ObservedEventsDescriptor(#'ObservedEventsDescriptor'{requestId        = Id,
                                                               observedEventLst = Events},
                                   State) when list (Events) ->
    #'ObservedEventsDescriptor'{requestId        = transform_RequestID(Id, State),
                                observedEventLst = [transform_ObservedEvent(E, State) || E <- Events]}.

%% ;time per event, because it might be buffered
%% observedEvent        = [ TimeStamp LWSP COLON] LWSP 
%%                        pkgdName [ LBRKT observedEventParameter
%%                        *(COMMA observedEventParameter) RBRKT ]
%% 
%% ;at-most-once eventStream, every eventParameterName at most once
%% observedEventParameter = eventStream / eventOther

transform_ObservedEvent(#'ObservedEvent'{eventName    = Name,
                                         streamID     = Id,
                                         eventParList = Parms,
                                         timeNotation = Time},
                        State) ->
    #'ObservedEvent'{eventName    = transform_EventName(Name, State),
                     streamID     = transform_opt_StreamID(Id, State),
                     eventParList = transform_EventParameter(Parms, State),
                     timeNotation = transform_opt_TimeNotation(Time, State)}.

transform_EventName(Name, State) ->
    Name2 = transform_PkgdName(Name, State),
    resolve(event, Name2, State).

transform_EventParameter(#'EventParameter'{eventParameterName = Name,
                                           value              = Value},
                         State) ->
    %% BUGBUG: event parameter name
    Name2 = transform_Name(Name, State),
    #'EventParameter'{eventParameterName = resolve(event, Name2, State),
                      value              = transform_Value(Value, State)}.

transform_ServiceChangeRequest(#'ServiceChangeRequest'{terminationID      = IdList,
                                                       serviceChangeParms = Parms},
                               State) ->
    #'ServiceChangeRequest'{terminationID      = [transform_TerminationID(Id, State) || Id <- IdList],
                            serviceChangeParms = transform_ServiceChangeParm(Parms, State)}.

%% serviceChangeReply   = ServiceChangeToken EQUAL TerminationID
%%                        [LBRKT (errorDescriptor / 
%%                        serviceChangeReplyDescriptor) RBRKT]
%% serviceChangeReplyDescriptor = ServicesToken LBRKT
%%                        servChgReplyParm *(COMMA servChgReplyParm) RBRKT
%% 
%% ;at-most-once. Version is REQUIRED on first ServiceChange response
%% servChgReplyParm     = (serviceChangeAddress / serviceChangeMgcId /
%%                        serviceChangeProfile / serviceChangeVersion )
transform_ServiceChangeReply(#'ServiceChangeReply'{terminationID       = IdList,
                                                   serviceChangeResult = Res},
                             State) ->
    #'ServiceChangeReply'{terminationID       = [transform_TerminationID(Id, State) || Id <- IdList],
                          serviceChangeResult = transform_ServiceChangeResult(Res, State)}.

transform_ServiceChangeResult({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            errorDescriptor       -> transform_ErrorDescriptor(Val, State);
            serviceChangeResParms -> transform_ServiceChangeResParm(Val, State)
        end,
    {Tag, Val2}.

%% TerminationID        = "ROOT" / pathNAME / "$" / "*"
%% ; Total length of pathNAME must not exceed 64 chars.
%% pathNAME             = ["*"] NAME *("/" / "*"/ ALPHA / DIGIT /"_" / "$" ) 
%%                        ["@" pathDomainName ]

transform_TerminationID(TermId, State) when State#state.mode /= verify ->
    resolve(term_id, TermId, State);
transform_TerminationID(#'TerminationID'{wildcard = Wild,
                                         id       = Id},
                        State) ->
    #'TerminationID'{wildcard = Wild,
                     id       = Id};
transform_TerminationID(#megaco_term_id{contains_wildcards = IsWild,
                                        id                 = Id},
                        State) ->
    #megaco_term_id{contains_wildcards = transform_bool(IsWild, State),
                    id                 = [transform_term_id_component(Sub, State) || Sub <- Id]}.

transform_defaulted_bool(undefined, State)  -> false;
transform_defaulted_bool(Bool, State)       -> transform_bool(Bool, State).

transform_bool(true, State)  -> true;
transform_bool(false, State) -> false.

transform_term_id_component(Sub, State) ->
    case Sub of
        all    -> all;
        choose -> choose;
        Char when integer(Char) -> Char
    end.

%% mediaDescriptor      = MediaToken LBRKT mediaParm *(COMMA mediaParm) RBRKT
%% ; at-most-once per item
%% ; and either streamParm or streamDescriptor but not both
%% mediaParm            = (streamParm / streamDescriptor / 
%%                         terminationStateDescriptor)
%% ; at-most-once
%% streamParm           = ( localDescriptor / remoteDescriptor / 
%%                         localControlDescriptor )
%% streamDescriptor     = StreamToken EQUAL StreamID LBRKT streamParm 
%%                        *(COMMA streamParm) RBRKT
transform_MediaDescriptor(#'MediaDescriptor'{termStateDescr = TermState,
                                             streams        = Streams},
                          State) ->
    #'MediaDescriptor'{termStateDescr = transform_opt_TerminationStateDescriptor(TermState, State),
                       streams        = transform_streams(Streams, State)}.

transform_streams({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            oneStream   -> transform_StreamParms(Val, State);
            multiStream -> transform_StreamDescriptor(Val, State)
        end,
    {Tag, Val2}.

transform_StreamParms(#'StreamParms'{localControlDescriptor = Control,
                                     localDescriptor        = Local,
                                     remoteDescriptor       = Remote},
                      State) ->
    #'StreamParms'{localControlDescriptor = transform_opt_LocalControlDescriptor(Control, State),
                   localDescriptor        = transform_opt_LocalRemoteDescriptor(Local, State),
                   remoteDescriptor       = transform_opt_LocalRemoteDescriptor(Remote, State)}.

transform_StreamDescriptor(#'StreamDescriptor'{streamID    = Id,
                                               streamParms = Parms},
                           State) ->
    #'StreamDescriptor'{streamID    = transform_StreamID(Id, State),
                        streamParms = transform_StreamParms(Parms, State)}.

%% localControlDescriptor = LocalControlToken LBRKT localParm 
%%                          *(COMMA localParm) RBRKT
%% 
%% ; at-most-once per item
%% localParm            = ( streamMode / propertyParm /
%%                          reservedValueMode  / reservedGroupMode ) 
%% reservedValueMode       = ReservedValueToken EQUAL ( "ON" / "OFF" ) 
%% reservedGroupMode       = ReservedGroupToken EQUAL ( "ON" / "OFF" ) 
%% 
%% reservedMode      = ReservedToken EQUAL ( "ON" / "OFF" )
%% 
%% streamMode           = ModeToken EQUAL streamModes
transform_opt_LocalControlDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_LocalControlDescriptor(#'LocalControlDescriptor'{streamMode    = Mode,
                                                               reserveGroup  = Group,
                                                               reserveValue  = Value,
                                                               propertyParms = Props},
                                     State) ->
    #'LocalControlDescriptor'{streamMode    = transform_opt_StreamMode(Mode, State),
                              reserveGroup  = transform_defaulted_bool(Group, State),
                              reserveValue  = transform_defaulted_bool(Value, State),
                              propertyParms = [transform_PropertyParm(P, State) || P <- Props]}.

transform_opt_StreamMode(Mode, State) ->
    case Mode of
        asn1_NOVALUE -> asn1_NOVALUE;
        sendOnly     -> sendOnly;
        recvOnly     -> recvOnly;
        sendRecv     -> sendRecv;
        inactive     -> inactive;
        loopBack     -> loopBack
    end.

transform_Name(Name, State) ->
    %% BUGBUG: transform
    %% BUGBUG: NAME = ALPHA *63(ALPHA / DIGIT / "_" )
    transform_STRING(Name, State, 1, 64).

transform_PkgdName(Name, State) ->
    %% BUGBUG: transform
    %% BUGBUG:  pkgdName =  (NAME / "*")  SLASH  (ItemID / "*" )
    transform_OCTET_STRING(Name, State, 1, 64).

%% When text encoding the protocol, the descriptors consist of session
%% descriptions as defined in SDP (RFC2327), except that the "s=", "t="
%% and "o=" lines are optional. When multiple session descriptions are
%% provided in one descriptor, the "v=" lines are required as delimiters;
%% otherwise they are optional.  Implementations shall accept session
%% descriptions that are fully conformant to RFC2327. When binary
%% encoding the protocol the descriptor consists of groups of properties
%% (tag-value pairs) as specified in Annex C.  Each such group may
%% contain the parameters of a session description.
transform_opt_LocalRemoteDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_LocalRemoteDescriptor(#'LocalRemoteDescriptor'{propGrps = Groups},
                                    State) ->
    #'LocalRemoteDescriptor'{propGrps = [transform_PropertyGroup(G, State) || G <- Groups]}.

transform_PropertyGroup(Props, State) ->
    [transform_PropertyGroupParm(P, State) || P <- Props].

transform_PropertyGroupParm(#'PropertyParm'{name  = Name,
                                            value = Value},
                            State) ->
    #'PropertyParm'{name  = transform_PkgdName(Name, State),
                    value = transform_OCTET_STRING(Value, State, 0, infinity)}.

%% propertyParm         = pkgdName parmValue
%% parmValue            = (EQUAL alternativeValue/ INEQUAL VALUE)
%% alternativeValue     = ( VALUE / LSBRKT VALUE *(COMMA VALUE) RSBRKT  / 
%%                        LSBRKT VALUE DOT DOT VALUE RSBRKT )
transform_PropertyParm(#'PropertyParm'{name      = Name,
                                       value     = Values,
                                       extraInfo = Extra},
                       State) ->
    Name2 = transform_PkgdName(Name, State),
    #'PropertyParm'{name      = resolve(property, Name2, State),
                    value     = [transform_Value(V, State) || V <- Values], % BUGBUG: ???
                    extraInfo = transform_opt_extraInfo(Extra, State)}.

transform_opt_extraInfo(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_extraInfo({relation, Rel}, State) ->
    Rel2 = 
        case Rel of
            greaterThan -> greaterThan;
            smallerThan -> smallerThan;
            unequalTo   -> unequalTo
        end,
    {relation, Rel2};
transform_opt_extraInfo({range, Range}, State) ->
    Range2 = transform_bool(Range, State),
    {range, Range2};
transform_opt_extraInfo({sublist, Sub}, State) ->
    Sub2 = transform_bool(Sub, State),
    {sublist, Sub2}.

transform_opt_TerminationStateDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_TerminationStateDescriptor(#'TerminationStateDescriptor'{propertyParms      = Props,
                                                                       eventBufferControl = Control,
                                                                       serviceState       = Service},
                                         State) ->
    #'TerminationStateDescriptor'{propertyParms      = [transform_PropertyParm(P, State) || P <- Props],
                                  eventBufferControl = transform_opt_EventBufferControl(Control, State),
                                  serviceState       = transform_opt_ServiceState(Service, State)}.

transform_opt_EventBufferControl(Control, State) ->
    case Control of
        asn1_NOVALUE -> asn1_NOVALUE;
        off          -> off;
        lockStep     -> lockStep
    end.

transform_opt_ServiceState(Service, State) ->
    case Service of
        asn1_NOVALUE -> asn1_NOVALUE;
        test         -> test;
        outOfSvc     -> outOfSvc;
        inSvc        -> inSvc
    end.

transform_MuxDescriptor(#'MuxDescriptor'{muxType  = Type,
                                         termList = IdList},
                        State) ->
    #'MuxDescriptor'{muxType  = transform_MuxType(Type, State),
                     termList = [transform_TerminationID(Id, State) || Id <- IdList]}.

transform_MuxType(Type, State) ->
    case Type of
        h221 -> h221;
        h223 -> h223;
        h226 -> h226;
        v76  -> v76
    end.

transform_opt_StreamID(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_StreamID(Id, State) ->
    transform_StreamID(Id, State).

transform_StreamID(Id, State) ->
    transform_UINT16(Id, State).

transform_EventsDescriptor(#'EventsDescriptor'{requestID = Id,
                                               eventList = Events},
                           State) ->
    #'EventsDescriptor'{requestID = transform_RequestID(Id, State),
                        eventList = [transform_RequestedEvent(E, State) || E <- Events]}.

transform_RequestedEvent(#'RequestedEvent'{pkgdName    = Name,
                                           streamID    = Id,
                                           evParList   = Parms,
                                           eventAction = Actions},
                         State)  ->
    Name2 = transform_PkgdName(Name, State),
    #'RequestedEvent'{pkgdName    = resolve(event, Name2, State),
                      streamID    = transform_opt_StreamID(Id, State),
                      eventAction = transform_opt_RequestedActions(Actions, State),
                      evParList   = [transform_EventParameter(P, State) || P <- Parms]}.

transform_opt_RequestedActions(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_RequestedActions(#'RequestedActions'{keepActive        = Keep,
                                                   eventDM           = DM,
                                                   secondEvent       = Event,
                                                   signalsDescriptor = SigDesc},
                               State) ->
    #'RequestedActions'{keepActive        = transform_keepActive(Keep, State),
                        eventDM           = transform_opt_EventDM(DM, State),
                        secondEvent       = transform_SecondEventsDescriptor(Event, State),
                        signalsDescriptor = transform_opt_SignalsDescriptor(SigDesc, State)}.

transform_opt_keepActive(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_keepActive(Keep, State) ->
    transform_keepActive(Keep, State).

transform_keepActive(Keep, State) ->
    transform_bool(Keep, State).

transform_opt_EventDM(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_EventDM({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            digitMapName  -> transform_DigitMapName(Val, State);
            digitMapValue -> transform_DigitMapValue(Val, State)
        end,
    {Tag, Val2}.

transform_SecondEventsDescriptor(#'SecondEventsDescriptor'{requestID = Id,
                                                           eventList = Events},
                                 State) ->
    #'SecondEventsDescriptor'{requestID = transform_RequestID(Id, State),
                              eventList = [transform_SecondRequestedEvent(E, State) || E <- Events]}.

transform_SecondRequestedEvent(#'SecondRequestedEvent'{pkgdName    = Name,
                                                       streamID    = Id,
                                                       evParList   = Parms,
                                                       eventAction = Actions},
                               State) ->
    Name2 = transform_PkgdName(Name, State),
    #'SecondRequestedEvent'{pkgdName    = resolve(event, Name2, State),
                            streamID    = transform_opt_StreamID(Id, State),
                            eventAction = transform_opt_SecondRequestedActions(Actions, State),
                            evParList   = [transform_EventParameter(P, State) || P <- Parms]}.


transform_opt_SecondRequestedActions(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_SecondRequestedActions(#'SecondRequestedActions'{keepActive        = Keep,
                                                               eventDM           = DM,
                                                               signalsDescriptor = SigDesc},
                                     State) ->
    #'SecondRequestedActions'{keepActive        = transform_keepActive(Keep, State),
                              eventDM           = transform_opt_EventDM(DM, State),
                              signalsDescriptor = transform_opt_SignalsDescriptor(SigDesc, State)}.

transform_EventBufferDescriptor(EventSpecs, State) ->
    [transform_EventSpec(ES, State) || ES <- EventSpecs].

transform_EventSpec(#'EventSpec'{eventName    = Name,
                                 streamID     = Id,
                                 eventParList = Parms},
                    State) ->
    #'EventSpec'{eventName    = transform_EventName(Name, State),
                 streamID     = transform_opt_StreamID(Id, State),
                 eventParList = transform_EventParameter(Parms, State)}.

transform_opt_SignalsDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_SignalsDescriptor(SigDesc, State) ->
    transform_SignalsDescriptor(SigDesc, State).

transform_SignalsDescriptor(SigDesc, State)  when list(SigDesc) ->
    [transform_SignalRequest(SigReq, State) || SigReq <- SigDesc].

transform_SignalRequest({Tag, Val}, State) ->
    Val2 =
        case Tag of
            signal     -> transform_Signal(Val, State);
            seqSigList -> transform_SeqSigList(Val, State)
        end,
    {Tag, Val2}.


transform_SeqSigList(#'SeqSigList'{id         = Id,
                                   signalList = SigList},
                     State) when list(SigList) ->
    #'SeqSigList'{id         = transform_UINT16(Id, State),
                  signalList = [transform_Signal(Sig, State) || Sig <- SigList]}.

transform_Signal(#'Signal'{signalName       = Name,
                           streamID         = Id,
                           sigType          = Type,
                           duration         = Dur,
                           notifyCompletion = Compl,
                           keepActive       = Keep,
                           sigParList       = Parms},
                 State) ->
    #'Signal'{signalName       = transform_SignalName(Name, State),
              streamID         = transform_opt_StreamID(Id, State),
              sigType          = transform_opt_SignalType(Type, State),
              duration         = transform_opt_duration(Dur, State),
              notifyCompletion = transform_opt_NotifyCompletion(Compl, State),
              keepActive       = transform_opt_keepActive(Keep, State),
              sigParList       = [transform_SigParameter(P, State) || P <- Parms]}.

transform_opt_duration(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_duration(Dur, State) ->
    transform_UINT16(Dur, State).

transform_opt_NotifyCompletion(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_NotifyCompletion(Items, State) when list(Items) ->
    [transform_notifyCompletionItem(I, State) || I <- Items].

transform_notifyCompletionItem(Item, State) ->
    case Item of
        onTimeOut                   -> onTimeOut;
        onInterruptByEvent          -> onInterruptByEvent;
        onInterruptByNewSignalDescr -> onInterruptByNewSignalDescr;
        otherReason                 -> otherReason
    end.

transform_opt_SignalType(Type, State) ->
    case Type of
        asn1_NOVALUE -> asn1_NOVALUE;
        brief        ->   brief;
        onOff        ->   onOff;
        timeOut      -> timeOut
    end.

transform_SignalName(Name, State) ->
    Name2 = resolve(signal, Name, State),
    transform_PkgdName(Name2, State).

transform_SigParameter(#'SigParameter'{sigParameterName = Name,
                                       value            = Value},
                       State) ->
    Name2 = transform_Name(Name, State),
    %% BUGBUG: transform
    #'SigParameter'{sigParameterName = resolve(signal_name, Name2, State),
                    value            = transform_Value(Value, State)}.

transform_RequestID(Id, State) ->
    transform_UINT32(Id, State).

transform_ModemDescriptor(#'ModemDescriptor'{mtl = Types,
                                             mpl = Props},
                          State) when list(Types), list(Props) -> 
    %% BUGBUG: Does not handle extensionParameter
    #'ModemDescriptor'{mtl = [transform_ModemType(T, State) || T <- Types],
                       mpl = [transform_PropertyParm(P, State) || P <- Props]}.

transform_ModemType(Type, State) ->
    %% BUGBUG: Does not handle extensionParameter
    case Type of
        v18       -> v18;
        v22       -> v22;
        v22bis    -> v22bis;
        v32       -> v32;
        v32bis    -> v32bis;
        v34       -> v34;
        v90       -> v90;
        v91       -> v91;
        synchISDN -> synchISDN
    end.

transform_DigitMapDescriptor(#'DigitMapDescriptor'{digitMapName  = Name,
                                                   digitMapValue = Value},
                             State) ->
    #'DigitMapDescriptor'{digitMapName  = transform_opt_DigitMapName(Name, State),
                          digitMapValue = transform_opt_DigitMapValue(Value, State)}.

transform_opt_DigitMapName(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_DigitMapName(Name, State) ->
    transform_DigitMapName(Name, State).

transform_DigitMapName(Name, State) ->
    Name2 = transform_Name(Name, State),
    resolve(dialplan, Name2, State).

transform_opt_DigitMapValue(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_DigitMapValue(Value, State) ->
    transform_DigitMapValue(Value, State).

transform_DigitMapValue(#'DigitMapValue'{digitMapBody = Body,
                                         startTimer   = Start,
                                         shortTimer   = Short,
                                         longTimer    = Long},
                        State) ->
    #'DigitMapValue'{startTimer   = transform_opt_timer(Start, State),
                     shortTimer   = transform_opt_timer(Short, State),
                     longTimer    = transform_opt_timer(Long, State),
                     digitMapBody = transform_STRING(Body, State)}. %% BUGBUG: digitMapBody not handled at all

transform_opt_timer(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_timer(Timer, State) ->
    transform_DIGIT(Timer, State, 0, 99).

transform_ServiceChangeParm(#'ServiceChangeParm'{serviceChangeMethod  = Method, 
                                                 serviceChangeAddress = Addr, 
                                                 serviceChangeVersion = Version, 
                                                 serviceChangeProfile = Profile, 
                                                 serviceChangeReason  = Reason, 
                                                 serviceChangeDelay   = Delay, 
                                                 serviceChangeMgcId   = MgcId, 
                                                 timeStamp            = Time},
                            State) ->
    #'ServiceChangeParm'{serviceChangeMethod  = transform_ServiceChangeMethod(Method, State),
                         serviceChangeAddress = transform_opt_ServiceChangeAddress(Addr, State),
                         serviceChangeVersion = transform_opt_serviceChangeVersion(Version, State),
                         serviceChangeProfile = transform_opt_ServiceChangeProfile(Profile, State),
                         serviceChangeReason  = transform_serviceChangeReason(Reason, State),
                         serviceChangeDelay   = transform_opt_serviceChangeDelay(Delay, State),
                         serviceChangeMgcId   = transform_opt_serviceChangeMgcId(MgcId, State),
                         timeStamp            = transform_opt_TimeNotation(Time, State)}.

transform_ServiceChangeMethod(Method, State) ->
    case Method of
        failover      -> failover;
        forced        -> forced;
        graceful      -> graceful;
        restart       -> restart;
        disconnected  -> disconnected;
        handOff       -> handOff
    end. %% BUGBUG: extension

transform_opt_ServiceChangeAddress(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_ServiceChangeAddress({Tag, Val}, State) ->
    Val2 = 
        case Tag of
            portNumber -> transform_portNumber(Val, State);
            ip4Address -> transform_IP4Address(Val, State);
            ip6Address -> transform_IP6Address(Val, State);
            domainName -> transform_DomainName(Val, State);
            deviceName -> transform_PathName(Val, State);
            mtpAddress -> transform_mtpAddress(Val, State)
        end,
    {Tag, Val2}.

transform_opt_serviceChangeVersion(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_serviceChangeVersion(Version, State) ->
    transform_version(Version, State).

transform_opt_ServiceChangeProfile(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_ServiceChangeProfile(#'ServiceChangeProfile'{profileName = Name,
                                                           version     = Version},
                                   State) ->
    Name2 = transform_Name(Name, State),
    #'ServiceChangeProfile'{profileName = resolve(profile, Name2, State),
                            version     = transform_version(Version, State)}.

transform_serviceChangeReason(Reason, State) ->
    transform_Value(Reason, State).

transform_opt_serviceChangeDelay(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_serviceChangeDelay(Delay, State) ->
    transform_UINT32(Delay, State).

transform_opt_serviceChangeMgcId(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_serviceChangeMgcId(MgcId, State) ->
    transform_MId(MgcId, State).

transform_opt_portNumber(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_portNumber(Port, State) ->
    transform_portNumber(Port, State).

transform_portNumber(Port, State) when integer(Port), Port >= 0 ->
    transform_UINT16(Port, State).

transform_ServiceChangeResParm(#'ServiceChangeResParm'{serviceChangeMgcId   = MgcId, 
                                                       serviceChangeAddress = Addr, 
                                                       serviceChangeVersion = Version, 
                                                       serviceChangeProfile = Profile}, 
                               State) ->
    #'ServiceChangeResParm'{serviceChangeMgcId   = transform_opt_serviceChangeMgcId(MgcId, State),
                            serviceChangeAddress = transform_opt_ServiceChangeAddress(Addr, State),
                            serviceChangeVersion = transform_opt_serviceChangeVersion(Version, State),
                            serviceChangeProfile = transform_opt_ServiceChangeProfile(Profile, State)}.

transform_PackagesDescriptor(Items, State) when list(Items) ->
    [transform_PackagesItem(I, State) || I <- Items].

transform_PackagesItem(#'PackagesItem'{packageName    = Name,
                                       packageVersion = Version},
                       State) ->
    Name2 = transform_Name(Name, State),
    #'PackagesItem'{packageName    = resolve(package, Name2, State),
                    packageVersion = transform_UINT16(Version, State)}.

transform_StatisticsDescriptor(Parms, State) when list(Parms) ->
    [transform_StatisticsParameter(P, State) || P <- Parms].

transform_StatisticsParameter(#'StatisticsParameter'{statName  = Name,
                                                     statValue = Value},
                              State) ->
    Name2 = transform_PkgdName(Name, State),
    #'StatisticsParameter'{statName  = resolve(statistics, Name2, State),
                           statValue = transform_Value(Value, State)}.

transform_opt_TimeNotation(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
transform_opt_TimeNotation(#'TimeNotation'{date = Date,
                                           time = Time},
                           State) ->
    #'TimeNotation'{date = transform_STRING(Date, State), % "yyyymmdd"
                    time = transform_STRING(Time, State)}.% "hhmmssss"

%% BUGBUG: Does not verify that string must contain at least one char
%% BUGBUG: This violation of the is required in order to comply with
%% BUGBUG: the dd/ce ds parameter that may possibly be empty.
transform_Value(String, State) when list(String) ->
    String.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Encode an octet string, escape } by \ if necessary 
transform_OCTET_STRING(String, State, Min, Max) when list(String) ->
    verify_count(length(String), Min, Max),
    String.

transform_QUOTED_STRING(String, State) when list(String) ->
    verify_count(length(String), 1, infinity),
    String.

%% The internal format of hex digits is a list of octets
%% Min and Max means #hexDigits
%% Leading zeros are prepended in order to fulfill Min
transform_HEXDIG(Octets, State, Min, Max) when list(Octets) ->
    verify_count(length(Octets), Min, Max),
    Octets.

transform_DIGIT(Val, State, Min, Max) ->
    transform_integer(Val, State, Min, Max).

transform_STRING(String, State) when list(String) ->
    String.

transform_STRING(String, State, Min, Max) when list(String) ->
    verify_count(length(String), Min, Max),
    String.

transform_UINT16(Val, State) ->
    transform_integer(Val, State, 0, 65535).

transform_UINT32(Val, State) ->
    transform_integer(Val, State, 0, 4294967295).

transform_integer(Int, State, Min, Max) ->
    verify_count(Int, Min, Max),
    Int.

%% Verify that Count is within the range of Min and Max
verify_count(Count, Min, Max) ->
    if
        integer(Count) ->
            if
                integer(Min), Count >= Min ->
                    if
                        integer(Max), Count =< Max ->
                            Count;
                        Max == infinity ->
                            Count;
                        true ->
                            exit({count_too_large, Count, Max})
                    end;
                true ->
                    exit({count_too_small, Count, Min})
            end;
        true ->
            exit({count_not_an_integer, Count})
    end.
