%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%%----------------------------------------------------------------------
%% Purpose: Transform internal form of Megaco/H.248 messages
%%----------------------------------------------------------------------

-module(megaco_transform).

-include_lib("megaco/include/megaco.hrl").
-include_lib("megaco/include/megaco_message_v1.hrl").

-export([enc_MegacoMessage/1,
	 enc_Transaction/1,
	 enc_CommandRequest/1,
	 enc_ActionReply/1]).

-define(META_ENC(Type, Item), Item) .
%% -define(META_ENC(Type, Item), megaco_meta_package:encode(text, Type, Item)).
%% -define(META_DEC(Type, Item), megaco_meta_package:decode(text, Type, Item)).

enc_MegacoMessage(Mega) ->
    State = undefined,
    enc_MegacoMessage(Mega, State).

enc_Transaction(Trans) ->
    State = undefined,
    enc_Transaction(Trans, State).

enc_CommandRequest(CmdReq) ->
    State = undefined,
    enc_CommandRequest(CmdReq, State).

enc_ActionReply(ActRep) ->
    State = undefined,
    enc_ActionReply(ActRep, State).

enc_MegacoMessage(#'MegacoMessage'{authHeader = Auth,
				   mess       = Mess},
		  State) ->
    #'MegacoMessage'{authHeader = enc_opt_AuthenticationHeader(Auth, State),
		     mess       = enc_Message(Mess, State)}.

enc_opt_AuthenticationHeader(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_AuthenticationHeader(#'AuthenticationHeader'{secParmIndex = SPI,
						 seqNum       = SN,
						 ad           = AuthData},
			 State) ->
    #'AuthenticationHeader'{secParmIndex = enc_SecurityParmIndex(SPI, State),
			    seqNum       = enc_SequenceNum(SN, State),
			    ad           = enc_AuthData(AuthData, State)}.

enc_SecurityParmIndex(SPI, State) ->
    enc_HEXDIG(SPI, State, 8, 8). % BUGBUG: Mismatch between ASN.1 and ABNF

enc_SequenceNum(SN, State) ->
    enc_HEXDIG(SN, State, 8, 8).  % BUGBUG: Mismatch between ASN.1 and ABNF

enc_AuthData(AuthData, State) ->
    enc_HEXDIG(AuthData, State, 32, 64).  % BUGBUG: Mismatch between ASN.1 and ABNF

enc_Message(#'Message'{version     = Version,
		       mId         = MID,
		       messageBody = Body},
	    State) ->
    #'Message'{version     = enc_version(Version, State),
	       mId         = enc_MId(MID, State),
	       messageBody = enc_Message_messageBody(Body, State)}.

enc_version(Version, State) ->
    enc_DIGIT(Version, State, 0, 99).

enc_Message_messageBody({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    messageError -> enc_ErrorDescriptor(Val, State);
	    transactions when list(Val) -> [enc_Transaction(T, State) || T <- Val]
	end,
    {Tag, Val2}.

enc_MId({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    ip4Address -> enc_IP4Address(Val, State);
	    ip6Address -> enc_IP6Address(Val, State);
	    domainName -> enc_DomainName(Val, State);
	    deviceName -> enc_PathName(Val, State);
	    mtpAddress -> enc_mtpAddress(Val, State)
	end,
    {Tag, Val2}.

enc_mtpAddress(MtpAddr, State) ->
    enc_OCTET_STRING(MtpAddr, State, 5, 5). % BUBGBUG: length?

enc_DomainName(#'DomainName'{name       = Name,
			     portNumber = Port},
	       State) ->
    %% BUGBUG: (ALPHA / DIGIT) *63(ALPHA / DIGIT / "-" / ".")
    #'DomainName'{name       = enc_STRING(Name, State, 1, 64),
		  portNumber = enc_opt_portNumber(Port, State)}.

enc_IP4Address(#'IP4Address'{address    = [A1, A2, A3, A4],
			     portNumber = Port},
	       State) ->
    #'IP4Address'{address    = [enc_V4hex(A1, State),
				enc_V4hex(A2, State),
				enc_V4hex(A3, State),
				enc_V4hex(A4, State)],
		  portNumber = enc_opt_portNumber(Port, State)}.

enc_V4hex(Val, State) ->
    enc_DIGIT(Val, State, 0, 255).

enc_IP6Address(Val, State) ->
    exit(ipv6_not_supported). %% BUGBUG: nyi

enc_PathName(Path, State) ->
    %% BUGBUG: ["*"] NAME *("/" / "*"/ ALPHA / DIGIT /"_" / "$" ) 
    %% BUGBUG: ["@" pathDomainName ]
    enc_STRING(Path, State, 1, 64).

enc_Transaction({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    transactionRequest ->     enc_TransactionRequest(Val, State);
	    transactionPending ->     enc_TransactionPending(Val, State);
	    transactionReply ->	      enc_TransactionReply(Val, State);
	    transactionResponseAck -> enc_TransactionResponseAck(Val, State)
	end,
    {Tag, Val2}.

enc_TransactionResponseAck(#'TransactionResponseAck'{firstAck = First,
						     lastAck  = Last},
			   State) ->
    #'TransactionResponseAck'{firstAck = enc_TransactionId(First, State),
			      lastAck  = enc_opt_TransactionId(Last, State)}.

enc_opt_TransactionId(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_TransactionId(Id, State) ->
    enc_TransactionId(Id, State).

enc_TransactionId(Id, State) ->
    enc_UINT32(Id, State).

enc_TransactionRequest(#'TransactionRequest'{transactionId = Id,
					     actions       = Actions},
		       State) when list(Actions) ->

    #'TransactionRequest'{transactionId = enc_TransactionId(Id, State),
			  actions       = [enc_ActionRequest(ActReq, State) || ActReq <- Actions]}.

enc_TransactionPending(#'TransactionPending'{transactionId = Id},
		       State) ->
    #'TransactionPending'{transactionId = enc_TransactionId(Id, State)}.

enc_TransactionReply(#'TransactionReply'{transactionId     = Id,
					 immAckRequired    = ImmAck,
					 transactionResult = TransRes},
		     State) ->
    #'TransactionReply'{transactionId     = enc_TransactionId(Id, State),
			immAckRequired    = enc_opt_null(ImmAck, State),
			transactionResult = enc_TransactionReply_transactionResult(TransRes, State)}.

enc_opt_null(asn1_NOVALUE, State) -> asn1_NOVALUE;
enc_opt_null('NULL', State)       -> 'NULL'.

enc_TransactionReply_transactionResult({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    transactionError ->
		enc_ErrorDescriptor(Val, State);
	    actionReplies when list(Val), Val /= [] ->
		[enc_ActionReply(ActRep, State) || ActRep <- Val]
	end,
    {Tag, Val2}.

enc_opt_ErrorDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_ErrorDescriptor(ErrDesc, State) ->
    enc_ErrorDescriptor(ErrDesc, State).

enc_ErrorDescriptor(#'ErrorDescriptor'{errorCode = Code,
				       errorText = Text},
		    State) ->
    #'ErrorDescriptor'{errorCode = enc_ErrorCode(Code, State),
		       errorText = enc_opt_ErrorText(Text, State)}.

enc_ErrorCode(Code, State) ->
    enc_DIGIT(Code, State, 0, 999).

enc_opt_ErrorText(asn1_NOVALUE, State)  ->
    asn1_NOVALUE;
enc_opt_ErrorText(Text, State)  ->
    enc_QUOTED_STRING(Text, State).

enc_ContextID(CtxId, State) ->
    case CtxId of
	?megaco_all_context_id    -> ?megaco_all_context_id;
	?megaco_null_context_id   -> ?megaco_null_context_id;
	?megaco_choose_context_id -> ?megaco_choose_context_id;
	Int when integer(Int)     -> enc_UINT32(Int, State)
    end.

enc_ActionRequest(#'ActionRequest'{contextId           = CtxId,
				   contextRequest      = CtxReq,
				   contextAttrAuditReq = CtxAuditReq,
				   commandRequests     = CmdReqList},
		  State) ->
    #'ActionRequest'{contextId           = enc_ContextID(CtxId, State),
		     contextRequest      = enc_opt_ContextRequest(CtxReq, State),
		     contextAttrAuditReq = enc_opt_ContextAttrAuditRequest(CtxAuditReq, State),
		     commandRequests     = [enc_CommandRequest(CmdReq, State) || CmdReq <- CmdReqList]}.

enc_ActionReply(#'ActionReply'{contextId       = CtxId,
			       errorDescriptor = ErrDesc,
			       contextReply    = CtxRep,
			       commandReply    = CmdRepList},
		State) ->
    CmdRepList2 = [enc_CommandReply(CmdRep, State) || CmdRep <- CmdRepList],
    #'ActionReply'{contextId       = enc_ContextID(CtxId, State),
		   errorDescriptor = enc_opt_ErrorDescriptor(ErrDesc, State),
		   contextReply    = enc_opt_ContextRequest(CtxRep, State),
		   commandReply    = CmdRepList2}.

enc_opt_ContextRequest(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_ContextRequest(#'ContextRequest'{priority    = Prio,
				     emergency   = Em,
				     topologyReq = TopReqList},
		   State) ->
    Prio2 = 
	case Prio of
	    asn1_NOVALUE -> asn1_NOVALUE;
	    _            -> enc_integer(Prio, State, 0, 15)
	end,
    Em2 = 
	case Em of
	    asn1_NOVALUE -> asn1_NOVALUE;
	    false        -> false;
	    true         -> true
	end,
    TopReqList2 = 
	case TopReqList of
	    asn1_NOVALUE -> asn1_NOVALUE;
	    _            -> [enc_TopologyRequest(TopReq, State) ||
				TopReq <- TopReqList]
     end,
    #'ContextRequest'{priority    = Prio2,
		      emergency   = Em2,
		      topologyReq = TopReqList2}.

enc_opt_ContextAttrAuditRequest(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_ContextAttrAuditRequest(#'ContextAttrAuditRequest'{topology = Top,
							   emergency = Em,
							   priority = Prio},
			    State) ->
    #'ContextAttrAuditRequest'{topology  = enc_opt_null(Top, State),
			       emergency = enc_opt_null(Em, State),
			       priority  = enc_opt_null(Prio, State)}.

enc_CommandRequest(#'CommandRequest'{command        = Cmd,
				     optional       = Opt,
				     wildcardReturn = Wild},
		   State) ->
    #'CommandRequest'{optional       = enc_opt_null(Opt, State),
		      wildcardReturn = enc_opt_null(Wild, State),
		      command        = enc_Command(Cmd, State)}.

enc_Command({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    addReq ->		 enc_AmmRequest(Val, State);
	    moveReq ->		 enc_AmmRequest(Val, State);
	    modReq ->		 enc_AmmRequest(Val, State);
	    subtractReq ->	 enc_SubtractRequest(Val, State);
	    auditCapRequest ->	 enc_AuditRequest(Val, State);
	    auditValueRequest -> enc_AuditRequest(Val, State);
	    notifyReq ->	 enc_NotifyRequest(Val, State);
	    serviceChangeReq ->  enc_ServiceChangeRequest(Val, State)
	end,
    {Tag, Val2}.

enc_CommandReply({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    addReply ->		  enc_AmmsReply(Val, State);
	    moveReply ->	  enc_AmmsReply(Val, State);
	    modReply ->		  enc_AmmsReply(Val, State);
	    subtractReply ->	  enc_AmmsReply(Val, State);
	    auditCapReply ->	  enc_AuditReply(Val, State);
	    auditValueReply ->	  enc_AuditReply(Val, State);
	    notifyReply ->	  enc_NotifyReply(Val, State);
	    serviceChangeReply -> enc_ServiceChangeReply(Val, State)
	end,
    {Tag, Val2}.

enc_TopologyRequest(#'TopologyRequest'{terminationFrom   = From,
				       terminationTo     = To,
				       topologyDirection = Dir},
		    State) ->
    Dir2 = 
	case Dir of
	    bothway -> bothway;
	    isolate -> isolate;
	    oneway ->  oneway
	end,
    #'TopologyRequest'{terminationFrom   = enc_TerminationID(From, State),
		       terminationTo     = enc_TerminationID(To, State),
		       topologyDirection = Dir2}.

enc_AmmRequest(#'AmmRequest'{terminationID = IdList,
			     descriptors   = DescList},
	       State) ->
    #'AmmRequest'{terminationID = [enc_TerminationID(Id, State) ||
				      Id <- IdList],
		  descriptors   = [enc_ammDescriptor(Desc, State) ||
				      Desc <- DescList]}.

enc_ammDescriptor({Tag, Desc}, State) ->
    Desc2 = 
	case Tag of
	    mediaDescriptor       -> enc_MediaDescriptor(Desc, State);
	    modemDescriptor       -> enc_ModemDescriptor(Desc, State);      
	    muxDescriptor         -> enc_MuxDescriptor(Desc, State);   
	    eventsDescriptor      -> enc_EventsDescriptor(Desc, State);      
	    eventBufferDescriptor -> enc_EventBufferDescriptor(Desc, State); 
	    signalsDescriptor     -> enc_SignalsDescriptor(Desc, State);    
	    digitMapDescriptor    -> enc_DigitMapDescriptor(Desc, State);    
	    auditDescriptor       -> enc_AuditDescriptor(Desc, State)
	end,
    {Tag, Desc2}.

enc_AmmsReply(#'AmmsReply'{terminationID    = IdList,
			   terminationAudit = TermAudit},
	      State) ->
    TermAudit2 =
	case TermAudit of
	    asn1_NOVALUE -> asn1_NOVALUE;
	    _            -> enc_TerminationAudit(TermAudit, State)
	end,
    #'AmmsReply'{terminationID    = [enc_TerminationID(Id, State) ||
					Id <- IdList],
		 terminationAudit = TermAudit2}.

enc_SubtractRequest(#'SubtractRequest'{terminationID   = IdList,
				       auditDescriptor = Desc},
		    State) ->
    Desc2 =
	case Desc of
	    asn1_NOVALUE -> asn1_NOVALUE;
	    _            -> enc_AuditDescriptor(Desc, State)
	end,
    #'SubtractRequest'{terminationID   = [enc_TerminationID(Id, State) ||
					     Id <- IdList],
		       auditDescriptor = Desc2}.

enc_AuditRequest(#'AuditRequest'{terminationID = IdList,
				 auditDescriptor = Desc},
		 State) ->
    Desc2 =
	case Desc of
	    asn1_NOVALUE -> asn1_NOVALUE;
	    _            -> enc_AuditDescriptor(Desc, State)
	end,
    #'AuditRequest'{terminationID = [enc_TerminationID(Id, State) ||
					Id <- IdList],
		    auditDescriptor = Desc2}.

%% auditReply           = (AuditValueToken / AuditCapToken ) 
%% 			  ( contextTerminationAudit  / auditOther)
%% auditOther           = EQUAL TerminationID LBRKT 
%% 			  terminationAudit RBRKT
%% terminationAudit     = auditReturnParameter *(COMMA auditReturnParameter) 
%% 
%% contextTerminationAudit = EQUAL CtxToken ( terminationIDList / 
%% 			  LBRKT errorDescriptor RBRKT )
enc_AuditReply(#'AuditReply'{terminationID = Id,
			     auditResult   = AuditRes},
	       State) ->
    #'AuditReply'{terminationID = enc_TerminationID(Id, State),
		  auditResult   = enc_auditResult(AuditRes, State)}.

enc_auditResult({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    contextAuditResult ->
		%% contextTerminationAudit
		%% BUGBUG: Mismatch between ASN.1 and ABNF
		%% BUGBUG: LBRKT errorDescriptor RBRKT
		[enc_TerminationID(Id, State) || Id <- Val];
	    terminationAuditResult ->
		%% auditOther
		enc_TerminationAudit(Val, State)
	end,
    {Tag, Val2}.

enc_AuditDescriptor(#'AuditDescriptor'{auditToken = Tokens},
		    State) ->
     case Tokens of
	 asn1_NOVALUE -> asn1_NOVALUE;
	 _            -> [enc_auditItem(Token, State) || Token <- Tokens]
     end.

enc_auditItem(Token, State) ->
    case Token of
	muxToken               -> muxToken;
	modemToken             -> modemToken;
	mediaToken             -> mediaToken;
	eventsToken            -> eventsToken;
	signalsToken           -> signalsToken;
	digitMapToken          -> digitMapToken;
	statsToken             -> statsToken;
	observedEventsToken    -> observedEventsToken;
	packagesToken          -> packagesToken;
	eventBufferToken       -> eventBufferToken
    end.

enc_TerminationAudit(ParmList, State) when list(ParmList) ->
    [enc_AuditReturnParameter(Parm, State) || Parm <- ParmList].

enc_AuditReturnParameter({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    errorDescriptor ->
		enc_ErrorDescriptor(Val, State);
	    mediaDescriptor ->
		enc_MediaDescriptor(Val, State);
	    modemDescriptor ->
		enc_ModemDescriptor(Val, State);
	    muxDescriptor ->
		enc_MuxDescriptor(Val, State);
	    eventsDescriptor ->
		enc_EventsDescriptor(Val, State);
	    eventBufferDescriptor ->
		enc_EventBufferDescriptor(Val, State);
	    signalsDescriptor ->
		enc_SignalsDescriptor(Val, State);
	    digitMapDescriptor ->
		enc_DigitMapDescriptor(Val, State);
	    observedEventsDescriptor ->
		enc_ObservedEventsDescriptor(Val, State);
	    statisticsDescriptor ->
		enc_StatisticsDescriptor(Val, State);
	    packagesDescriptor ->
		enc_PackagesDescriptor(Val, State);
	    emptyDescriptors ->
		enc_EmptyDescriptors(Val, State)
	end,
    {Tag, Val2}.

enc_EmptyDescriptors(#'AuditDescriptor'{auditToken = Tokens},
		     State) ->
     case Tokens of
	 asn1_NOVALUE -> asn1_NOVALUE;
	 _            -> [enc_auditItem(Token, State) || Token <- Tokens]
     end.

enc_NotifyRequest(#'NotifyRequest'{terminationID            = IdList,
				   observedEventsDescriptor = ObsDesc,
				   errorDescriptor          = ErrDesc},
		  State) ->
    %% BUGBUG: Mismatch between ASN.1 and ABNF
    %% BUGBUG: The following ought to be a 'choice'
    #'NotifyRequest'{terminationID            = [enc_TerminationID(Id, State) ||
						    Id <- IdList],
		     observedEventsDescriptor = enc_ObservedEventsDescriptor(ObsDesc, State),
		     errorDescriptor          = enc_opt_ErrorDescriptor(ErrDesc, State)}.

enc_NotifyReply(#'NotifyReply'{terminationID = IdList,
			       errorDescriptor = ErrDesc},
		State) ->
    #'NotifyReply'{terminationID = enc_opt_TerminationIDList(IdList, State),
		   errorDescriptor = enc_opt_ErrorDescriptor(ErrDesc, State)}.

enc_opt_TerminationIDList(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_TerminationIDList(IdList, State) when list(IdList) ->
    [enc_TerminationID(Id, State) || Id <- IdList].

enc_ObservedEventsDescriptor(#'ObservedEventsDescriptor'{requestId        = Id,
							 observedEventLst = Events},
 			     State) when list (Events) ->
    #'ObservedEventsDescriptor'{requestId        = enc_RequestID(Id, State),
				observedEventLst = [enc_ObservedEvent(E, State) || E <- Events]}.

%% ;time per event, because it might be buffered
%% observedEvent        = [ TimeStamp LWSP COLON] LWSP 
%% 			  pkgdName [ LBRKT observedEventParameter
%% 			  *(COMMA observedEventParameter) RBRKT ]
%% 
%% ;at-most-once eventStream, every eventParameterName at most once
%% observedEventParameter = eventStream / eventOther

enc_ObservedEvent(#'ObservedEvent'{eventName    = Name,
				   streamID     = Id,
				   eventParList = Parms,
				   timeNotation = Time},
		  State) ->
    #'ObservedEvent'{eventName    = enc_EventName(Name, State),
		     streamID     = enc_opt_StreamID(Id, State),
		     eventParList = enc_EventParameter(Parms, State),
		     timeNotation = enc_opt_TimeNotation(Time, State)}.

enc_EventName(Name, State) ->
    PkgdName = ?META_ENC(event, Name),
    enc_PkgdName(PkgdName, State).

enc_EventParameter(#'EventParameter'{eventParameterName = Name,
				     value              = Value},
	       State) ->
    PkgdName = ?META_ENC(event, Name),
    #'EventParameter'{eventParameterName = enc_Name(Name, State),
		      value              = enc_Value(Value, State)}.

enc_ServiceChangeRequest(#'ServiceChangeRequest'{terminationID      = IdList,
						 serviceChangeParms = Parms},
			 State) ->
    #'ServiceChangeRequest'{terminationID      = [enc_TerminationID(Id, State) || Id <- IdList],
			    serviceChangeParms = enc_ServiceChangeParm(Parms, State)}.

%% serviceChangeReply   = ServiceChangeToken EQUAL TerminationID
%% 			  [LBRKT (errorDescriptor / 
%% 			  serviceChangeReplyDescriptor) RBRKT]
%% serviceChangeReplyDescriptor = ServicesToken LBRKT
%% 			  servChgReplyParm *(COMMA servChgReplyParm) RBRKT
%% 
%% ;at-most-once. Version is REQUIRED on first ServiceChange response
%% servChgReplyParm     = (serviceChangeAddress / serviceChangeMgcId /
%% 			  serviceChangeProfile / serviceChangeVersion )
enc_ServiceChangeReply(#'ServiceChangeReply'{terminationID       = IdList,
					     serviceChangeResult = Res},
		       State) ->
  #'ServiceChangeReply'{terminationID       = [enc_TerminationID(Id, State) || Id <- IdList],
			serviceChangeResult = enc_ServiceChangeResult(Res, State)}.

enc_ServiceChangeResult({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    errorDescriptor       -> enc_ErrorDescriptor(Val, State);
	    serviceChangeResParms -> enc_ServiceChangeResParm(Val, State)
	end,
    {Tag, Val2}.

%% TerminationID        = "ROOT" / pathNAME / "$" / "*"
%% ; Total length of pathNAME must not exceed 64 chars.
%% pathNAME             = ["*"] NAME *("/" / "*"/ ALPHA / DIGIT /"_" / "$" ) 
%% 			  ["@" pathDomainName ]

enc_TerminationID(#'TerminationID'{wildcard = Wild,
				   id       = Id},
		  State) ->
    %% BUGBUG:  ?META_ENC(term_id, TermId),
    #'TerminationID'{wildcard = Wild,
		     id       = Id};
enc_TerminationID(#megaco_term_id{contains_wildcards = IsWild,
				  id                 = Id},
		  State) ->
    %% BUGBUG:  ?META_ENC(term_id, TermId),
    #megaco_term_id{contains_wildcards = enc_bool(IsWild, State),
		    id                 = [enc_term_id_component(Sub, State) || Sub <- Id]}.

enc_bool(true, State)  -> true;
enc_bool(false, State) -> false.

enc_term_id_component(Sub, State) ->
    case Sub of
	all    -> all;
	choose -> choose;
	Char when integer(Char) -> Char
    end.

%% mediaDescriptor      = MediaToken LBRKT mediaParm *(COMMA mediaParm) RBRKT
%% ; at-most-once per item
%% ; and either streamParm or streamDescriptor but not both
%% mediaParm            = (streamParm / streamDescriptor / 
%% 			   terminationStateDescriptor)
%% ; at-most-once
%% streamParm           = ( localDescriptor / remoteDescriptor / 
%% 			   localControlDescriptor )
%% streamDescriptor     = StreamToken EQUAL StreamID LBRKT streamParm 
%% 			  *(COMMA streamParm) RBRKT
enc_MediaDescriptor(#'MediaDescriptor'{termStateDescr = TermState,
				       streams        = Streams},
		    State) ->
  #'MediaDescriptor'{termStateDescr = enc_opt_TerminationStateDescriptor(TermState, State),
		     streams        = enc_streams(Streams, State)}.

enc_streams({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    oneStream   -> enc_StreamParms(Val, State);
	    multiStream -> enc_StreamDescriptor(Val, State)
	end,
    {Tag, Val2}.

enc_StreamParms(#'StreamParms'{localControlDescriptor = Control,
			       localDescriptor        = Local,
			       remoteDescriptor       = Remote},
		State) ->
    #'StreamParms'{localControlDescriptor = enc_opt_LocalControlDescriptor(Control, State),
		   localDescriptor        = enc_opt_LocalRemoteDescriptor(Local, State),
		   remoteDescriptor       = enc_opt_LocalRemoteDescriptor(Remote, State)}.

enc_StreamDescriptor(#'StreamDescriptor'{streamID    = Id,
					 streamParms = Parms},
		     State) ->
    #'StreamDescriptor'{streamID    = enc_StreamID(Id, State),
			streamParms = enc_StreamParms(Parms, State)}.

%% localControlDescriptor = LocalControlToken LBRKT localParm 
%% 			    *(COMMA localParm) RBRKT
%% 
%% ; at-most-once per item
%% localParm            = ( streamMode / propertyParm /
%%                          reservedValueMode  / reservedGroupMode ) 
%% reservedValueMode       = ReservedValueToken EQUAL ( "ON" / "OFF" ) 
%% reservedGroupMode       = ReservedGroupToken EQUAL ( "ON" / "OFF" ) 
%% 
%% reservedMode	     = ReservedToken EQUAL ( "ON" / "OFF" )
%% 
%% streamMode           = ModeToken EQUAL streamModes
enc_opt_LocalControlDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_LocalControlDescriptor(#'LocalControlDescriptor'{streamMode    = Mode,
							 reserveGroup  = Group,
							 reserveValue  = Value,
							 propertyParms = Props},
			   State) ->
  #'LocalControlDescriptor'{streamMode    = enc_opt_StreamMode(Mode, State),
			    reserveGroup  = enc_bool(Group, State),
			    reserveValue  = enc_bool(Value, State),
			    propertyParms = [enc_PropertyParm(P, State) || P <- Props]}.

enc_opt_StreamMode(Mode, State) ->
     case Mode of
	 asn1_NOVALUE -> asn1_NOVALUE;
	 sendOnly     -> sendOnly;
	 recvOnly     -> recvOnly;
	 sendRecv     -> sendRecv;
	 inactive     -> inactive;
	 loopBack     -> loopBack
     end.

enc_Name(Name, State) ->
    %% BUGBUG: ?META_ENC
    %% BUGBUG: NAME = ALPHA *63(ALPHA / DIGIT / "_" )
    enc_STRING(Name, State, 1, 64).

enc_PkgdName(Name, State) ->
    %% BUGBUG: ?META_ENC
    %% BUGBUG:  pkgdName =  (NAME / "*")  SLASH  (ItemID / "*" )
    enc_OCTET_STRING(Name, State, 1, 64).

%% When text encoding the protocol, the descriptors consist of session
%% descriptions as defined in SDP (RFC2327), except that the "s=", "t="
%% and "o=" lines are optional. When multiple session descriptions are
%% provided in one descriptor, the "v=" lines are required as delimiters;
%% otherwise they are optional.  Implementations shall accept session
%% descriptions that are fully conformant to RFC2327. When binary
%% encoding the protocol the descriptor consists of groups of properties
%% (tag-value pairs) as specified in Annex C.  Each such group may
%% contain the parameters of a session description.
enc_opt_LocalRemoteDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_LocalRemoteDescriptor(#'LocalRemoteDescriptor'{propGrps = Groups},
			      State) ->
    #'LocalRemoteDescriptor'{propGrps = [enc_PropertyGroup(G, State) || G <- Groups]}.

enc_PropertyGroup(Props, State) ->
    [enc_PropertyGroupParm(P, State) || P <- Props].

enc_PropertyGroupParm(#'PropertyParm'{name  = Name,
				      value = Value},
		      State) ->
    #'PropertyParm'{name  = enc_PkgdName(Name, State),
		    value = enc_OCTET_STRING(Value, State, 0, infinity)}.

%% propertyParm         = pkgdName parmValue
%% parmValue            = (EQUAL alternativeValue/ INEQUAL VALUE)
%% alternativeValue     = ( VALUE / LSBRKT VALUE *(COMMA VALUE) RSBRKT  / 
%% 			  LSBRKT VALUE DOT DOT VALUE RSBRKT )
enc_PropertyParm(#'PropertyParm'{name  	   = Name,
				 value 	   = Values,
				 extraInfo = Extra},
		 State) ->
    PkgdName = ?META_ENC(property, Name),
    #'PropertyParm'{name      = enc_PkgdName(PkgdName, State),
		    value     = [enc_Value(V, State) || V <- Values], % BUGBUG: ???
		    extraInfo = enc_opt_extraInfo(Extra, State)}.
     
enc_opt_extraInfo(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_extraInfo({relation, Rel}, State) ->
    Rel2 = 
	case Rel of
	    greaterThan -> greaterThan;
	    smallerThan -> smallerThan;
	    unequalTo   -> unequalTo
	end,
    {relation, Rel2};
enc_opt_extraInfo({range, Range}, State) ->
    Range2 = enc_bool(Range, State),
    {range, Range2};
enc_opt_extraInfo({sublist, Sub}, State) ->
    Sub2 = enc_bool(Sub, State),
    {sublist, Sub2}.

enc_opt_TerminationStateDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_TerminationStateDescriptor(#'TerminationStateDescriptor'{propertyParms      = Props,
								 eventBufferControl = Control,
								 serviceState       = Service},
			       State) ->
  #'TerminationStateDescriptor'{propertyParms      = [enc_PropertyParm(P, State) || P <- Props],
				eventBufferControl = enc_opt_EventBufferControl(Control, State),
				serviceState       = enc_opt_ServiceState(Service, State)}.

enc_opt_EventBufferControl(Control, State) ->
     case Control of
	 asn1_NOVALUE -> asn1_NOVALUE;
	 off          -> off;
	 lockStep     -> lockStep
     end.
    
enc_opt_ServiceState(Service, State) ->
     case Service of
	 asn1_NOVALUE -> asn1_NOVALUE;
	 test         -> test;
	 outOfSvc     -> outOfSvc;
	 inSvc        -> inSvc
     end.

enc_MuxDescriptor(#'MuxDescriptor'{muxType  = Type,
				   termList = IdList},
		  State) ->
    #'MuxDescriptor'{muxType  = enc_MuxType(Type, State),
		     termList = [enc_TerminationID(Id, State) || Id <- IdList]}.

enc_MuxType(Type, State) ->
    case Type of
	h221 -> h221;
	h223 -> h223;
	h226 -> h226;
	v76  -> v76
    end.

enc_opt_StreamID(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_StreamID(Id, State) ->
    enc_StreamID(Id, State).

enc_StreamID(Id, State) ->
    enc_UINT16(Id, State).

enc_EventsDescriptor(#'EventsDescriptor'{requestID = Id,
					 eventList = Events},
		     State) ->
    #'EventsDescriptor'{requestID = enc_RequestID(Id, State),
			eventList = [enc_RequestedEvent(E, State) || E <- Events]}.

enc_RequestedEvent(#'RequestedEvent'{pkgdName    = Name,
				     streamID    = Id,
				     evParList   = Parms,
				     eventAction = Actions},
		   State)  ->
    PkgdName = ?META_ENC(event, Name),
    #'RequestedEvent'{pkgdName    = enc_PkgdName(PkgdName, State),
		      streamID    = enc_opt_StreamID(Id, State),
		      eventAction = enc_opt_RequestedActions(Actions, State),
		      evParList   = [enc_EventParameter(P, State) || P <- Parms]}.

enc_opt_RequestedActions(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_RequestedActions(#'RequestedActions'{keepActive        = Keep,
					     eventDM           = DM,
					     secondEvent       = Event,
					     signalsDescriptor = SigDesc},
		    State) ->
  #'RequestedActions'{keepActive        = enc_keepActive(Keep, State),
		      eventDM    	= enc_opt_EventDM(DM, State),
		      secondEvent       = enc_SecondEventsDescriptor(Event, State),
		      signalsDescriptor = enc_opt_SignalsDescriptor(SigDesc, State)}.

enc_opt_keepActive(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_keepActive(Keep, State) ->
    enc_keepActive(Keep, State).

enc_keepActive(Keep, State) ->
    enc_bool(Keep, State).
    
enc_opt_EventDM(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_EventDM({Tag, Val}, State) ->
    Val2 = 
    case Tag of
	digitMapName  -> enc_DigitMapName(Val, State);
	digitMapValue -> enc_DigitMapValue(Val, State)
    end,
    {Tag, Val2}.

enc_SecondEventsDescriptor(#'SecondEventsDescriptor'{requestID = Id,
						     eventList = Events},
			   State) ->
    #'SecondEventsDescriptor'{requestID = enc_RequestID(Id, State),
			      eventList = [enc_SecondRequestedEvent(E, State) || E <- Events]}.

enc_SecondRequestedEvent(#'SecondRequestedEvent'{pkgdName    = Name,
						 streamID    = Id,
						 evParList   = Parms,
						 eventAction = Actions},
			 State) ->
    PkgdName = ?META_ENC(event, Name),
    #'SecondRequestedEvent'{pkgdName    = enc_PkgdName(PkgdName, State),
			    streamID    = enc_opt_StreamID(Id, State),
			    eventAction = enc_opt_SecondRequestedActions(Actions, State),
			    evParList   = [enc_EventParameter(P, State) || P <- Parms]}.


enc_opt_SecondRequestedActions(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_SecondRequestedActions(#'SecondRequestedActions'{keepActive        = Keep,
							 eventDM           = DM,
							 signalsDescriptor = SigDesc},
			       State) ->
    #'SecondRequestedActions'{keepActive        = enc_keepActive(Keep, State),
			      eventDM    	= enc_opt_EventDM(DM, State),
			      signalsDescriptor = enc_opt_SignalsDescriptor(SigDesc, State)}.

enc_EventBufferDescriptor(EventSpecs, State) ->
    [enc_EventSpec(ES, State) || ES <- EventSpecs].

enc_EventSpec(#'EventSpec'{eventName    = Name,
			   streamID     = Id,
			   eventParList = Parms},
	      State) ->
    #'EventSpec'{eventName    = enc_EventName(Name, State),
		 streamID     = enc_opt_StreamID(Id, State),
		 eventParList = enc_EventParameter(Parms, State)}.

enc_opt_SignalsDescriptor(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_SignalsDescriptor(SigDesc, State) ->
    enc_SignalsDescriptor(SigDesc, State).

enc_SignalsDescriptor(SigDesc, State)  when list(SigDesc) ->
    [enc_SignalRequest(SigReq, State) || SigReq <- SigDesc].

enc_SignalRequest({Tag, Val}, State) ->
    Val2 =
	case Tag of
	    signal     -> enc_Signal(Val, State);
	    seqSigList -> enc_SeqSigList(Val, State)
	end,
    {Tag, Val2}.


enc_SeqSigList(#'SeqSigList'{id         = Id,
			     signalList = SigList},
	       State) when list(SigList) ->
    #'SeqSigList'{id         = enc_UINT16(Id, State),
		  signalList = [enc_Signal(Sig, State) || Sig <- SigList]}.

enc_Signal(#'Signal'{signalName       = Name,
		     streamID         = Id,
		     sigType          = Type,
		     duration         = Dur,
		     notifyCompletion = Compl,
		     keepActive       = Keep,
		     sigParList       = Parms},
	   State) ->
    #'Signal'{signalName       = enc_SignalName(Name, State),
	      streamID         = enc_opt_StreamID(Id, State),
	      sigType          = enc_opt_SignalType(Type, State),
	      duration         = enc_opt_duration(Dur, State),
	      notifyCompletion = enc_opt_NotifyCompletion(Compl, State),
	      keepActive       = enc_opt_keepActive(Keep, State),
	      sigParList       = [enc_SigParameter(P, State) || P <- Parms]}.

enc_opt_duration(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_duration(Dur, State) ->
     enc_UINT16(Dur, State).

enc_opt_NotifyCompletion(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_NotifyCompletion(Items, State) when list(Items) ->
    [enc_notifyCompletionItem(I, State) || I <- Items].

enc_notifyCompletionItem(Item, State) ->
    case Item of
	onTimeOut                   -> onTimeOut;
        onInterruptByEvent          -> onInterruptByEvent;
        onInterruptByNewSignalDescr -> onInterruptByNewSignalDescr;
        otherReason                 -> otherReason
    end.

enc_opt_SignalType(Type, State) ->
    case Type of
	asn1_NOVALUE -> asn1_NOVALUE;
	brief        ->   brief;
	onOff        ->   onOff;
	timeOut      -> timeOut
    end.

enc_SignalName(Name, State) ->
    PkgdName = ?META_ENC(signal, Name),
    enc_PkgdName(PkgdName, State).

enc_SigParameter(#'SigParameter'{sigParameterName = Name,
				 value            = Value},
		 State) ->
    %% BUGBUG: ?META_ENC
    #'SigParameter'{sigParameterName = enc_Name(Name, State),
		    value            = enc_Value(Value, State)}.

enc_RequestID(Id, State) ->
    enc_UINT32(Id, State).

enc_ModemDescriptor(#'ModemDescriptor'{mtl = Types,
				       mpl = Props},
		    State) when list(Types), list(Props) -> 
    %% BUGBUG: Does not handle extensionParameter
     #'ModemDescriptor'{mtl = [enc_ModemType(T, State) || T <- Types],
			mpl = [enc_PropertyParm(P, State) || P <- Props]}.

enc_ModemType(Type, State) ->
    %% BUGBUG: Does not handle extensionParameter
    case Type of
        v18    	  -> v18;
        v22    	  -> v22;
        v22bis 	  -> v22bis;
        v32    	  -> v32;
        v32bis 	  -> v32bis;
        v34    	  -> v34;
        v90    	  -> v90;
        v91    	  -> v91;
        synchISDN -> synchISDN
    end.

enc_DigitMapDescriptor(#'DigitMapDescriptor'{digitMapName  = Name,
					     digitMapValue = Value},
		       State) ->
    #'DigitMapDescriptor'{digitMapName  = enc_opt_DigitMapName(Name, State),
			  digitMapValue = enc_opt_DigitMapValue(Value, State)}.

enc_opt_DigitMapName(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_DigitMapName(Name, State) ->
    enc_DigitMapName(Name, State).

enc_DigitMapName(Name, State) ->
    enc_Name(Name, State).

enc_opt_DigitMapValue(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_DigitMapValue(Value, State) ->
    enc_DigitMapValue(Value, State).

enc_DigitMapValue(#'DigitMapValue'{digitMapBody = Body,
				   startTimer   = Start,
				   shortTimer   = Short,
				   longTimer    = Long},
		  State) ->
    #'DigitMapValue'{startTimer   = enc_opt_timer(Start, State),
		     shortTimer   = enc_opt_timer(Short, State),
		     longTimer    = enc_opt_timer(Long, State),
		     digitMapBody = enc_STRING(Body, State)}. %% BUGBUG: digitMapBody not handled at all

enc_opt_timer(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_timer(Timer, State) ->
    enc_DIGIT(Timer, State, 0, 99).

enc_ServiceChangeParm(#'ServiceChangeParm'{serviceChangeMethod  = Method, 
					   serviceChangeAddress = Addr, 
					   serviceChangeVersion = Version, 
					   serviceChangeProfile = Profile, 
					   serviceChangeReason  = Reason, 
					   serviceChangeDelay   = Delay, 
					   serviceChangeMgcId   = MgcId, 
					   timeStamp            = Time},
		      State) ->
    #'ServiceChangeParm'{serviceChangeMethod  = enc_ServiceChangeMethod(Method, State),
			 serviceChangeAddress = enc_opt_ServiceChangeAddress(Addr, State),
			 serviceChangeVersion = enc_opt_serviceChangeVersion(Version, State),
			 serviceChangeProfile = enc_opt_ServiceChangeProfile(Profile, State),
			 serviceChangeReason  = enc_serviceChangeReason(Reason, State),
			 serviceChangeDelay   = enc_opt_serviceChangeDelay(Delay, State),
			 serviceChangeMgcId   = enc_opt_serviceChangeMgcId(MgcId, State),
			 timeStamp            = enc_opt_TimeNotation(Time, State)}.

enc_ServiceChangeMethod(Method, State) ->
    case Method of
        failover      -> failover;
        forced        -> forced;
        graceful      -> graceful;
        restart       -> restart;
        disconnected  -> disconnected;
        handOff       -> handOff
    end. %% BUGBUG: extension

enc_opt_ServiceChangeAddress(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_ServiceChangeAddress({Tag, Val}, State) ->
    Val2 = 
	case Tag of
	    portNumber -> enc_portNumber(Val, State);
	    ip4Address -> enc_IP4Address(Val, State);
	    ip6Address -> enc_IP6Address(Val, State);
	    domainName -> enc_DomainName(Val, State);
	    deviceName -> enc_PathName(Val, State);
	    mtpAddress -> enc_mtpAddress(Val, State)
	end,
    {Tag, Val2}.

enc_opt_serviceChangeVersion(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_serviceChangeVersion(Version, State) ->
    enc_version(Version, State).

enc_opt_ServiceChangeProfile(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_ServiceChangeProfile(#'ServiceChangeProfile'{profileName = Name,
						     version     = Version},
			 State) ->
    ProfName = ?META_ENC(profile, Name),
    #'ServiceChangeProfile'{profileName = enc_Name(ProfName, State),
			    version     = enc_version(Version, State)}.

enc_serviceChangeReason(Reason, State) ->
     enc_Value(Reason, State).

enc_opt_serviceChangeDelay(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_serviceChangeDelay(Delay, State) ->
    enc_UINT32(Delay, State).

enc_opt_serviceChangeMgcId(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_serviceChangeMgcId(MgcId, State) ->
     enc_MId(MgcId, State).

enc_opt_portNumber(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_portNumber(Port, State) ->
    enc_portNumber(Port, State).
     
enc_portNumber(Port, State) when integer(Port), Port >= 0 ->
    enc_UINT16(Port, State).
     
enc_ServiceChangeResParm(#'ServiceChangeResParm'{serviceChangeMgcId   = MgcId, 
						 serviceChangeAddress = Addr, 
						 serviceChangeVersion = Version, 
						 serviceChangeProfile = Profile}, 
			 State) ->
  #'ServiceChangeResParm'{serviceChangeMgcId   = enc_opt_serviceChangeMgcId(MgcId, State),
			  serviceChangeAddress = enc_opt_ServiceChangeAddress(Addr, State),
			  serviceChangeVersion = enc_opt_serviceChangeVersion(Version, State),
			  serviceChangeProfile = enc_opt_ServiceChangeProfile(Profile, State)}.

enc_PackagesDescriptor(Items, State) when list(Items) ->
    [enc_PackagesItem(I, State) || I <- Items].

enc_PackagesItem(#'PackagesItem'{packageName    = Name,
				 packageVersion = Version},
		 State) ->
    PkgdName = ?META_ENC(package, Name),
    #'PackagesItem'{packageName    = enc_Name(PkgdName, State),
		    packageVersion = enc_UINT16(Version, State)}.

enc_StatisticsDescriptor(Parms, State) when list(Parms) ->
    [enc_StatisticsParameter(P, State) || P <- Parms].

enc_StatisticsParameter(#'StatisticsParameter'{statName  = Name,
					       statValue = Value},
			State) ->
    PkgdName = ?META_ENC(statistics, Name),
    #'StatisticsParameter'{statName  = enc_PkgdName(PkgdName, State),
			   statValue = enc_Value(Value, State)}.

enc_opt_TimeNotation(asn1_NOVALUE, State) ->
    asn1_NOVALUE;
enc_opt_TimeNotation(#'TimeNotation'{date = Date,
				     time = Time},
		 State) ->
    #'TimeNotation'{date = enc_STRING(Date, State), % "yyyymmdd"
		    time = enc_STRING(Time, State)}.% "hhmmssss"

%% BUGBUG: Does not verify that string must contain at least one char
%% BUGBUG: This violation of the is required in order to comply with
%% BUGBUG: the dd/ce ds parameter that may possibly be empty.
enc_Value(String, State) when list(String) ->
    String.
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Encode an octet string, escape } by \ if necessary 
enc_OCTET_STRING(String, State, Min, Max) when list(String) ->
    verify_count(length(String), Min, Max),
    String.

enc_QUOTED_STRING(String, State) when list(String) ->
    verify_count(length(String), 1, infinity),
    String.

%% The internal format of hex digits is a list of octets
%% Min and Max means #hexDigits
%% Leading zeros are prepended in order to fulfill Min
enc_HEXDIG(Octets, State, Min, Max) when list(Octets) ->
    verify_count(length(Octets), Min, Max),
    Octets.

enc_DIGIT(Val, State, Min, Max) ->
    enc_integer(Val, State, Min, Max).

enc_STRING(String, State) when list(String) ->
    String.

enc_STRING(String, State, Min, Max) when list(String) ->
    verify_count(length(String), Min, Max),
    String.

enc_UINT16(Val, State) ->
    enc_integer(Val, State, 0, 65535).

enc_UINT32(Val, State) ->
    enc_integer(Val, State, 0, 4294967295).

enc_integer(Int, State, Min, Max) ->
    verify_count(Int, Min, Max),
    Int.

%% Verify that Count is within the range of Min and Max
verify_count(Count, Min, Max) ->
    if
	integer(Count) ->
	    if
		integer(Min), Count >= Min ->
		    if
			integer(Max), Count =< Max ->
			    Count;
			Max == infinity ->
			    Count;
			true ->
			    exit({count_too_large, Count, Max})
		    end;
		true ->
		    exit({count_too_small, Count, Min})
	    end;
	true ->
	    exit({count_not_an_integer, Count})
    end.
