%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%%----------------------------------------------------------------------
%% Purpose: Lightweight test server
%%----------------------------------------------------------------------

-module(test_lib).

-compile(export_all).

-include("test_lib.hrl").

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Evaluates a test case or test suite
%% Returns a list of failing test cases:
%% 
%%     {Mod, Fun, ExpectedRes, ActualRes}
%%----------------------------------------------------------------------

t(Case) ->
    t(Case, default_config()).

t({Mod, Fun}, Config) when atom(Mod), atom(Fun) ->
    case catch apply(Mod, Fun, [suite]) of
	[] ->
	    io:format("Eval:   ~p~n", [{Mod, Fun}]),
	    eval(Mod, Fun, Config);
	Cases when list(Cases) ->
	    io:format("Expand: ~p ...~n", [{Mod, Fun}]),
	    Map = fun(Case) when atom(Case)-> {Mod, Case};
		     (Case) -> Case
		  end,
	    t(lists:map(Map, Cases), Config);
	_ ->
	    io:format("Eval:   ~p ???~n", [{Mod, Fun}]),
	    [{Mod, Fun, ok, nyi}]
    end;
t(Mod, Config) when atom(Mod) ->
    t({Mod, all}, Config);
t(Cases, Config) when list(Cases) ->
    Errors = [t(Case, Config) || Case <- Cases],
    lists:append(Errors);
t(Bad, Config) ->
    [{badarg, Bad, ok, nyi}].

eval(Mod, Fun, Config) ->
    TestCase = {?MODULE, Mod, Fun},
    Label = lists:concat(["TEST CASE: ", Fun]),
    event_tracer:report(40, ?MODULE, Mod, Label ++ " started",
			[TestCase, Config]),
    Pid = spawn_link(?MODULE, do_eval, [self(), Mod, Fun, Config]),
    KillAfter = get_kill_timer(Config),
    receive
	{done, Pid, Errors} ->
	    if
		Errors == [] ->
		    event_tracer:report(40, Mod, ?MODULE, Label ++ " ok",
					[TestCase, Config, Errors]);
		true ->
		    event_tracer:report(20, Mod, ?MODULE, Label ++ " failed",
					[TestCase, Config, Errors])
	    end,
	    Errors;
	{'EXIT', Pid, Reason} -> 
	    event_tracer:report(20, Mod, ?MODULE, Label ++ " crashed",
				[TestCase, Config, {'EXIT', Reason}]),
	    [Mod, Fun, ok, {'EXIT', Reason}]
    after KillAfter ->
	    Error = {?MODULE, eval, {kill_after, KillAfter}},
	    event_tracer:report(20, Mod, ?MODULE, Label ++ " timeout",
				[TestCase, Config, Error]),
    
	    Errors = steal_test_lib_errors(Pid),
	    unlink(Pid),
	    exit(Pid, KillAfter),
	    lists:reverse([Error | Errors])
    end.

steal_test_lib_errors(Pid) ->
    case process_info(Pid, dictionary) of
	{dictionary, Dict} ->
	    case lists:keysearch(test_lib_errors, 1, Dict) of
		{value, {test_lib_errors, Errors}} ->
		    Errors;
		false ->
		    []
	    end;
	undefined ->
	    []
    end.

do_eval(ReplyTo, Mod, Fun, Config) ->
    put(test_lib_errors, []),
    Res = (catch apply(Mod, Fun, [Config])),
    Errors = lists:reverse(get(test_lib_errors)),
    case Res == ok of	    
	true -> 
	    ReplyTo ! {done, self(), Errors};
	false ->
	    BadRes = {Mod, Fun, ok, Res},
	    ReplyTo ! {done, self(), Errors ++ [BadRes]}
    end,
    unlink(ReplyTo),
    exit(test_case_eval_done).
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Verify that the actual result of a test case matches the exected one
%% Returns the actual result
%% Stores the result in the process dictionary if mismatch

error(Actual, Mod, Line) ->
    log("<ERROR> Bad result: ~p~n", [Actual], Mod, Line),
    Label = lists:concat([Mod, "(", Line, ") unexpected result"]),
    event_tracer:report(60, Mod, Mod, Label,
			[{line, Mod, Line}, {error, Actual}]),
    Error = {Mod, Line, Actual},
    put(test_lib_errors, [Error | get(test_lib_errors)]),
    Actual.

log(Format, Args, Mod, Line) ->
    ok = io:format("~p(~p): " ++ Format, [Mod, Line] ++ Args).

skip(Actual, File, Line) ->
    error({skipped, Actual}, File, Line),
    Format = "Skipping test case ~p(~p):",
    String = lists:flatten(io_lib:format(Format, [File, Line])),
    exit({skipped, String}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Flush the message queue and return its messages
flush() ->
    receive
	Msg ->
	    [Msg | flush()]
    after 1000 ->
	    []
    end.
	    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Check if process is alive and kicking
still_alive(Pid) ->   
    case catch erlang:is_process_alive(Pid) of % New BIF in Erlang/OTP R5
	true -> 
	    true;
	false -> 
	    false;
	{'EXIT', _} -> % Pre R5 backward compatibility 
	    case process_info(Pid, message_queue_len) of
		undefined -> false;
		_ -> true
	    end 
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The proxy process

proxy_start(ProxyId) ->
    spawn_link(?MODULE, proxy_init, [ProxyId, self()]).

proxy_start(Node, ProxyId) ->
    spawn_link(Node, ?MODULE, proxy_init, [ProxyId, self()]).

proxy_init(ProxyId, Controller) ->
    process_flag(trap_exit, true),
    proxy_loop(ProxyId, Controller).

proxy_loop(OwnId, Controller) ->
    receive
	{'EXIT', Controller, Reason} ->
	    exit(Reason);
	{apply, Fun} ->
	    Controller ! {res, OwnId, Fun()},
	    proxy_loop(OwnId, Controller);
	OtherMsg ->
	    Controller ! {msg, OwnId, OtherMsg},
	    proxy_loop(OwnId, Controller)
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Test server callbacks
init_per_testcase(Case, Config) ->
    Config.

fin_per_testcase(Case, Config) ->
    ok.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Set kill timer

get_kill_timer(Config) ->
    case init:get_argument(test_lib_timeout) of
	{ok, _ } ->
	    infinity;
	_ ->
	    case lists:keysearch(tc_timeout, 1, Config) of
		false ->
		    timer:minutes(5);
		{Value, {tc_timeout, KillAfter}} ->
		    KillAfter
	    end
    end.
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

prepare_test_case(Actions, N, Config, File, Line) ->
    OrigNodes = lookup_config(nodes, Config),
    TestNodes = lookup_config(nodenames, Config), %% For testserver
    This = node(),
    SomeNodes = OrigNodes ++ (TestNodes -- OrigNodes),
    AllNodes = [This | (SomeNodes -- [This])],
    Nodes = pick_n_nodes(N, AllNodes, File, Line),
    start_nodes(Nodes, File, Line),
    do_prepare_test_case(Actions, Nodes, Config, File, Line).

do_prepare_test_case([init | Actions], Nodes, Config, File, Line) ->
    process_flag(trap_exit, true),
    test_lib:flush(),
    do_prepare_test_case(Actions, Nodes, Config, File, Line);
do_prepare_test_case([{stop_app, App} | Actions], Nodes, Config, File, Line) ->
    Res = rpc:multicall(Nodes, application, stop, [App]),
    do_prepare_test_case(Actions, Nodes, Config, File, Line);
do_prepare_test_case([], Nodes, Config, File, Line) ->
    Nodes.

pick_n_nodes(all, AllNodes, File, Line) ->
    AllNodes;
pick_n_nodes(N, AllNodes, File, Line) when integer(N), length(AllNodes) >= N ->
    AllNodes -- lists:nthtail(N, AllNodes);
pick_n_nodes(N, AllNodes, File, Line) ->
    skip({too_few_nodes, N, AllNodes}, File, Line).
   
lookup_config(Key,Config) ->
    case lists:keysearch(Key, 1, Config) of
	{value,{Key,Val}} ->
	    Val;
	_ ->
	    []
    end.

default_config() ->
    [{nodes, default_nodes()}].

default_nodes() ->    
    mk_nodes(2, []).

mk_nodes(0, Nodes) ->
    Nodes;
mk_nodes(N, []) ->
    mk_nodes(N - 1, [node()]);
mk_nodes(N, Nodes) when N > 0 ->
    Head = hd(Nodes),
    [Name, Host] = node_to_name_and_host(Head),
    Nodes ++ [mk_node(I, Name, Host) || I <- lists:seq(1, N)].

mk_node(N, Name, Host) ->
    list_to_atom(lists:concat([Name ++ integer_to_list(N) ++ "@" ++ Host])).
    
%% Returns [Name, Host]    
node_to_name_and_host(Node) ->
    string:tokens(atom_to_list(Node), [$@]).

start_nodes([Node | Nodes], File, Line) ->
    case net_adm:ping(Node) of
	pong ->
	    start_nodes(Nodes, File, Line);
	pang ->
	    [Name, Host] = node_to_name_and_host(Node),
	    Args = [],
	    case slave:start_link(Host, Name) of
		{ok, NewNode} when NewNode == Node ->
		    Path = code:get_path(),
		    {ok, Cwd} = file:get_cwd(),
		    true = rpc:call(Node, code, set_path, [Path]),
		    ok = rpc:call(Node, file, set_cwd, [Cwd]),
		    true = rpc:call(Node, code, set_path, [Path]),
		    {_, []} = rpc:multicall(global, sync, []),
		    start_nodes(Nodes, File, Line);
		Other ->
		    skip({cannot_start_node, Node, Other}, File, Line)
	    end
    end;
start_nodes([], File, Line) ->
    ok.

