%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(coast).

%% External exports
-export([start/0, stop/0,
	 compile/1, compile/2, compile_all/0, compile_all/1, compile_all/2,
	 run/3,
	 mod_coverage/1, func_coverage/1, clause_coverage/1,
	 mod_calls/1, func_calls/1, clause_calls/1,
	 analyse_to_file/1,
	 known_modules/0,
	 source_file/1,
	 clear/1, clear_all/0]).
-export([bump/6]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
	 code_change/3]).

%% Internal exports
-export([code_runner/4]).

-include_lib("kernel/include/file.hrl").

-record(state, {modules=[]}).

-record(bump, {module   = '_',                  % atom()
	       function = '_',                  % atom()
	       arity    = '_',                  % integer()
	       clause   = '_',                  % integer()
	       index    = '_',                  % integer()
	       line     = '_'}).                % integer()

-record(vars, {module,                          % atom() Module name
	       file,                            % string() File name
	       obj_dir,                         % string()
	       
	       pretty_mode,                     % pretty | not_pretty
	       index,                           % int()
	       init_info,                       % [{M,F,A,C,L}]

	       function,                        % atom()
	       arity,                           % int()
	       clause,                          % int()
	       line,                            % int()
	       lines}).                         % int()

-define(TABLE, 'coast_internal_data_table').

%%%----------------------------------------------------------------------
%%% External exports
%%%----------------------------------------------------------------------

start() ->
    gen_server:start({local, coast_server}, ?MODULE, [], []).

stop() ->
    gen_server:cast(coast_server, stop).

compile(File) ->
    compile(File, []).
compile(ModFile, Options) ->
    File1 = if
		atom(ModFile) -> atom_to_list(ModFile);
		true -> ModFile
	    end,
    File2 = case filename:extension(File1) of
		".erl" -> File1;
		_ -> File1++".erl"
	    end,
    call({compile, File2, Options}).

compile_all() ->
    case file:get_cwd() of
	{ok, Dir} ->
	    compile_all(Dir, []);
	Error ->
	    Error
    end.
compile_all(Dir) ->
    compile_all(Dir, []).
compile_all(Dir, Options) ->
    case file:list_dir(Dir) of
	{ok, Files} ->
	    Pred = fun("coast.erl") -> false;
		      ("coast_server.erl") -> false;
		      ("coast_analyse_funcs.erl") -> false;
		      ("coast_compile_funcs.erl") -> false;
		      (File) ->
			   case filename:extension(File) of
			       ".erl" ->
				   case string:str(File, "COAST") of
				       0 -> true;
				       _ -> false
				   end;
			       _ -> false
			   end
		   end,
	    Files2 = lists:filter(Pred, Files),
	    Files3 = lists:map(fun(File) ->
				       filename:join(Dir, File)
			       end,
			       Files2),
	    lists:map(fun(File) ->
			      compile(File, Options)
		      end,
		      Files3);
	Error ->
	    Error
    end.

run(Mod, Func, Args) ->
    case call({run, [Mod, Func, Args]}) of
	%% If apply(Mod,Func,Args) fails, this function should fail as well
	{'EXIT', Reason} ->
	    exit(Reason);
	Result ->
	    Result
    end.

mod_coverage(Module) ->
    call({mod_coverage, Module}).

func_coverage(Module) ->
    call({func_coverage, Module}).

clause_coverage(Module) ->
    call({clause_coverage, Module}).

mod_calls(Module) ->
    call({mod_calls, Module}).

func_calls(Module) ->
    call({func_calls, Module}).

clause_calls(Module) ->
    call({clause_calls, Module}).

analyse_to_file(Module) ->
    call({analyse_to_file, Module}).

known_modules() ->
   call(known_modules).

source_file(Module) ->
    call({source_file, Module}).

clear(Module) ->
    call({clear, Module}).

clear_all() ->
    call(clear_all).

bump(Module, Function, Arity, Clause, Index, Line) ->
    Key = #bump{module=Module, function=Function, arity=Arity, clause=Clause,
		index=Index, line=Line},

    case catch ets:update_counter(?TABLE, Key, 1) of
	N when integer(N) ->
	    done;
	{'EXIT', _Reason1} ->
	    case catch ets:insert(?TABLE, {Key, 1}) of
		true ->
		    done;
		{'EXIT', _Reason2} ->
		    exit({coast_compiled, Module})
	    end
    end.

call(Request) ->
    case catch gen_server:call(coast_server, Request, infinity) of
	{'EXIT', {noproc, _Call}} ->
	    start(),
	    catch gen_server:call(coast_server, Request, infinity);
	Reply ->
	    Reply
    end.


%%%----------------------------------------------------------------------
%%% gen_server callbacks
%%%----------------------------------------------------------------------

init(_Args) ->
    ets:new(?TABLE, [set, public, named_table]),
    {ok, #state{}}.

handle_call({compile, File, Options}, _From, State) ->
    case compile_internal(File, Options) of
	{ok, Module} ->
	    clear_db(Module),
	    Modules = add_module(Module, State#state.modules),
	    {reply, {ok, Module}, State#state{modules=Modules}};
	Error ->
	    {reply, Error, State}
    end;

handle_call({run, [Mod, Func, Args]}, From, State) ->
    spawn(?MODULE, code_runner, [From, Mod, Func, Args]),
    {noreply, State}; % reply sent from code_runner/4

handle_call(known_modules, _From, State) ->
    Reply = State#state.modules,
    {reply, Reply, State};

handle_call({source_file, Module}, _From, State) ->
    case beam_lib:chunks(Module, [attributes]) of
	{ok, {Module, [{attributes, Attributes}]}} ->
	    case lists:keysearch(coast_src_file, 1, Attributes) of
		false ->
		    {reply, {error, {not_coast_compiled,Module}}, State};
		{value, {coast_src_file,File}} ->
		    {reply, {ok,File}, State}
	    end;
	Error ->
	    {reply, {error, {not_coast_compiled,Module}}, State}
    end;

handle_call({clear, Module}, _From, State) ->
    clear_db(Module),
    {reply, ok, State};
handle_call(clear_all, _From, State) ->
    ets:delete(?TABLE),
    ets:new(?TABLE, [set, public, named_table]),
    {reply, ok, State};

handle_call({Analyse, Module}, _From, State) ->
    case catch Module:'coast_init_info_$$_private'() of
	{'EXIT', _Reason} ->
	    {reply, {error, {not_coast_compiled,Module}}, State};
	Clauses ->
	    Reply = case Analyse of
			mod_coverage -> mod_coverage(Module, Clauses);
			func_coverage -> func_coverage(Module, Clauses);
			clause_coverage -> clause_coverage(Module, Clauses);
			mod_calls -> mod_calls(Module, Clauses);
			func_calls -> func_calls(Module, Clauses);
			clause_calls -> clause_calls(Module, Clauses);
			analyse_to_file -> analyse_to_file(Module, Clauses)
		    end,
	    {reply, Reply, State}
    end.

handle_cast(stop, State) ->
    {stop, normal, State}.

handle_info(null, State) ->
    {noreply, State}.

terminate(Reason, State) ->
    ok.

code_change(OldVsn, State, Extra) ->
    {ok, State}.


%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

code_runner(From, Mod, Func, Args) ->
    Result = (catch apply(Mod, Func, Args)),
    gen_server:reply(From, Result).


%%%--Auxiliary-----------------------------------------------------------

clear_db(Module) ->
    ets:match_delete(?TABLE, {#bump{module=Module},'_'}).

add_module(Module, Modules) ->
    case lists:member(Module, Modules) of
	true ->
	    Modules;
	false ->
	    Modules ++ [Module]
    end.

%%%--Compilation---------------------------------------------------------

%% compile_internal(File, Options) -> {ok, Module} | {error, Error}
%%   File = string()  (including directory path and .erl extension)
%%   Options =
%%   Module = atom()  (module name only)
%%   Error = {What, File, Reason} | {What, File}
%%     What = file | directory | transform | compile
%%     Reason = term()
compile_internal(File, Options) ->
    
    %% Check that the file exists, is a regular file and is accessible
    case check_file(File) of
	ok ->
	    ModuleS = filename:basename(File, ".erl"),
	    Module = list_to_atom(ModuleS),

	    %% Sort out the options to get object file directory,
	    %% include directories and defines.
	    {ok, CWD} = file:get_cwd(),
	    SrcDir = filename:dirname(File),
	    Default = {CWD, [SrcDir], []},
	    {ObjDir, IncDirs, Defs} = sort_out_options(Options, Default),
	    
	    Vars = #vars{module = Module, file = File, obj_dir = ObjDir},

	    CoastDir = filename:join(ObjDir, "COAST"),
	    CoastSrcDir = filename:join(ObjDir, "COAST/tmp_code"),
	    PrettyFile = filename:join(CoastSrcDir,
				       ModuleS++".COAST.pretty.erl"),
	    SrcFile = filename:join(CoastSrcDir, ModuleS++".COAST.erl"),
	    ResObjFile = filename:join(ObjDir, ModuleS++".COAST.beam"),
	    ObjMod = filename:join(ObjDir, ModuleS),
	    ObjFile = ObjMod++".beam",

	    %% Create coast directories in the output directory
	    case make_dirs(CoastDir, CoastSrcDir) of
		ok ->
		    case make_srcfile(File, PrettyFile, SrcFile, IncDirs,
				      Defs, Vars) of
			ok ->
			    Options2 = [report_errors,
					report_warnings | Options],
			    case compile:file(SrcFile, Options2) of
				{ok, Module} ->
				    file:rename(ResObjFile, ObjFile),
				    code:purge(Module),
				    code:load_abs(ObjMod),
				    {ok, Module};
				error ->
				    {error, {compile, File}}
			    end;
			{error, Reason} ->
			    {error, {transform, File, Reason}}
		    end;
		{error, Reason} ->
		    {error, {directory, File, Reason}}
	    end;
	{error, Reason} ->
	    {error, {file, File, Reason}}
    end.

%% check_file(File) -> ok | {error,Reason}
%%   File = string()
%%   Reason = term()
%% Returns ok if File is a readable file.
check_file(File) ->
    case file:read_file_info(File) of
	{ok, FileInfo} when FileInfo#file_info.type == regular,
			    FileInfo#file_info.access /= none ->
	    ok;
	{ok, FileInfo} when FileInfo#file_info.type /= regular ->
	    {error, {bad_file_type, FileInfo#file_info.type}};
	{ok, FileInfo} when FileInfo#file_info.access == none ->
	    {error, no_access};
	Error ->
	    Error
    end.

%% sort_out_options(Options, Default) -> Actual
%%   Options = [Option]
%%     Option = <compiler option>, see compile:file/2
%%   Default = Actual = {ObjDir, IncDirs, Defs}
%%     ObjDir = string()
%%     IncDirs = [string()]
%%     Defs = [term()]
%% Takes a list of compiler options and return a tuple with:
%% * object/output directory as specified by the option key 'outdir',
%%   replacing the default value.
%% * a list of directories to search for include files as specified by
%%   the option key 'i', appended to the default value.
%% * a list of terms to define as specified by the option key 'd', appended
%%   to the default value.
sort_out_options([{outdir,ObjDir} | Options], {_ObjDir, IncDirs, Defs}) ->
    sort_out_options(Options, {ObjDir, IncDirs, Defs});
sort_out_options([{i,Dir} | Options], {ObjDir, IncDirs, Defs}) ->
    sort_out_options(Options, {ObjDir, IncDirs++[Dir], Defs});
sort_out_options([{d,Macro} | Options], {ObjDir, IncDirs, Defs}) ->
    sort_out_options(Options, {ObjDir, IncDirs, Defs++[{Macro,true}]});
sort_out_options([{d,Macro,Value} | Options], {ObjDir, IncDirs, Defs}) ->
    sort_out_options(Options, {ObjDir, IncDirs, Defs++[{Macro,Value}]});
sort_out_options([Option | Options], Result) ->
    sort_out_options(Options, Result);
sort_out_options([], Result) ->
    Result.

%% make_dirs(CoastDir, CoastSrcDir) -> ok | {error,Reason}
%%   CoastDir = CoastSrcDir = string()
%%   Reason = term()
%% Create the coast directories.
make_dirs(CoastDir, CoastSrcDir) ->
    case file:make_dir(CoastDir) of
	ok ->
	    file:make_dir(CoastSrcDir);
	{error, eexist} ->
	    ok;
	Error ->
	    Error
    end.

%% make_srcfile(File, PrettyFile, SrcFile, IncDirs, Defs, Vars)
%%                                                    -> ok | {error, Reason}
%%   File = PrettyFile = SrcFile = string()  (incl dir path and .erl ext)
%%   IncDirs = [string()]
%%   Defs = [{atom(),term()}]
%%   Vars = #vars{}
%%   Reason = term()
%% Create a source file.
make_srcfile(File, PrettyFile, SrcFile, IncDirs, Defs, Vars) ->
    case transform(pretty, File, PrettyFile, IncDirs, Defs, Vars) of
	ok ->
	    transform(not_pretty, PrettyFile, SrcFile, [], [], Vars);
	Error ->
	    Error
    end.

%% transform(Mode, InFile, OutFile, IncDirs, Defs, Vars)
%%                                                    -> ok | {error, Reason}
%%   Mode = pretty | not_pretty
%%   InFile = OutFile = string()  (including directory path and extension)
%%   IncDirs = [string()]
%%   Defs = [{atom(),term()}]
%%   Vars = #vars{}
%%   Reason = term()
transform(Mode, InFile, OutFile, IncDirs, Defs, Vars) ->

    %% Open files.
    %% Note: No error check, at this stage it should already have been made
    %% certain that the files are readable/writable!
    {ok, InFd} = epp:open(InFile, IncDirs, Defs),
    {ok, OutFd} = file:open(OutFile, write),

    Vars1 = Vars#vars{pretty_mode = Mode, index = 0, init_info = []},
    Result = translate(Mode, InFd, OutFd, Vars1),
    epp:close(InFd),
    case Result of
	{transformed, Vars2} ->
	    windup(Mode, OutFd, Vars2),
	    file:close(OutFd),
	    ok;
	Error ->
	    file:close(OutFd),
	    Error
    end.

%% translate(Mode, InFd, OutFd, Vars) -> {transformed,Vars2} | {error,Reason}
%%   Mode = pretty | not_pretty
%%   InFd = OutFd = file descriptor (pid())
%%   Vars = Vars2 = #vars{}
%%   Reason = term()
translate(Mode, InFd, OutFd, Vars) ->
    AbsForm = epp:parse_erl_form(InFd),
    translation_loop(Mode, AbsForm, InFd, OutFd, Vars).

translation_loop(pretty,{ok,{attribute,Line,file,Mod}},InFd,OutFd,Vars) ->
    translate(pretty, InFd, OutFd, Vars);
translation_loop(pretty, {ok,{attribute,Line,module,Mod}},InFd,OutFd,Vars) ->
    io:format(OutFd, "~s", [erl_pp:form({attribute,Line,module,Mod})]),
    translate(pretty, InFd, OutFd, Vars);
translation_loop(pretty,{ok,{attribute,Line,AName,AData}},InFd,OutFd,Vars) ->
    io:format(OutFd, "~s", [erl_pp:form({attribute,Line,AName,AData})]),
    translate(pretty, InFd, OutFd, Vars);
translation_loop(pretty,{ok,Form},InFd,OutFd,Vars) ->
    {MungedForm, NewVars} = munge(Form, Vars),
    io:format(OutFd, "~n~s", [erl_pp:form(MungedForm)]),
    translate(pretty, InFd, OutFd, NewVars);

translation_loop(not_pretty,
		 {ok,{attribute,Line,module,Mod}},InFd,OutFd,Vars) ->
    io:format(OutFd, "~s", [erl_pp:form({attribute,Line,module,Mod})]),
    io:format(OutFd, "~s", [erl_pp:form({attribute,Line,coast_src_file,
					 Vars#vars.file})]),
    io:format(OutFd, "~s", [erl_pp:form({attribute,Line,coast_obj_dir,
					 Vars#vars.obj_dir})]),
    io:format(OutFd, "~s~n",
	      [erl_pp:form({attribute,Line,export,
			    [{'coast_init_info_$$_private', 0}]})]),
    translate(not_pretty, InFd, OutFd, Vars);
translation_loop(not_pretty,{ok,Form},InFd,OutFd,Vars) ->
    {MungedForm, NewVars} = munge(Form, Vars),
    io:format(OutFd, "~s", [erl_pp:form(MungedForm)]),
    translate(not_pretty, InFd, OutFd, NewVars);

translation_loop(_Mode,{eof,Line},InFd,OutFd,Vars) ->
    {transformed, Vars};
translation_loop(_Mode, {error,{Line,Mod,Error}},InFd,OutFd,Vars) ->
    {error, {parse_error,Line,Error}}.

%% munge(AbsForm, Vars) -> AbsForm2
%%   AbsForm = AbsForm2 = term()
%%   Vars = #vars{}
munge({attribute,Line,module,Mod}, Vars) ->
    {{attribute,Line,module,Mod}, Vars};
munge({function,Line,Function,Arity,Clauses}, Vars) ->
    Vars1 = Vars#vars{function=Function, arity=Arity, clause=1, line=Line,
		      lines=0},
    {MungeRes, Vars2} = munge_clauses(Clauses, Vars1),
    {{function,Line,Function,Arity,MungeRes}, Vars2};
munge(X, Vars) ->
    {X, Vars}.

munge_clauses(ClauseForm, Vars) ->
    munge_clauses(ClauseForm, Vars, []).

munge_clauses([{clause,Line,Head,Guard,Body} | T], Vars, MungeAcc) ->
    {BodyRes, NewVars} = 
	munge_body(Body, Vars#vars{line=Line}),
    InitInfo = [{NewVars#vars.module, NewVars#vars.function,
		 NewVars#vars.arity, NewVars#vars.clause,
		 NewVars#vars.lines} | NewVars#vars.init_info],
    munge_clauses(T,
		  NewVars#vars{init_info=InitInfo,
			       clause=(NewVars#vars.clause)+1, lines=0},
		  [{clause,Line,Head,Guard,BodyRes} | MungeAcc]);
munge_clauses([], Vars, MungeAcc) -> 
    {lists:reverse(MungeAcc), Vars}.

munge_body([H | T], Vars) ->
    {Bump, NewVars1} = 
	case Vars#vars.pretty_mode of
	    pretty ->
		{[], Vars};
	    not_pretty ->
		add_bump(Vars)
	end,
    {ExprRes, NewVars2} = munge_expr(H, NewVars1),
    {BodyRes, NewVars3} = munge_body(T, NewVars2),
    {Bump ++ [ExprRes | BodyRes], NewVars3};
munge_body([], Vars) ->
    {[], Vars}.

% The following expressions are of interest:
%
%  {'if', Line, Opts}
%  {'case', Line, Func, Opts}
%  {'receive', Line, Opts, Func, TimeOutBody}
%  {'receive', Line, Opts}
%  {'catch', Line, E0}
%  {match,Line,LS,RS}
%  {call,Line,Label,Args}
%  {'fun', Line, Body}
%  {tuple, Line, Es0}
%  {op, Line, Op, A0}
%  {op, Line, Op, L0, R0}
%  {remote, Line, M0, F0}
%  {var, Line, VariabelName}
%  {TermName, Line, Value}, TermName = integer | float | atom | string
%  {nil, Line}
%
% The following expressions are probably not of interest:
%
%  {record_index,Line,Name,Field0}
%  {record,Line,Name,Inits0}
%  {record_field,Line,Rec0,Name,Field0}
%  {record,Line,Rec0,Name,Upds0}
%  {record_field,Line,Rec0,Field0}
%  {block,Line,Es0}
%
% (For further information about the forms that exist, 
% see 'erl_id_trans.erl' in the stdlib!)

munge_expr({integer,Line,Value}, Vars) ->
    {{integer,Line,Value}, Vars#vars{line=Line}};
munge_expr({float, Line, Value}, Vars) ->
    {{float,Line,Value}, Vars#vars{line=Line}};
munge_expr({atom,Line,Value}, Vars) ->
    {{atom,Line,Value}, Vars#vars{line=Line}};
munge_expr({string,Line,Value}, Vars) ->
    {{string,Line,Value}, Vars#vars{line=Line}};
munge_expr({tuple,Line,Es0}, Vars) ->
    {Es1, NewVars} = munge_tuple_expr_list(Es0, Vars#vars{line=Line}),
    {{tuple,Line,Es1}, NewVars};
munge_expr({var,Line,Value}, Vars) ->
    {{var,Line,Value}, Vars#vars{line=Line}};
munge_expr({op,Line,Op,A0}, Vars) ->
    {A1, NewVars} = munge_expr(A0, Vars#vars{line=Line}),
    {{op,Line,Op,A1}, NewVars};
munge_expr({op,Line,Op,L0,R0}, Vars) ->
    {L1, NewVars1} = munge_expr(L0, Vars#vars{line=Line}),
    {R1, NewVars2} = munge_expr(R0, NewVars1),
    {{op,Line,Op,L1,R1}, NewVars2};
munge_expr({match,Line,LS,RS}, Vars) ->
    {ExprRes1, NewVars1} = munge_expr(LS, Vars#vars{line=Line}),
    {ExprRes2, NewVars2} = munge_expr(RS, NewVars1),
    {{match,Line,ExprRes1,ExprRes2}, NewVars2};
munge_expr({call,Line,Label,Args}, Vars) ->
    {ExprRes, NewVars} = munge_call_expr_list(Args, Vars#vars{line=Line}),
    {{call,Line,Label,ExprRes}, NewVars};
munge_expr({'case',Line,Func,Opts}, Vars) -> 
    {OptionRes, NewVars} = munge_option_list(Opts, Vars#vars{line=Line}),
    %% Since the 'end' word isn't shown in the parsing, 
    %% we have to manually count that line.
    NewLine = NewVars#vars.line + 1,
    {{'case',Line,Func,OptionRes}, NewVars#vars{line=NewLine}};
munge_expr({'if',Line,Opts}, Vars) -> 
    {OptionRes, NewVars} = munge_option_list(Opts, Vars#vars{line=Line}),
    %% Since the 'end' word isn't shown in the parsing, 
    %% we have to manually count that line.
    NewLine = NewVars#vars.line + 1,
    {{'if',Line,OptionRes}, NewVars#vars{line=NewLine}};
munge_expr({'receive',Line,Opts}, Vars) -> 
    {OptionRes, NewVars} = munge_option_list(Opts, Vars#vars{line=Line}),
    %% Since the 'end' word isn't shown in the parsing, 
    %% we have to manually count that line.
    NewLine = NewVars#vars.line + 1,
    {{'receive',Line,OptionRes}, NewVars#vars{line=NewLine}};
munge_expr({'receive',Line,Opts,Func,TimeOutBody}, Vars) ->
    {OptionRes, NewVars1} = munge_option_list(Opts, Vars#vars{line=Line}),
    {FuncRes, NewVars2}   = munge_expr(Func, NewVars1),
    {BodyRes, NewVars3}   = munge_body(TimeOutBody, NewVars2),
    %% Since the 'end' word isn't shown in the parsing, 
    %% we have to manually count that line.
    NewLine = NewVars3#vars.line + 1,
    {{'receive',Line,OptionRes,Func,BodyRes},NewVars3#vars{line=NewLine}};
munge_expr({'catch',Line,E0}, Vars) ->
    {E1, NewVars} = munge_expr(E0, Vars#vars{line=Line}),
    {{'catch',Line,E1}, NewVars};
munge_expr({'fun',Line,{clauses,Cs0}}, Vars) ->
    {Cs1, NewVars} = munge_fun_clauses(Cs0, Vars#vars{line=Line}),
    {{'fun',Line,{clauses,Cs1}}, NewVars};
munge_expr({'fun',Line,FunctionForm}, Vars) ->
    {{'fun',Line,FunctionForm}, Vars#vars{line=Line}};
munge_expr({record,Line,Name,Inits}, Vars) ->
    {{record,Line,Name,Inits}, Vars#vars{line=Line}};
munge_expr({record,Line,Rec,Name,Upds}, Vars) ->
    {{record,Line,Rec,Name,Upds}, Vars#vars{line=Line}};
munge_expr({record_field,Line,Rec,Name,Field}, Vars) ->
    {{record_field,Line,Rec,Name,Field}, Vars#vars{line=Line}};
munge_expr({record_index,Line,Name,Field}, Vars) ->
    {{record_index,Line,Name,Field}, Vars#vars{line=Line}};
munge_expr({block,Line,Es}, Vars) ->
    {{block,Line,Es}, Vars#vars{line=Line}};
munge_expr({remote,Line,M0,F0}, Vars) ->
    {M1, NewVars1} = munge_expr(M0, Vars#vars{line=Line}),
    {F1, NewVars2} = munge_expr(F0, NewVars1),
    {{remote,Line,M1,F1}, NewVars2};
munge_expr({cons,Line,H0,T0}, Vars) ->
    {H1, NewVars1} = munge_expr(H0, Vars#vars{line=Line}),
    {T1, NewVars2} = munge_expr(T0, NewVars1),
    {{cons,Line,H1,T1}, NewVars2};
munge_expr({lc,Line,E0,Qs0}, Vars) ->
    {E1, NewVars1}  = munge_expr(E0, Vars#vars{line=Line}),
    {Qs1, NewVars2} = munge_lc_quals(Qs0, NewVars1),
    {{lc,Line,E1,Qs1}, NewVars2};
munge_expr({nil,Line}, Vars) ->
    {{nil, Line}, Vars#vars{line=Line}};
munge_expr({TermName,Line,Value}, Vars) ->  
    %% If we have missed some term names above...
    {{TermName,Line,Value}, Vars#vars{line=Line}};
munge_expr(X, Vars) ->       
    {X, Vars}.

munge_tuple_expr_list(ExprList, Vars) ->
    munge_tuple_expr_list(ExprList, Vars, []).

munge_tuple_expr_list([E0 | Es], Vars, Acc) ->
    {E1, NewVars} = munge_expr(E0, Vars),
    munge_tuple_expr_list(Es, NewVars, [E1 | Acc]);
munge_tuple_expr_list([], Vars, Acc) ->
    {lists:reverse(Acc), Vars}.

munge_call_expr_list(ExprList, Vars) ->
    munge_call_expr_list(ExprList, Vars, []).

munge_call_expr_list([E0 | Es], Vars, Acc) ->
    {E1, NewVars} = munge_expr(E0, Vars),
    munge_call_expr_list(Es, NewVars, [E1 | Acc]);
munge_call_expr_list([], Vars, Acc) ->
    {lists:reverse(Acc), Vars}.

munge_option_list(OptionList, Vars) ->
    munge_option_list(OptionList, Vars, []).

munge_option_list([{clause,Line,Guard,Body} | T], Vars, Acc) ->
    {BodyRes, NewVars} = munge_body(Body, Vars#vars{line=Line}),
    munge_option_list(T, NewVars, [{clause,Line,Guard,BodyRes} | Acc]);
munge_option_list([{clause,Line,Pattern,Guard,Body} | T], Vars, Acc) ->
    {BodyRes, NewVars} = munge_body(Body, Vars#vars{line=Line}),
    munge_option_list(T, NewVars,
		      [{clause, Line, Pattern, Guard, BodyRes} | Acc]);
munge_option_list([], Vars, Acc) -> 
    {lists:reverse(Acc), Vars};
munge_option_list(X, Vars, Acc) ->
    throw({error, {unknown_option, X}}).

munge_lc_quals(LcQualsList, Vars) ->
    munge_lc_quals(LcQualsList, Vars, []).

munge_lc_quals([{generate,Line,P0,E0} | Qs], Vars, Acc) ->
    {P1, NewVars1} = munge_expr(P0, Vars#vars{line=Line}),
    {E1, NewVars2} = munge_expr(E0, NewVars1),
    munge_lc_quals(Qs, NewVars2, [{generate,Line,P1,E1} | Acc]);
munge_lc_quals([E0 | Qs], Vars, Acc) ->
    {E1, NewVars} = munge_expr(E0, Vars),
    munge_lc_quals(Qs, NewVars, [E1 | Acc]);
munge_lc_quals([], Vars, Acc) ->
       %% Since the closing bracket isn't shown in the parsing, 
       %% we have to add that line manually!
    NewLine = Vars#vars.line + 1,
    {lists:reverse(Acc), Vars#vars{line=NewLine}}.

% {clause, LineNo, ListOfArgForms, ListOfGuardForms, ListOfBodyForms}
% This function is essentially the same as the function clause function, 
% the only difference being we don't want any clause numbering here!

munge_fun_clauses(FunClauseList, Vars) ->
    munge_fun_clauses(FunClauseList, Vars, []).

munge_fun_clauses([{clause,Line,Head,Guard,Body} | T], Vars, Acc) ->
    {BodyRes, NewVars} = munge_body(Body, Vars#vars{line=Line}),
    munge_fun_clauses(T, NewVars, 
		      [{clause,Line,Head,Guard,BodyRes} | Acc]);
munge_fun_clauses([], Vars, Acc) ->
    {lists:reverse(Acc), Vars}.

add_bump(Vars) ->
    {[{call, 0, {remote, 0, {atom,0,coast}, {atom,0,bump}},
       [{atom,    0, Vars#vars.module},
	{atom,    0, Vars#vars.function},
	{integer, 0, Vars#vars.arity},
	{integer, 0, Vars#vars.clause}, 
	{integer, 0, Vars#vars.index},
	{integer, 0, Vars#vars.line}]
      }],
     Vars#vars{index=(Vars#vars.index)+1, lines=(Vars#vars.lines)+1}
    }.

windup(Mode, OutFd, Vars) ->
    case Mode of
	pretty ->
	    ignore;
	not_pretty ->
	    InitInfo = lists:reverse(Vars#vars.init_info),
	    Form1 = {function,0,'coast_init_info_$$_private',0, 
		     [{clause,0,[],[],
		       [erl_parse:abstract(InitInfo)]}]},
	    io:format(OutFd, "~s~n", [erl_pp:form(Form1)])
    end.


%%%--Analysis------------------------------------------------------------

%% mod_coverage(Module, Clauses) -> {ok, Result}
%%   Module = atom()
%%   Clauses = [{Module,Function,Arity,Clause,Lines}]
%%   Result = {Module,Cov,NotCov}
mod_coverage(Module, Clauses) ->
    {ok, FuncCoverage} = func_coverage(Module, Clauses),
    {CovTot, NotCovTot} =
	lists:foldl(fun({_MFA,Cov,NotCov}, {AccCov,AccNotCov}) ->
			    {AccCov+Cov, AccNotCov+NotCov}
		    end,
		    {0, 0},
		    FuncCoverage),
    {ok, {Module, CovTot, NotCovTot}}.

%% func_coverage(Module) -> {ok, Result}
%%   Module = atom()
%%   Clauses = [{Module,Function,Arity,Clause,Lines}]
%%   Result = [{{Module,Func,Arity},Cov,NotCov}]
func_coverage(Module, Clauses) ->
    {ok, ClauseCoverage} = clause_coverage(Module, Clauses),
    Result = func_coverage1(ClauseCoverage),
    {ok, Result}.

func_coverage1([{{M,F,A,C1}, Cov1, NotCov1},
		{{M,F,A,C2}, Cov2, NotCov2} | T]) ->
    func_coverage1([{{M,F,A,C2}, Cov1+Cov2, NotCov1+NotCov2} |T]);
func_coverage1([{{M,F,A,C}, Cov, NotCov} | T]) ->
    [{{M,F,A}, Cov, NotCov} | func_coverage1(T)];
func_coverage1([]) ->
    [].

%% clause_coverage(Module) -> {ok, Result}
%%   Module = atom()
%%   Clauses = [{Module,Function,Arity,Clause,Lines}]
%%   Result = [{{Module,Func,Arity,Clause},Cov,NotCov}]
clause_coverage(Module, Clauses) ->
    Fun = fun({M,F,A,C,L}) ->
		  Pttern = {#bump{module=M,function=F,arity=A,clause=C},'_'},
		  case ets:match_object(?TABLE, Pttern) of
		      [] ->
			  {{M,F,A,C}, 0, L};
		      Bumps ->
			  Cov = length(Bumps),
			  {{M,F,A,C}, Cov, L-Cov}
		  end
	  end,
    Result = lists:map(Fun, Clauses),
    {ok, Result}.

%% mod_calls(Module, Clauses) -> {ok, Result}
%%   Module = atom()
%%   Clauses = [{Module,Function,Arity,Clause,Lines}]
%%   Result = {Module,N}
mod_calls(Module, Clauses) ->
    {ok, FuncCalls} = func_calls(Module, Clauses),
    Ntot = lists:foldl(fun({_MFA, N}, AccN) -> AccN+N end, 0, FuncCalls),
    {ok, {Module, Ntot}}.

%% func_calls(Module, Clauses) -> {ok, Result}
%%   Module = atom()
%%   Clauses = [{Module,Function,Arity,Clause,Lines}]
%%   Result = [{{Module,Func,Arity},N}]
func_calls(Module, Clauses) ->
    {ok, ClauseCalls} = clause_calls(Module, Clauses),
    Result = func_calls1(ClauseCalls),
    {ok, Result}.

func_calls1([{{M,F,A,C1}, N1}, {{M,F,A,C2}, N2} | T]) ->
    func_calls1([{{M,F,A,C2}, N1+N2} | T]);
func_calls1([{{M,F,A,C}, N} | T]) ->
    [{{M,F,A}, N} | func_calls1(T)];
func_calls1([]) ->
    [].

%% clause_calls(Module, Clauses) -> {ok, Result}
%%   Module = atom()
%%   Clauses = [{Module,Function,Arity,Clause,Lines}]
%%   Result = [{{Module,Func,Arity,Clause},N}]
clause_calls(Module, Clauses) ->
    Fun = fun({M,F,A,C,L}) ->
		  Pttern = {#bump{module=M,function=F,arity=A,clause=C},'_'},
		  case ets:match_object(?TABLE, Pttern) of
		      [] ->
			  {{M,F,A,C}, 0};
		      Bumps ->
			  {_Bump, N} = hd(lists:keysort(1, Bumps)),
			  {{M,F,A,C}, N}
		  end
	  end,
    Result = lists:map(Fun, Clauses),
    {ok, Result}.

%% analyse_to_file(Module, Clauses) -> {ok, File} | {error,Error}
%%   Module = atom()
%%   File = string()
%%   Clauses = [{Module,Function,Arity,Clause,Lines}]
%%   Error = {Where, File, Reason}
%%     Where = file
%%     Reason = term()
analyse_to_file(Module, Clauses) ->
    Attributes = Module:module_info(attributes),
    {value, {coast_obj_dir,ObjDir}} =
	lists:keysearch(coast_obj_dir, 1, Attributes),

    ModuleS = atom_to_list(Module),
    OutDir = filename:join(ObjDir, "COAST"),
    PrettyDir = filename:join(OutDir, "tmp_code"),
    OutFile = filename:join(OutDir, ModuleS++".COAST.out"),
    PrettyFile = filename:join(PrettyDir, ModuleS++".COAST.pretty.erl"),

    case file:read_file_info(PrettyFile) of
	{ok, FileInfo} when FileInfo#file_info.type == regular,
			    FileInfo#file_info.access /= none ->
	    analyse_to_file(Module, Clauses, PrettyFile, OutFile,Attributes);
	{ok, FileInfo} when FileInfo#file_info.type /= regular ->
	    {error,
	     {file, PrettyFile, {bad_file_type, FileInfo#file_info.type}}};
	{ok, FileInfo} when FileInfo#file_info.access == none ->
	    {error, {file, PrettyFile, no_access}};
	{error, Reason} ->
	    {error, {file, PrettyFile, Reason}}
    end.
    
analyse_to_file(Module, Clauses, PrettyFile, OutFile, Attributes) ->
    {ok, InFd}  = file:open(PrettyFile, read),
    {ok, OutFd} = file:open(OutFile, write),
    
    {value, {coast_src_file, SrcFile}} =
	lists:keysearch(coast_src_file, 1, Attributes),
    io:format(OutFd, "SOURCE FILE: ~s~n", [SrcFile]),
    case lists:keysearch(vsn, 1, Attributes) of
	false ->
	    ignore;
	{value, {vsn, Version}} ->
	    io:format(OutFd, "VERSION: ~p~n", [Version])
    end,
    io:format(OutFd, "~n", []),
    
    print_coverage(OutFd, func_coverage(Module, Clauses)),
    print_calls(OutFd, func_calls(Module, Clauses)),

    io:format(OutFd, "NUMBER OF TIMES EACH~nLINE HAS BEEN EXECUTED:~n", []),
    print_separator(OutFd),
    
    case ets:match(?TABLE, {#bump{module=Module, index=0}, '$1'}) of
	[] ->
	    ignore;
	[[FirstL]] -> 
	    scan_to_line(FirstL, 1, InFd),
	    print_lines(Module, InFd, OutFd, FirstL-1)
    end,
    file:close(InFd),
    file:close(OutFd),
    {ok, OutFile}.

print_calls(OutFd, {ok,[]}) ->
    ignore;
print_calls(OutFd, {ok,FuncCalls}) ->
    MapFun = 
	fun({{M,F,A},N}) ->
		{atom_to_list(M)++":"++atom_to_list(F)++
		 "/"++integer_to_list(A),
		 integer_to_list(N)}
	end,
    LenFun = fun({S,N}, {M1,M2}) ->
		     {max(length(S),M1), max(length(N),M2)}
	     end,
    LenFunStartTuple = {length("Function"), length("Calls")},
    IoStrFun = fun({M1,M2}) ->
		       "  ~-" ++ integer_to_list(M1) ++ "s" ++ "     ~" ++ 
			   integer_to_list(M2) ++ "s~n"
	       end,
    HeaderIoArgs = ["Function","Calls"],
    ForeachFun = fun(IoStr, Fd) ->
			 fun({MFA,N}) ->
				 io:format(Fd, IoStr, [MFA,N]);
			    (X) ->
				 done
			 end
		 end,

    io:format(OutFd, "NUMBER OF CALLS TO EACH FUNCTION:~n", []),
    print_generic(OutFd, MapFun, LenFun, LenFunStartTuple, ForeachFun, 
		  IoStrFun, HeaderIoArgs, FuncCalls),
   
    io:format(OutFd, "~n", []).
    
print_coverage(OutFd, {ok,[]}) ->
    ignore;
print_coverage(OutFd, {ok,FuncCalls}) ->
    MapFun = 
	fun({{M,F,A},Cov,NotCov}) ->
		{atom_to_list(M)++":"++atom_to_list(F)++
		 "/"++integer_to_list(A),
		 integer_to_list(Cov) ++ "     ", 
		 integer_to_list(NotCov) ++ "     "}
	end,
    LenFun = 
	fun({S,C,UC}, {M1,M2,M3}) ->
		{max(length(S),M1), max(length(C),M2), max(length(UC),M3)}
	end,
    LenFunStartTuple = {length("Function"),
			length("Covered lines"),
			length("Not covered lines")
		       },
    IoStrFun = 
	fun({M1,M2,M3}) ->
		"  ~-" ++ integer_to_list(M1) ++ "s" ++ "        ~" ++ 
		    integer_to_list(M2) ++ "s        ~" ++
		    integer_to_list(M3) ++ "s~n"
	end,
    ForeachFun = 
	fun(IoStr, Fd) ->
		fun({MFA,Cov,NotCov}) ->
			io:format(Fd, IoStr, [MFA,Cov,NotCov]);
		   (X) ->
			done
		end
	end,
    HeaderIoArgs = ["Function", "Covered lines", "Not covered lines"],
    io:format(OutFd, "COVERAGE DATA:~n", []),
    print_generic(OutFd, MapFun, LenFun, LenFunStartTuple, ForeachFun, 
		  IoStrFun, HeaderIoArgs, FuncCalls),
    io:format(OutFd, "~n", []).

max(Curr, Max) when Curr > Max ->
    Curr;
max(Curr, Max) ->
    Max.

print_generic(Fd, MapFun, LenFun, LenFunStartTuple, 
	      ForeachFun, IoStrFun, HeaderIoArgs, DataList) ->
    FcdL2    = lists:map(MapFun, DataList),
    MaxTuple = lists:foldl(LenFun, LenFunStartTuple, FcdL2),
    IoStr    = IoStrFun(MaxTuple),
    print_separator(Fd),
    io:format(Fd, IoStr ++ "~n", HeaderIoArgs),
    lists:foreach(ForeachFun(IoStr, Fd), FcdL2),
    print_separator(Fd).
    
print_separator(OutFd) ->
        io:format(OutFd, 
		  "**************************************"
		  "**************************************~n", []).

scan_to_line(L, L, InFd) -> ok;
scan_to_line(L, CurrL, InFd) ->
    io:get_line(InFd, ''),
    scan_to_line(L, CurrL+1, InFd).

tab() ->  "        |  ".
fill() ->       "..|  ".

print_lines(Module, InFd, OutFd, L) ->
    case io:get_line(InFd, '') of
	eof ->
	    ignore;
	Line ->
	    case ets:match(?TABLE, {#bump{module=Module, line=L}, '$1'}) of
		[[N]] ->
		    Str = string:right(integer_to_list(N), 6, 32),
		    io:format(OutFd, "~s", [Str++fill()++Line]);
		[] ->
		    io:format(OutFd, "~s", [tab()++Line])
	    end,
	    print_lines(Module, InFd, OutFd, L+1)
    end.
