%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(coast_analyse_funcs).

%% External exports
-export([mod_calls/1, func_calls/1, clause_calls/1, 
	 mod_coverage/1, func_coverage/1, clause_coverage/1,
	 analyse_to_file/1]).

-include_lib("kernel/include/file.hrl").
-include("coast_server.hrl").

%%%----------------------------------------------------------------------
%%% External exports
%%%----------------------------------------------------------------------

%% mod_calls(Modules) -> {module_calls, Result} |
%%                       {error, {not_coast_compiled,Module}}
%%   Modules = [Module]
%%     Module = atom()
%%   Result = [{Module,N}]
mod_calls(Modules) ->
    case func_calls(Modules) of
	{function_calls, Result} ->
	    mod_calls(Result, []);
	Error ->
	    Error
    end.

%% func_calls(Modules) -> {function_calls, Result} |
%%                        {error, {not_coast_compiled,Module}}
%%   Modules = [Module]
%%     Module = atom()
%%   Result = [{{Module,Func,Arity},N}]
func_calls(Modules) ->
    case clause_calls(Modules) of
	{clause_calls, Result} ->
	    func_calls(Result, []);
	Error ->
	    Error
    end.

%% clause_calls(Modules) -> {clause_calls, Result} |
%%                          {error, {not_coast_compiled,Module}}
%%   Modules = [Module]
%%     Module = atom()
%%   Result = [{{Module,Func,Arity,Clause},N}]
clause_calls(Modules) ->
    clause_calls(Modules, []).

%% mod_coverage(Modules) -> {module_coverage, Result} |
%%                          {error, {not_coast_compiled,Module}}
%%   Modules = [Module]
%%     Module = atom()
%%   Result = [{Module,Cov,NotCov}]
mod_coverage(Modules) ->
    case func_coverage(Modules) of
	{function_coverage, Result} ->
	    mod_coverage(Result, []);
	Error ->
	    Error
    end.

%% func_coverage(Modules) -> {function_coverage, Result} |
%%                           {error, {not_coast_compiled,Module}}
%%   Modules = [Module]
%%     Module = atom()
%%   Result = [{{Module,Func,Arity},Cov,NotCov}]
func_coverage(Modules) ->
    case clause_coverage(Modules) of
	{clause_coverage, Result} ->
	    func_coverage(Result, []);
	Error ->
	    Error
    end.

%% clause_coverage(Modules) -> {clause_coverage, Result} |
%%                             {error, {not_coast_compiled,Module}}
%%   Modules = [Module]
%%     Module = atom()
%%   Result = [{{Module,Func,Arity,Clause},Cov,NotCov}]
clause_coverage(Modules) ->
    clause_coverage(Modules, []).

%% analyse_to_file(Modules) -> {ok, Files} | {error,Error}
%%   Modules = [Module]
%%     Module = atom()
%%   Files = [File]
%%     File = string()
%%   Error = {not_coast_compiled,Module} | {Where, File, Reason}
%%     Where = file
%%     Reason = term()
analyse_to_file(Modules) ->
    case ok_to_analyse(Modules, []) of
	{ok, Result} ->
	    analyse_to_file(Result, []);
	Error ->
	    Error
    end.


%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

mod_calls([{{Module,_F,_A}, N1}, {{Module,F,A}, N2} | T], Acc) ->
    mod_calls([{{Module,F,A}, N1+N2} | T], Acc);
mod_calls([{{Module,_F,_A}, N}  | T], Acc) ->
    mod_calls(T, [{Module, N} | Acc]);
mod_calls([], Acc) ->
    {module_calls, lists:keysort(1, Acc)}.

func_calls([{{M,F,A,C1}, N1}, {{M,F,A,C2}, N2} | T], Acc) ->
    func_calls([{{M,F,A,C2}, N1+N2} | T], Acc);
func_calls([{{M,F,A,_C}, N} | T], Acc) ->
    func_calls(T, [{{M,F,A}, N} | Acc]);
func_calls([], Acc) ->
    {function_calls, lists:keysort(1, Acc)}.

clause_calls([Module | T], Acc) ->
    case get_call_data_list(Module) of
	L1 when list(L1) ->
	    L2 = lists:sort(L1),
	    L3 = merge_clause_calls(L2, []),
	    clause_calls(T, Acc ++ L3);
	Error ->
	    Error
    end;
clause_calls([], Acc) ->
    {clause_calls, lists:keysort(1, Acc)}.

merge_clause_calls([[M,F,A,C,I,N], [M,F,A,C,_,_] | T], Acc) ->
    merge_clause_calls([[M,F,A,C,I,N] | T], Acc);
merge_clause_calls([[M,F,A,C,I,N] | T], Acc) ->
    merge_clause_calls(T, [{{M,F,A,C}, N} | Acc]);
merge_clause_calls([], Acc) ->
    Acc.

mod_coverage([{{Module,_F,_A}, Cov1, NotCov1}, 
	      {{Module,F,A},   Cov2, NotCov2} | T], Acc) ->
    mod_coverage([{{Module,F,A}, Cov1+Cov2, NotCov1+NotCov2} | T], Acc);
mod_coverage([{{Module,_F,_A}, Cov, NotCov} | T], Acc) ->
    mod_coverage(T, [{Module, Cov, NotCov} | Acc]);
mod_coverage([], Acc) ->
    {module_coverage, lists:keysort(1, Acc)}.

func_coverage([{{M,F,A,C1}, Cov1, NotCov1}, 
	       {{M,F,A,C2}, Cov2, NotCov2} | T], Acc) ->
    func_coverage([{{M,F,A,C2}, Cov1+Cov2, NotCov1+NotCov2} | T], Acc);
func_coverage([{{M,F,A,C},  Cov,  NotCov} | T], Acc) ->
    func_coverage(T, [{{M,F,A}, Cov, NotCov} | Acc]);
func_coverage([], Acc) ->
    {function_coverage, lists:keysort(1, Acc)}.

clause_coverage([Module | T], Acc) ->
    case get_call_data_list(Module) of
	L1 when list(L1) ->
	    L2 = lists:sort(L1),
	    L3 = merge_clause_coverage(L2, 0, 0, []),
	    clause_coverage(T, Acc ++ L3);
	Error ->
	    Error
    end;
clause_coverage([], Acc) ->
    {clause_coverage, lists:keysort(1, Acc)}.

merge_clause_coverage([[M,F,A,C,_,0], [M,F,A,C,I,N] | T], Cov, NotCov, Acc)->
    merge_clause_coverage([[M,F,A,C,I,N] | T], Cov, NotCov+1, Acc);
merge_clause_coverage([[M,F,A,C,_,_], [M,F,A,C,I,N] | T], Cov, NotCov, Acc)->
    merge_clause_coverage([[M,F,A,C,I,N] | T], Cov+1, NotCov, Acc);
merge_clause_coverage([[M,F,A,C,I,0] | T], Cov, NotCov, Acc) ->
    merge_clause_coverage(T, 0, 0, [{{M,F,A,C}, Cov, NotCov+1} | Acc]);
merge_clause_coverage([[M,F,A,C,I,N] | T], Cov, NotCov, Acc) ->
    merge_clause_coverage(T, 0, 0, [{{M,F,A,C}, Cov+1, NotCov} | Acc]);
merge_clause_coverage([], _, _, Acc) ->
    Acc.

get_call_data_list(Module) ->
    case catch Module:'coast_init_info_$$_private'() of
	{'EXIT', _Reason} ->
	    {error, {not_coast_compiled, Module}};
	IndexList ->
	    case ets:lookup(?TABLE_NAME, {module_initialised, Module}) of
		[{{module_initialised, Module}, true}] ->
		    ignore;
		[] ->
		    initialise_table(IndexList),
		    ets:insert(?TABLE_NAME,
			       {{module_initialised, Module}, true})
	    end,

	    %% We want a list of lists, where each sublist has the format 
	    %% [Module, Function, Arity, Clause, StatementIndex, N]
	    %% The reason for keeping the StatementIndex is that we want to
	    %% keep the original relative order of statements, since
	    %% the first bump-call for each clause tells how many times that
	    %% specific clause has been called. The other bump-calls for
	    %% that clause may be uncertain since, e.g. there may be
	    %% statements that never are entered! 
	    L = ets:match(?TABLE_NAME,
			  {{Module,'$2','$3','$5','$4','_','_'}, '$6'}),
	    lists:map(fun(DataList) ->
			      [Module|DataList]
		      end,
		      L)
    end.

initialise_table([Index | T]) ->
    case ets:lookup(?TABLE_NAME, Index) of
	[] ->
	    ets:insert(?TABLE_NAME, {Index, 0});
	_Other ->
	    done
    end,
    initialise_table(T);
initialise_table([]) ->
    done.
    
ok_to_analyse([Module | T], Acc) ->
    case catch apply(Module, 'coast_init_info_$$_private', []) of
	{'EXIT', Reason} ->
	    {error, {not_coast_compiled, Module}};
	IndexList ->
	    case ets:lookup(?TABLE_NAME, {module_initialised, Module}) of
		[{{module_initialised, Module}, true}] ->
		    ignore;
		[] ->
		    initialise_table(IndexList),
		    ets:insert(?TABLE_NAME,
			       {{module_initialised, Module}, true})
	    end,
	    
	    ModuleInfo = Module:module_info(attributes),
	    {value, {coast_obj_dir, ObjDir}} =
		lists:keysearch(coast_obj_dir, 1, ModuleInfo),
	    
	    CoastSrcDir = filename:join(ObjDir, ?SRC_DIR),
	    CoastDir = filename:join(ObjDir, ?DIR),
	    
	    ModuleS = atom_to_list(Module),
	    Pretty = filename:join(CoastSrcDir, ModuleS++?PRETTY_SUFFIX),
	    OutFile = filename:join(CoastDir, ModuleS++?ANALYSE_SUFFIX),
	    case file:read_file_info(Pretty) of
		{ok, FileInfo} when FileInfo#file_info.type == regular,
				    FileInfo#file_info.access /= none ->

		    Info = {Module, Pretty, OutFile, ModuleInfo},
		    ok_to_analyse(T, [Info | Acc]);
		{error, Reason} ->
		    {error, {file, Pretty, Reason}}
	    end
    end;
ok_to_analyse([], Acc) ->
    {ok, Acc}.

analyse_to_file([Info | T], Acc) ->
    case analyse_to_file2(Info) of
	{ok, OutFile} ->
	    analyse_to_file(T, [OutFile | Acc]);
	Error ->
	    Error
    end;
analyse_to_file([], Acc) ->
    {ok, Acc}.
    
analyse_to_file2({Module, Pretty, OutFile, ModuleInfo}) ->
    {ok, InFd}  = file:open(Pretty, read),
    {ok, OutFd} = file:open(OutFile, write),
    
    io:format(OutFd, "~n", []),
    
    {value, {coast_compiled, [{Module, SrcFile}]}} =
	lists:keysearch(coast_compiled, 1, ModuleInfo),
    io:format(OutFd, "   SOURCE FILE:  ~s~n~n~n~n", [SrcFile]),
    
    print_call_statistics(OutFd, func_calls([Module])),
    print_func_cover_data(OutFd, func_coverage([Module])),

    io:format(OutFd,
	      "   NUMBER OF TIMES EACH~n"
	      "   LINE HAS BEEN EXECUTED:~n", []),
    print_separator(OutFd),
    case lists:keysearch(vsn, 1, ModuleInfo) of
	false ->
	    ignore;
	{value, {vsn, Version}} ->
	    io:format(OutFd, "   Version:      ~p~n~n~n~n~n", [Version])
    end,

    case ets:match(?TABLE_NAME,
		   {{Module,'_','_',0,'_','_','$1'},'_'}) of
	[] ->
	    ignore;
	[[FirstFcnLine]] -> 
	    MaxNofCalls = hd(lists:reverse(
			       lists:sort(
				 lists:flatten(
				   ets:match(?TABLE_NAME, 
					     {{'_','_','_','_','_','_','_'}, 
					      '$1'}))))),
	    MaxNofCallsStrLength = length(integer_to_list(MaxNofCalls)),
	    scan_to_line(FirstFcnLine, 1, InFd),
	    read_file(InFd, io:get_line(InFd, ''), FirstFcnLine, OutFd,
		      Module, undefined, undefined, MaxNofCallsStrLength)
    end,
    file:close(InFd),
    file:close(OutFd),
    {ok, OutFile}.

print_call_statistics(OutFd, {_,[]}) ->
    done;
print_call_statistics(OutFd, {_,FcdL}) ->
    MapFun = 
	fun({{M,F,A},N}) ->
		{atom_to_list(M)++":"++atom_to_list(F)++
		 "/"++integer_to_list(A),
		 integer_to_list(N)}
	end,
    LenFun = fun({S,N}, {M1,M2}) ->
		     {max(length(S),M1), max(length(N),M2)}
	     end,
    LenFunStartTuple = {length("Function"), length("Calls")},
    IoStrFun = fun({M1,M2}) ->
		       "   ~-" ++ integer_to_list(M1) ++ "s" ++ "     ~" ++ 
			   integer_to_list(M2) ++ "s~n"
	       end,
    HeaderIoArgs = ["Function","Calls"],
    ForeachFun = fun(IoStr, Fd) ->
			 fun({MFA,N}) ->
				 io:format(Fd, IoStr, [MFA,N]);
			    (X) ->
				 done
			 end
		 end,

    io:format(OutFd, "   NUMBER OF CALLS TO EACH FUNCTION:~n", []),
    print_generic(OutFd, MapFun, LenFun, LenFunStartTuple, ForeachFun, 
		  IoStrFun, HeaderIoArgs, FcdL),
   
    io:format(OutFd, "~n~n~n", []).
    
print_func_cover_data(_OutFd, {_,[]}) ->
    done;
print_func_cover_data(OutFd, {_,FcdL}) ->
    MapFun = 
	fun({{M,F,A},Cov,UnCov}) ->
		{atom_to_list(M)++":"++atom_to_list(F)++
		 "/"++integer_to_list(A),
		 integer_to_list(Cov) ++ "     ", 
		 integer_to_list(UnCov) ++ "     "}
	end,
    LenFun = 
	fun({S,C,UC}, {M1,M2,M3}) ->
		{max(length(S),M1), max(length(C),M2), max(length(UC),M3)}
	end,
    LenFunStartTuple = {length("Function"),
			length("Covered lines"),
			length("Uncovered lines")
		       },
    IoStrFun = 
	fun({M1,M2,M3}) ->
		"   ~-" ++ integer_to_list(M1) ++ "s" ++ "        ~" ++ 
		    integer_to_list(M2) ++ "s        ~" ++
		    integer_to_list(M3) ++ "s~n"
	end,
    ForeachFun = 
	fun(IoStr, Fd) ->
		fun({MFA,Cov,UnCov}) ->
			io:format(Fd, IoStr, [MFA,Cov,UnCov]);
		   (X) ->
			done
		end
	end,
    HeaderIoArgs = ["Function", "Covered lines", "Uncovered lines"],
    io:format(OutFd, "   COVERAGE DATA:~n", []),
    print_generic(OutFd, MapFun, LenFun, LenFunStartTuple, ForeachFun, 
		  IoStrFun, HeaderIoArgs, FcdL),
    io:format(OutFd, "~n~n~n~n", []).

max(Curr, Max) when Curr > Max ->
    Curr;
max(Curr, Max) ->
    Max.

print_generic(Fd, MapFun, LenFun, LenFunStartTuple, 
	      ForeachFun, IoStrFun, HeaderIoArgs, DataList) ->
    FcdL2    = lists:map(MapFun, DataList),
    MaxTuple = lists:foldl(LenFun, LenFunStartTuple, FcdL2),
    IoStr    = IoStrFun(MaxTuple),
    print_separator(Fd),
    io:format(Fd, IoStr ++ "~n", HeaderIoArgs),
    lists:foreach(ForeachFun(IoStr, Fd), FcdL2),
    print_separator(Fd).
    
print_separator(OutFd) ->
        io:format(OutFd, 
		  "   ********************************************"
		  "*****************************~n", []).

scan_to_line(NofLines, CurrentLine, InFd) when CurrentLine < NofLines ->
    io:get_line(InFd, ''),
    scan_to_line(NofLines, CurrentLine + 1, InFd);
scan_to_line(_NofLines, _CurrentLine, _InFd) ->
    ok.

read_file(InFd, eof, LineNo, OutFd, Mod, Func, Arity, MaxLengthN) ->
    io:format(OutFd, "~n", []);
read_file(InFd, Line, LineNo, OutFd, Mod, Func, Arity, MaxLengthN) ->
    Tab1 = 3,
    Tab2 = MaxLengthN + 2,
    Tab3 = 7,
    Tab4 = 2,

    {NewFunc, NewArity} = 
	case ets:match(?TABLE_NAME,
		       {{Mod,'$1','$2','_','_','_',LineNo},'$3'}) of
	    [[F, A, N]] ->
		io:format(OutFd, "~s", 
			  [lists:duplicate(Tab1 + Tab2 + Tab3 + Tab4, " ") ++ 
			   "|  " ++ Line]),
		NStr = integer_to_list(N),
		io:format(OutFd, "~s|~n", [lists:duplicate(Tab1, " ") ++ NStr ++ 
					   lists:duplicate(Tab2 - length(NStr), 
							   " ") ++ 
					   lists:duplicate(Tab3 + Tab4, ".")]),
		{F, A};
	    [] ->
		case regexp:match(Line, "%%%%%--> End of function ") of
		    {match, _, _} ->
			io:format(OutFd, "~s~n", 
				  [lists:duplicate(Tab1 + Tab2 + Tab3 + 
						   Tab4, " ") ++ "|  "]),
			io:format(OutFd, "~s~n", 
				  [lists:duplicate(Tab1 + Tab2 + Tab3 + 
						   Tab4, " ") ++ "|  "]),
			{Func, Arity};
		    _Other ->
			io:format(OutFd, "~s", 
				  [lists:duplicate(Tab1 + Tab2 + Tab3 + 
						   Tab4, " ") ++ 
				   "|  " ++ Line]),
			{Func, Arity}
		end
	end,
    read_file(InFd, io:get_line(InFd, ''), LineNo + 1, OutFd, 
	      Mod, NewFunc, NewArity, MaxLengthN).
