%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(coast_compile_funcs).

%%%----------------------------------------------------------------------
%%%----------------------------------------------------------------------

%% External exports
-export([compile/2, remove_from_database/1]).

-include("coast_server.hrl").
-include_lib("kernel/include/file.hrl").

-record(vars, {module,        % atom() Module name
	       file,          % string() File name (incl dir and ext)
	       obj_dir,       % string() Object code directory
	       
	       pretty_mode,   % pretty | not_pretty
	       max_bumps,     % int()
	       init_info,     % [{Mod,Name,Arity,N,ClsNo,ClsLine,CodeLine}]
	       all_funcs,     % [{Name,Arity,N}]
	       line_no,       % int()
	       fcn_name,      % atom()
	       arity,         % int()
	       clause_no,     % int()
	       clause_line,   % int()
	       call_depth,    % int()
	       call_direction % internal | external  Used when call_depth=1
	      }).

%%%----------------------------------------------------------------------
%%% External exports
%%%----------------------------------------------------------------------

%% compile(File, Options) -> {ok, Module} | {error, Error}
%%   File = string()  (including directory path and .erl extension)
%%   Options =
%%   Module = atom()  (module name only)
%%   Error = {Where, File, Reason}
%%     Where = file | directory | transform | compile
%%     Reason = term()
compile(File, Options) ->
    
    %% Check that the file exists, is a regular file and is accessible
    case check_file(File) of
	ok ->
	    %% FIX Det mesta nedan borde fixas pa annan plats
	    
	    ModuleS = filename:basename(File, ".erl"),
	    Module = list_to_atom(ModuleS),

	    %% Remove earlier information about the module from the database
	    remove_from_database(Module),

	    %% Sort out the options to get object file directory,
	    %% include directories and defines.
	    {ok, CWD} = file:get_cwd(),
	    SrcDir = filename:dirname(File),
	    Default = {CWD, [SrcDir], []},
	    {ObjDir, IncDirs, Defs} = sort_out_options(Options, Default),
	    
	    Vars = #vars{module = Module, file = File, obj_dir = ObjDir},

	    CoastDir = filename:join(ObjDir, ?DIR),
	    CoastSrcDir = filename:join(ObjDir, ?SRC_DIR),
	    PrettyFile = filename:join(CoastSrcDir, ModuleS++?PRETTY_SUFFIX),
	    SrcFile = filename:join(CoastSrcDir, ModuleS++?SRC_SUFFIX),
	    ResObjFile = filename:join(ObjDir, ModuleS++?SUFFIX++".beam"),
	    ObjMod = filename:join(ObjDir, ModuleS),
	    ObjFile = ObjMod++".beam",

	    %% Create coast directories in the output directory
	    case make_dirs(CoastDir, CoastSrcDir) of
		ok ->
		    case make_srcfile(File, PrettyFile, SrcFile, IncDirs,
				      Defs, Vars) of
			ok ->
			    Options2 = [report_errors,
					report_warnings | Options],
			    case compile:file(SrcFile, Options2) of
				{ok, Module} ->
				    file:rename(ResObjFile, ObjFile),
				    code:purge(Module),
				    code:load_abs(ObjMod),
				    {ok, Module};
				{error, Reason} ->
				    {error, {compile, File, Reason}}
			    end;
			{error, Reason} ->
			    {error, {transform, File, Reason}}
		    end;
		{error, Reason} ->
		    {error, {directory, File, Reason}}
	    end;
	{error, Reason} ->
	    {error, {file, File, Reason}}
    end.

%% remove_from_database(Module)
%%   Module = atom()
%% Removes all information about Module from the ets table.
remove_from_database(Module) ->
    ets:delete(?TABLE_NAME, {module_initialised, Module}),
    ets:match_delete(?TABLE_NAME,
		     {{Module,'_','_','_','_','_','_'},'_'}).


%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

%% check_file(File) -> ok | {error,Reason}
%%   File = string()
%%   Reason = term()
%% Returns ok if File is a readable file.
check_file(File) ->
    case file:read_file_info(File) of
	{ok, FileInfo} when FileInfo#file_info.type == regular,
			    FileInfo#file_info.access /= none ->
	    ok;
	{ok, FileInfo} when FileInfo#file_info.type /= regular ->
	    {error, {bad_file_type, FileInfo#file_info.type}};
	{ok, FileInfo} when FileInfo#file_info.access == none ->
	    {error, no_access};
	Error ->
	    Error
    end.

%% sort_out_options(Options, Default) -> Actual
%%   Options = [Option]
%%     Option = <compiler option>, see compile:file/2
%%   Default = Actual = {ObjDir, IncDirs, Defs}
%%     ObjDir = string()
%%     IncDirs = [string()]
%%     Defs = [term()]
%% Takes a list of compiler options and return a tuple with:
%% * object/output directory as specified by the option key 'outdir',
%%   replacing the default value.
%% * a list of directories to search for include files as specified by
%%   the option key 'i', appended to the default value.
%% * a list of terms to define as specified by the option key 'd', appended
%%   to the default value.
sort_out_options([{outdir,ObjDir}|Options], {_ObjDir, IncDirs, Defs}) ->
    sort_out_options(Options, {ObjDir, IncDirs, Defs});
sort_out_options([{i,Dir}|Options], {ObjDir, IncDirs, Defs}) ->
    sort_out_options(Options, {ObjDir, IncDirs++[Dir], Defs});
sort_out_options([{d,Macro}|Options], {ObjDir, IncDirs, Defs}) ->
    sort_out_options(Options, {ObjDir, IncDirs, Defs++[{Macro,true}]});
sort_out_options([{d,Macro,Value}|Options], {ObjDir, IncDirs, Defs}) ->
    sort_out_options(Options, {ObjDir, IncDirs, Defs++[{Macro,Value}]});
sort_out_options([Option|Options], Result) ->
    sort_out_options(Options, Result);
sort_out_options([], Result) ->
    Result.

%% make_dirs(CoastDir, CoastSrcDir) -> ok | {error,Reason}
%%   CoastDir = CoastSrcDir = string()
%%   Reason = term()
%% Create the coast directories.
make_dirs(CoastDir, CoastSrcDir) ->
    case file:make_dir(CoastDir) of
	ok ->
	    file:make_dir(CoastSrcDir);
	{error, eexist} ->
	    ok;
	Error ->
	    Error
    end.

%% make_srcfile(File, SrcDir, ObjDir, IncDirs, Defs, Vars)
%%                                                    -> ok | {error, Reason}
%%   File = string()  (including directory path and .erl extension)
%%   SrcDir = string()
%%   ObjDir = string()
%%   IncDirs = [string()]
%%   Defs = [{atom(),term()}]
%%   Vars = #var{}
%%   Reason = term()
%% Create a source file.
make_srcfile(File, PrettyFile, SrcFile, IncDirs, Defs, Vars) ->
    case transform(pretty, File, PrettyFile, IncDirs, Defs, Vars) of
	ok ->
	    transform(not_pretty, PrettyFile, SrcFile, [], [], Vars);
	Error ->
	    Error
    end.

%% transform(Mode,InFile,OutFile,IncDirs,Defs,Vars) -> ok | {error, Reason}
%%   Mode = pretty | not_pretty
%%   InFile = OutFile = string()  (including directory path and extension)
%%   IncDirs = [string()]
%%   Defs = [{atom(),term()}]
%%   Vars = #var{}
transform(Mode, InFile, OutFile, IncDirs, Defs, Vars) ->

    %% Open files.
    %% Note: No error check, at this stage it should already have been made
    %% certain that the files are readable/writable!
    {ok, InFd} = epp:open(InFile, IncDirs, Defs),
    {ok, OutFd} = file:open(OutFile, write),
    
    Vars1 = Vars#vars{pretty_mode = Mode,
		      max_bumps   = 0,
		      init_info   = [],
		      all_funcs   = [],
		      call_depth  = 0},
    Result = translate(Mode, InFd, OutFd, Vars1),
    epp:close(InFd),
    case Result of
	{transformed, Vars2} ->
	    windup(Mode, OutFd, OutFile, Vars2),
	    file:close(OutFd),
	    ok;
	Error ->
	    file:close(OutFd),
	    Error
    end.

%% translate(Mode, InFd, OutFd, Vars) -> {transformed,Vars2} | {error,Reason}
%%   Mode = pretty | not_pretty
%%   InFd = OutFd = file descriptor (pid())
%%   Vars = Vars2 = #vars{}
%%   Reason = term()
translate(Mode, InFd, OutFd, Vars) ->
    AbsForm = epp:parse_erl_form(InFd),
    translation_loop(Mode, AbsForm, InFd, OutFd, Vars).

translation_loop(pretty,{ok,{attribute,Line,file,Mod}},InFd,OutFd,Vars) ->
    translate(pretty, InFd, OutFd, Vars);
translation_loop(pretty, {ok,{attribute,Line,module,Mod}},InFd,OutFd,Vars) ->
    NewVars = Vars#vars{module=Mod},
    io:format(OutFd, "~n~s", [erl_pp:form({attribute,Line,module,Mod})]),
    translate(pretty, InFd, OutFd, NewVars);
translation_loop(pretty,{ok,{attribute,Line,AName,AData}},InFd,OutFd,Vars) ->
    io:format(OutFd, "~n~s", [erl_pp:form({attribute,Line,AName,AData})]),
    translate(pretty, InFd, OutFd, Vars);
translation_loop(pretty,{ok,Form},InFd,OutFd,Vars) ->
    {MungedForm, NewVars} = munge(Form, Vars),
    io:format(OutFd, "~n~n~n~s", [erl_pp:form(MungedForm)]),
    case MungedForm of
	{function,Line,Name,Arity,MungeRes} ->
	    io:format(OutFd, 
		      "~n%%%%%--> End of function ~p/~p <--%%%%%~n", 
		      [Name,Arity]);
	_OtherForm ->
	    ignore
    end,
    translate(pretty, InFd, OutFd, NewVars);

translation_loop(not_pretty,
		 {ok,{attribute,Line,module,Mod}},InFd,OutFd,Vars) ->
    NewVars = Vars#vars{module=Mod},
    io:format(OutFd, "~s", [erl_pp:form({attribute,Line,module,Mod})]),
    io:format(OutFd, "~s", [erl_pp:form({attribute,Line,coast_compiled,
					 {Mod,NewVars#vars.file}})]),
    io:format(OutFd, "~s", [erl_pp:form({attribute,Line,coast_obj_dir,
					 NewVars#vars.obj_dir})]),
    io:format(OutFd, "~s~n",
	      [erl_pp:form({attribute,Line,export,
			    [{'coast_data_$$_private',0},
			     {'coast_init_info_$$_private', 0}]})]),
    translate(not_pretty, InFd, OutFd, NewVars);
translation_loop(not_pretty,{ok,Form},InFd,OutFd,Vars) ->
    {MungedForm, NewVars} = munge(Form, Vars),
    io:format(OutFd, "~s", [erl_pp:form(MungedForm)]),
    translate(not_pretty, InFd, OutFd, NewVars);

translation_loop(_Mode,{eof,Line},InFd,OutFd,Vars) ->
    {transformed, Vars};
translation_loop(_Mode, {error,{Line,Mod,Error}},InFd,OutFd,Vars) ->
    {error, {parse_error,Line,Error}}.

%% munge(AbsForm, Vars) -> AbsForm2
%%   AbsForm = AbsForm2 = term()
%%   Vars = #vars{}
munge({attribute,Line,module,Mod}, Vars) ->
    {{attribute,Line,module,Mod}, Vars#vars{module=Mod}};
munge({function,Line,Name,Arity,Clauses}, Vars) ->
    Vars1 = Vars#vars{line_no = Line,
		      fcn_name = Name,
		      arity = Arity,
		      clause_no = 1},
    {MungeRes, Vars2} = munge_clauses(Clauses, Vars1),
    {{function,Line,Name,Arity,MungeRes}, Vars2};
munge(X, Vars) ->
    {X, Vars}.

munge_clauses(ClauseForm, Vars) ->
    munge_clauses(ClauseForm, Vars, []).

munge_clauses([{clause,Line,Head,Guard,Body} | T], Vars, MungeAcc) ->
    {BodyRes, NewVars} = 
	munge_body(Body, Vars#vars{line_no=Line,clause_line=Line}),
    NewClauseNo = NewVars#vars.clause_no + 1,
    munge_clauses(T, NewVars#vars{clause_no=NewClauseNo},
		  [{clause,Line,Head,Guard,BodyRes} | MungeAcc]);
munge_clauses([], Vars, MungeAcc) -> 
    {lists:reverse(MungeAcc), Vars}.

munge_body([H|T], Vars) ->
    {Bump, NewVars1} = 
	case Vars#vars.pretty_mode of
	    pretty ->
		{[], Vars};
	    not_pretty ->
		CurrMod = Vars#vars.module,
		case H of
		    {call,_L1,{remote,_L2,{atom,_L3,CurrMod},_Func},_Args} ->
			add_bump(Vars#vars{call_direction=internal});
		    {call, _L1, {remote, _L2, _Mod, _Func}, _Args} ->
			add_bump(Vars#vars{call_direction=external});
		    {call,_L1,{tuple,_L2,[{atom,_L3,CurrMod},_Func]},_Args}->
			add_bump(Vars#vars{call_direction=internal});
		    {call, _L1, {tuple, _L2, [_Mod, _Func]}, _Args} ->
			add_bump(Vars#vars{call_direction=external});
		    {call,
		     _L1,{atom,_L2,apply},[{atom,_L3,CurrMod},_Func,_Args]}->
			add_bump(Vars#vars{call_direction=internal});
		    {call, _L1, {atom,_L2,apply}, [_Mod, _Func, _Args]} ->
			add_bump(Vars#vars{call_direction=external});
		    _OtherH ->
			%% Direction only interesting if external,
			%% set it to 'internal' otherwise!
			add_bump(Vars#vars{call_direction=internal})
		end
	end,
    {ExprRes, NewVars2} = munge_expr(H, NewVars1),
    {BodyRes, NewVars3} = munge_body(T, NewVars2),
    {Bump ++ [ExprRes | BodyRes], NewVars3};
munge_body([], Vars) ->
    {[], Vars}.

% The following expressions are of interest:
%
%  {'if', Line, Opts}
%  {'case', Line, Func, Opts}
%  {'receive', Line, Opts, Func, TimeOutBody}
%  {'receive', Line, Opts}
%  {'catch', Line, E0}
%  {match,Line,LS,RS}
%  {call,Line,Label,Args}
%  {'fun', Line, Body}
%  {tuple, Line, Es0}
%  {op, Line, Op, A0}
%  {op, Line, Op, L0, R0}
%  {remote, Line, M0, F0}
%  {var, Line, VariabelName}
%  {TermName, Line, Value}, TermName = integer | float | atom | string
%  {nil, Line}
%
% The following expressions are probably not of interest:
%
%  {record_index,Line,Name,Field0}
%  {record,Line,Name,Inits0}
%  {record_field,Line,Rec0,Name,Field0}
%  {record,Line,Rec0,Name,Upds0}
%  {record_field,Line,Rec0,Field0}
%  {block,Line,Es0}
%
% (For further information about the forms that exist, 
% see 'erl_id_trans.erl' in the stdlib!)

munge_expr({integer,Line,Value}, Vars) ->
    {{integer,Line,Value}, Vars#vars{line_no=Line}};
munge_expr({float, Line, Value}, Vars) ->
    {{float,Line,Value}, Vars#vars{line_no=Line}};
munge_expr({atom,Line,Value}, Vars) ->
    {{atom,Line,Value}, Vars#vars{line_no=Line}};
munge_expr({string,Line,Value}, Vars) ->
    {{string,Line,Value}, Vars#vars{line_no=Line}};
munge_expr({tuple,Line,Es0}, Vars) ->
    {Es1, NewVars} = munge_tuple_expr_list(Es0, Vars#vars{line_no=Line}),
    {{tuple,Line,Es1}, NewVars};
munge_expr({var,Line,Value}, Vars) ->
    {{var,Line,Value}, Vars#vars{line_no=Line}};
munge_expr({op,Line,Op,A0}, Vars) ->
    {A1, NewVars} = munge_expr(A0, Vars#vars{line_no=Line}),
    {{op,Line,Op,A1}, NewVars};
munge_expr({op,Line,Op,L0,R0}, Vars) ->
    {L1, NewVars1} = munge_expr(L0, Vars#vars{line_no=Line}),
    {R1, NewVars2} = munge_expr(R0, NewVars1),
    {{op,Line,Op,L1,R1}, NewVars2};
munge_expr({match,Line,LS,RS}, Vars) ->
    {ExprRes1, NewVars1} = munge_expr(LS, Vars#vars{line_no=Line}),
    {ExprRes2, NewVars2} = munge_expr(RS, NewVars1),
    {{match,Line,ExprRes1,ExprRes2}, NewVars2};
munge_expr({call,Line,Label,Args}, Vars) ->
    Depth = Vars#vars.call_depth,
    {ExprRes, NewVars} = munge_call_expr_list(Args,
					      Vars#vars{line_no=Line,
							call_depth=Depth+1}),
    {{call,Line,Label,ExprRes}, NewVars#vars{call_depth=Depth}};
munge_expr({'case',Line,Func,Opts}, Vars) -> 
    {OptionRes, NewVars} = munge_option_list(Opts, Vars#vars{line_no=Line}),
    %% Since the 'end' word isn't shown in the parsing, 
    %% we have to manually count that line.
    NewLine = NewVars#vars.line_no + 1,
    {{'case',Line,Func,OptionRes}, NewVars#vars{line_no=NewLine}};
munge_expr({'if',Line,Opts}, Vars) -> 
    {OptionRes, NewVars} = munge_option_list(Opts, Vars#vars{line_no=Line}),
    %% Since the 'end' word isn't shown in the parsing, 
    %% we have to manually count that line.
    %% Can't use variable 'Line' here!
    NewLine = NewVars#vars.line_no + 1,
    {{'if',Line,OptionRes}, NewVars#vars{line_no=NewLine}};
munge_expr({'receive',Line,Opts}, Vars) -> 
    {OptionRes, NewVars} = munge_option_list(Opts, Vars#vars{line_no=Line}),
    %% Since the 'end' word isn't shown in the parsing, 
    %% we have to manually count that line.
    NewLine = NewVars#vars.line_no + 1,
    {{'receive',Line,OptionRes}, NewVars#vars{line_no=NewLine}};
munge_expr({'receive',Line,Opts,Func,TimeOutBody}, Vars) ->
    {OptionRes, NewVars1} = munge_option_list(Opts, Vars#vars{line_no=Line}),
    {FuncRes, NewVars2}   = munge_expr(Func, NewVars1),
    {BodyRes, NewVars3}   = munge_body(TimeOutBody, NewVars2),
    %% Since the 'end' word isn't shown in the parsing, 
    %% we have to manually count that line.
    NewLine = NewVars3#vars.line_no + 1,
    {{'receive',Line,OptionRes,Func,BodyRes},NewVars3#vars{line_no=NewLine}};
munge_expr({'catch',Line,E0}, Vars) ->
    {E1, NewVars} = munge_expr(E0, Vars#vars{line_no=Line}),
    {{'catch',Line,E1}, NewVars};
munge_expr({'fun',Line,{clauses,Cs0}}, Vars) ->
    {Cs1, NewVars} = munge_fun_clauses(Cs0, Vars#vars{line_no=Line}),
    {{'fun',Line,{clauses,Cs1}}, NewVars};
munge_expr({'fun',Line,FunctionForm}, Vars) ->
    {{'fun',Line,FunctionForm}, Vars#vars{line_no=Line}};
munge_expr({record,Line,Name,Inits}, Vars) ->
    {{record,Line,Name,Inits}, Vars#vars{line_no=Line}};
munge_expr({record,Line,Rec,Name,Upds}, Vars) ->
    {{record,Line,Rec,Name,Upds}, Vars#vars{line_no=Line}};
munge_expr({record_field,Line,Rec,Name,Field}, Vars) ->
    {{record_field,Line,Rec,Name,Field}, Vars#vars{line_no=Line}};
munge_expr({record_index,Line,Name,Field}, Vars) ->
    {{record_index,Line,Name,Field}, Vars#vars{line_no=Line}};
munge_expr({block,Line,Es}, Vars) ->
    {{block,Line,Es}, Vars#vars{line_no=Line}};
munge_expr({remote,Line,M0,F0}, Vars) ->
    {M1, NewVars1} = munge_expr(M0, Vars#vars{line_no=Line}),
    {F1, NewVars2} = munge_expr(F0, NewVars1),
    {{remote,Line,M1,F1}, NewVars2};
munge_expr({cons,Line,H0,T0}, Vars) ->
    {H1, NewVars1} = munge_expr(H0, Vars#vars{line_no=Line}),
    {T1, NewVars2} = munge_expr(T0, NewVars1),
    {{cons,Line,H1,T1}, NewVars2};
munge_expr({lc,Line,E0,Qs0}, Vars) ->
    {E1, NewVars1}  = munge_expr(E0, Vars#vars{line_no=Line}),
    {Qs1, NewVars2} = munge_lc_quals(Qs0, NewVars1),
    {{lc,Line,E1,Qs1}, NewVars2};
munge_expr({nil,Line}, Vars) ->
    {{nil, Line}, Vars#vars{line_no=Line}};
munge_expr({TermName,Line,Value}, Vars) ->  
    %% If we have missed some term names above...
    {{TermName,Line,Value}, Vars#vars{line_no=Line}};
munge_expr(X, Vars) ->       
    {X, Vars}.

munge_tuple_expr_list(ExprList, Vars) ->
    munge_tuple_expr_list(ExprList, Vars, []).

munge_tuple_expr_list([E0 | Es], Vars, Acc) ->
    {E1, NewVars} = munge_expr(E0, Vars),
    munge_tuple_expr_list(Es, NewVars, [E1 | Acc]);
munge_tuple_expr_list([], Vars, Acc) ->
    {lists:reverse(Acc), Vars}.

munge_call_expr_list(ExprList, Vars) ->
    munge_call_expr_list(ExprList, Vars, []).

munge_call_expr_list([E0 | Es], Vars, Acc) ->
    {E1, NewVars} = munge_expr(E0, Vars),
    munge_call_expr_list(Es, NewVars, [E1 | Acc]);
munge_call_expr_list([], Vars, Acc) ->
    {lists:reverse(Acc), Vars}.

munge_option_list(OptionList, Vars) ->
    munge_option_list(OptionList, Vars, []).

munge_option_list([{clause,Line,Guard,Body} | T], Vars, Acc) ->
    {BodyRes, NewVars} = munge_body(Body, Vars#vars{line_no=Line}),
    munge_option_list(T, NewVars, [{clause,Line,Guard,BodyRes} | Acc]);
munge_option_list([{clause,Line,Pattern,Guard,Body} | T], Vars, Acc) ->
    {BodyRes, NewVars} = munge_body(Body, Vars#vars{line_no=Line}),
    munge_option_list(T, NewVars,
		      [{clause, Line, Pattern, Guard, BodyRes} | Acc]);
munge_option_list([], Vars, Acc) -> 
    {lists:reverse(Acc), Vars};
munge_option_list(X, Vars, Acc) ->
    throw({error, {unknown_option, X}}).

munge_lc_quals(LcQualsList, Vars) ->
    munge_lc_quals(LcQualsList, Vars, []).

munge_lc_quals([{generate,Line,P0,E0} | Qs], Vars, Acc) ->
    {P1, NewVars1} = munge_expr(P0, Vars#vars{line_no=Line}),
    {E1, NewVars2} = munge_expr(E0, NewVars1),
    munge_lc_quals(Qs, NewVars2, [{generate,Line,P1,E1} | Acc]);
munge_lc_quals([E0 | Qs], Vars, Acc) ->
    {E1, NewVars} = munge_expr(E0, Vars),
    munge_lc_quals(Qs, NewVars, [E1 | Acc]);
munge_lc_quals([], Vars, Acc) ->
       %% Since the closing bracket isn't shown in the parsing, 
       %% we have to add that line manually!
    NewLine = Vars#vars.line_no + 1,
    {lists:reverse(Acc), Vars#vars{line_no=NewLine}}.

% {clause, LineNo, ListOfArgForms, ListOfGuardForms, ListOfBodyForms}
% This function is essentially the same as the function clause function, 
% the only difference being we don't want any clause numbering here!

munge_fun_clauses(FunClauseList, Vars) ->
    munge_fun_clauses(FunClauseList, Vars, []).

munge_fun_clauses([{clause,Line,Head,Guard,Body} | T], Vars, Acc) ->
    {BodyRes, NewVars} = munge_body(Body, Vars#vars{line_no=Line}),
    munge_fun_clauses(T, NewVars, 
		      [{clause,Line,Head,Guard,BodyRes} | Acc]);
munge_fun_clauses([], Vars, Acc) ->
    {lists:reverse(Acc), Vars}.

% Results in 'coast:bump(Module, 
%                        Function, 
%                        Arity, 
%                        Index(?), 
%                        ClauseNumber, 
%                        ClauseLineNumber, 
%                        CodeLineNumber)'
% where ClauseLineNumber is the line where the specific clause starts.
%
% NOTE: ClauseLineNumber is probably unnecessary information, since the same
% information can be obtained by getting the lowest line number for a given
% clause. Maybe this information will be removed later on.

add_bump(Vars) ->
    #vars{module         = Mod,
	  max_bumps      = N,
	  init_info      = InitInfo,
	  all_funcs      = AllFuncs,
	  line_no        = CodeLine,
	  fcn_name       = Name,
	  arity          = Arity,
	  clause_no      = ClauseNo,
	  clause_line    = ClauseLine,
	  call_direction = CallDirection}  = Vars,

    %% Store information necessary for "initialisation" of the database. 
    %% Please note that this initialisation takes place afterwards, 
    %% when statistics for a certain module is requested!
    InitInfo1 = [{Mod,Name,Arity,N, ClauseNo,ClauseLine,CodeLine} |InitInfo],
    NewVars = Vars#vars{max_bumps = N + 1,
			init_info = InitInfo1,
			all_funcs = [{Name, Arity, N} | AllFuncs]},
    
    {[{call, 0, {remote, 0, {atom,0,coast}, {atom,0,bump}},
       [{atom,    0, Mod},
	{atom,    0, Name},
	{integer, 0, Arity},
	{integer, 0, N},
	{integer, 0, ClauseNo}, 
	{integer, 0, ClauseLine},
	{integer, 0, CodeLine},
	{atom,    0, CallDirection}
       ]
      }],
     NewVars}.

windup(pretty, OutFd, OutFileName, Vars) ->
    done;
windup(not_pretty, OutFd, OutFileName, Vars) ->
    %% Add the info necessary to "afterwards" initialise the database.
    InitInfo = Vars#vars.init_info,
    Form1    = {function,0,'coast_init_info_$$_private',0, 
		[{clause,0,[],[],
		  [erl_parse:abstract(InitInfo)]}
		]},
    io:format(OutFd, "~s~n", [erl_pp:form(Form1)]),

    AllFuncs = Vars#vars.all_funcs,
    Form2    = {function,0,'coast_data_$$_private',0,
		[{clause,0,[],[],
		  [erl_parse:abstract({OutFileName, AllFuncs})]}
		]},
    io:format(OutFd, "~s~n", [erl_pp:form(Form2)]).
