%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(coast_server).

%%%----------------------------------------------------------------------
%%%----------------------------------------------------------------------

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
	 code_change/3]).

%% Internal exports
-export([code_runner/4]).

-include("coast_server.hrl").

-record(state, {modules=[]}).

%%%----------------------------------------------------------------------
%%% gen_server callback
%%%----------------------------------------------------------------------

init(_Args) ->
    ets:new(?TABLE_NAME, [set, public, named_table]),
    {ok, #state{}}.

handle_call({mod_calls, Modules}, _From, State) ->
    Reply = coast_analyse_funcs:mod_calls(Modules),
    {reply, Reply, State};
handle_call({func_calls, Modules}, _From, State) ->
    Reply = coast_analyse_funcs:func_calls(Modules),
    {reply, Reply, State};
handle_call({clause_calls, Modules}, _From, State) ->
    Reply = coast_analyse_funcs:clause_calls(Modules),
    {reply, Reply, State};

handle_call({mod_coverage, Modules}, _From, State) ->
    Reply = coast_analyse_funcs:mod_coverage(Modules),
    {reply, Reply, State};
handle_call({func_coverage, Modules}, _From, State) ->
    Reply = coast_analyse_funcs:func_coverage(Modules),
    {reply, Reply, State};
handle_call({clause_coverage, Modules}, _From, State) ->
    Reply = coast_analyse_funcs:clause_coverage(Modules),
    {reply, Reply, State};

handle_call({analyse_to_file, Modules}, _From, State) ->
    Reply = coast_analyse_funcs:analyse_to_file(Modules),
    {reply, Reply, State};

handle_call({compile, File, Options}, _From, State) ->
    case coast_compile_funcs:compile(File, Options) of
	{ok, Module} ->
	    Modules = add_module(Module, State#state.modules),
	    {reply, {ok, Module}, State#state{modules=Modules}};
	Error ->
	    {reply, Error, State}
    end;

handle_call({run, [Mod, Func, Args]}, From, State) ->
    spawn(?MODULE, code_runner, [From, Mod, Func, Args]),
    {noreply, State}; % reply sent from code_runner/4

handle_call(known_modules, _From, State) ->
    Reply = State#state.modules,
    {reply, Reply, State};

handle_call({source_files, Modules}, _From, State) ->
    Reply = get_source_files(Modules),
    {reply, Reply, State};

handle_call({clear, Modules}, _From, State) ->
    lists:foreach(fun(Module) ->
			  coast_compile_funcs:remove_from_database(Module)
		  end,
		  Modules),
    {reply, ok, State};
handle_call(clear_all, _From, State) ->
    %% Easiest to just kill the table and create a new!
    ets:delete(?TABLE_NAME),
    ets:new(?TABLE_NAME, [set, public, named_table]),
    {reply, ok, State}.
	
handle_cast(stop, State) ->
    {stop, normal, State}.

handle_info(null, State) ->
    {noreply, State}.

terminate(Reason, State) ->
    ok.

code_change(OldVsn, State, Extra) ->
    {ok, State}.


%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

%% add_module(Module, Modules1) -> Modules2
%%   Module = atom()
%%   Modules1 = Modules2 = [Module]
add_module(Module, Modules) ->
    case lists:member(Module, Modules) of
	true ->
	    Modules;
	false ->
	    Modules ++ [Module]
    end.
	    
code_runner(From, Mod, Func, Args) ->
    Result = (catch apply(Mod, Func, Args)),
    gen_server:reply(From, Result).

%% get_source_files(Modules) -> Files
%%   Modules = [atom()]
%%   Files = [File]
%%     File = string() | {error,{no_such_module,Module}}
%%	               | {error,{not_coast_compiled,Module}}
get_source_files(Modules) ->
    get_source_files(Modules, []).

get_source_files([Module | Modules], Files) ->
    SrcFile = case catch Module:module_info(attributes) of
		  {'EXIT', Reason} ->
		      {error, {no_such_module, Module}};
		  ModuleInfo ->
		      case lists:keysearch(coast_compiled, 1, ModuleInfo) of
			  false ->
			      {error, {not_coast_compiled, Module}};
			  {value, {coast_compiled, [{Module, File}]}} ->
			      File
		      end
	      end,
    get_source_files(Modules, [SrcFile | Files]);
get_source_files([], Files) ->
    lists:reverse(Files).
